// grunt tasks for deploying to github pages, testing that deployment, and
// rolling back if tests fail
module.exports = function (grunt) {
  "use strict";
  var spawn = require('./util/spawn'),
    runCasper = require('./util/runCasper'),
    getLastModified = require('./util/getLastModified'),
    waitForPage = require('./util/waitForPage');

  // clear the output directory, clone the latest from github and remove all
  // local files to prepare for build
  grunt.registerTask("git-pull", function () {
    grunt.config.requires('outputDir', 'git.remote', 'git.branch');
    var done = this.async(),
      outputDir = grunt.config.get('outputDir'),
      remote = grunt.config.get('git.remote'),
      branch = grunt.config.get('git.branch');
    grunt.file["delete"](outputDir, {force: true});
    grunt.file.mkdir(outputDir);
    spawn(grunt, {
      dir: outputDir,
      cmds: [
        "git init .",
        "git remote add origin " + remote,
        "git fetch " + remote + " " + branch + ":refs/remotes/origin/" + branch,
        "git checkout " + branch,
        "git rm -r ."
      ],
      succ: done,
      err: function () { done(false); }
    });
  });

  // - push changes from output directory to github
  // - wait for github pages to deploy new version
  // - run casperjs tests against the deployed app
  // - if they fail, roll back to the previous commit
  grunt.registerTask("git-push", function () {
    grunt.config.requires(
      'outputDir',
      'git.branch',
      'git.working',
      'deploy.host',
      'deploy.path',
      'casperfiles'
    );
    var done = this.async(),
      casperfiles = grunt.config.get('casperfiles'),
      outputDir = grunt.config.get('outputDir'),
      working = grunt.config.get('git.working'),
      branch = grunt.config.get('git.branch'),
      host = grunt.config.get('deploy.host'),
      path = grunt.config.get('deploy.path'),
      lastModified,
      rollback = function () {
        grunt.log.error("Deployment tests failed, rolling back change");
        spawn(grunt, {
          dir: outputDir,
          cmds: [
            "git revert -n HEAD",
            "git commit -m rollback",
            "git push origin " + branch
          ],
          succ: function () {
            grunt.log.error("Deploy failed, but rollback succeeded - app now at last state");
            done(false);
          },
          err: function () {
            grunt.log.error("Rollback failed, deployed app broken!!");
            done(false);
          }
        });
      },
      runTests = function () {
        grunt.log.ok("Page available, running tests against it");
        runCasper(grunt, {
          args: {url: "http://" + host + path, remote: true},
          files: casperfiles,
          succ: function () {
            grunt.log.ok('Tests passed! app deployed.');
            done();
          },
          err: rollback
        });
      },
      waitForDeployment = function () {
        grunt.log.writeln("Waiting for http://" + host + path + " to become available, last modified " + lastModified);
        waitForPage({
          host: host,
          path: path,
          timeout: 20,
          succ: runTests,
          lastModified: lastModified,
          err: function () {
            grunt.log.err("Timed out getting page");
            done(false);
          }
        });
      };

    // kick off deployment by getting current last modified date of
    // index.html and starting push
    getLastModified(host, path, function (last) {
      grunt.log.writeln("http://" + host + path + " last modified " + last);
      lastModified = last;
      var fail = function (msg) {
          return function () {
            grunt.log.error(msg);
            done(false);
          };
        },
        push = function () {
          spawn(grunt, {
            dir: outputDir,
            cmds: ["git push origin " + branch],
            succ: waitForDeployment,
            err: fail("Failed pushing to remote")
          });
        };
      // grab the current revision to put into commit message
      grunt.util.spawn({
        cmd: 'git',
        args: ['rev-parse', 'HEAD']
      }, function (err, result, code) {
        if (code === 0) {
          spawn(grunt, {
            dir: outputDir,
            cmds: ["git add -A .", ["git", "commit", "-m", "publish " + working + "@" + result]],
            succ: push,
            err: fail("Failed committing - maybe no change from HEAD?")
          });
        } else {
          fail("Failed getting current revision")();
        }
      });
    }, grunt);
  });
};