// miscellaneous grunt tasks
module.exports = function (grunt) {
  "use strict";
  // replace text in a file
  var fs = require('fs');
  var path = require('path');
  var existsSync = fs.existsSync || path.existsSync;
  var runCasper = require('./util/runCasper');
  var connect = require('connect');

  // print size of compiled main.js
  grunt.registerTask("size", function () {
    var done = this.async();
    fs.stat(grunt.config.get("outputDir") + "/js/main.js", function (err, stats) {
      grunt.log.ok("main.js: " + Math.round(stats.size / 10.24) / 100 + "KB");
      done();
    });
  });

  // start up a web server on localhost from the src/ folder
  grunt.registerTask("server", function (keepalive) {
    connect(connect['static'](".."), connect.directory("..")).listen(80);
    if (keepalive) {
      this.async();
    }
  });

  // verify existence of non-existence of specific files
  grunt.registerMultiTask('exists', 'File Existence/Nonexistance', function() {
    var wrong = 0;
    this.data.forEach(function(file) {
      var shouldExist = !/^\!/.test(file);
      var fileName = file.replace(/^\!/, "");
      var exists = existsSync(fileName);
      if (exists && !shouldExist) {
        grunt.log.error("Banned file [" + fileName + "] exists but should not.");
        wrong += 1;
      } else if (!exists && shouldExist) {
        grunt.log.error("Required file [" + fileName + "] does not exist but should.");
        wrong += 1;
      }
    });
    if (wrong) {
      grunt.fatal("Existence of " + wrong + "/" + this.data.length + " files incorrect");
    } else {
      grunt.log.ok("Existence of " + this.data.length + " files correct");
    }
  });

  // invoke casperjs helper
  grunt.registerMultiTask("casper", function () {
    grunt.config.requires('casperfiles');
    var PORT = 9091;
    var done = this.async();
    var local = this.data.local;
    var remote = this.data.remote;
    var files = grunt.config.get('casperfiles');

    if ((remote && !local) || (local && !remote)) {
      // if "local", start up connect webserver before and shut down after
      if (local) {
        remote = "http://localhost:" + PORT;
        (function () {
          var oldDone = done,
            connect = require('connect'),
            server = connect(connect['static'](local)).listen(PORT);
          // override done to also shut down http server
          done = function () {
            server.close();
            oldDone.apply(null, arguments);
          };
        }());
      }

      // verify correct contents of CNAME
      var checkForCNAME = function () {
        require('http').get(remote + "/CNAME", function(res) {
          res.on('data', function (chunk) {
            if (chunk.toString() === 'onthegomap.com') {
              grunt.log.ok("CNAME correct: " + chunk);
              done();
            } else {
              grunt.log.error("CNAME incorrect: " + chunk);
              done(false);
            }
          });
        }).on('error', function(e) {
          grunt.log.error("Error getting CNAME: " + e);
          done(false);
        });
      };

      // kick off casperjs
      runCasper(grunt, {
        files: files,
        args: {url: remote, remote: !!this.data.remote},
        succ: grunt.option('dev') ? done : checkForCNAME,
        err: function () {
          done(false);
        }
      });
    } else {
      grunt.log.err('One and only one "local" and "remote" should be provided for casper task');
      done(false);
    }
  });
};