define([
    'jquery',
    'url',
    'shake',
    'log',
    'Modernizr',
    'controls/distance',
    'controls/legal',
    'controls/bicyclelayer',
    'controls/maptype',
    'controls/routetype',
    'controls/showlink',
    'controls/geocode',
    'controls/dropdown',
    'controls/elevation',
    'controls/social'
  ],

  function (
    $,
    URL,
    shake,
    LOG,
    Modernizr,
    bindDistance,
    bindLegal,
    bindBicycleLayer,
    bindMapType,
    bindRouteType,
    bindShowLink,
    bindGeocode,
    bindDropdowns,
    bindElevation,
    setupSocialButtons) {

    "use strict";

    // imports
    var MapTypeId = google.maps.MapTypeId;
    var positions = google.maps.ControlPosition;
    var TravelMode = google.maps.TravelMode;
    var events = google.maps.event;

    // constants
    var DROPDOWN_SPEED = 100;
    var BIKE_LAYER = ".bikelayer-ctrl";
    var ROAD = ".road-ctrl";
    var SATELLITE = ".sat-ctrl";
    var TERRAIN = ".terrain-ctrl";
    var WALK = ".walk";
    var BIKE = ".bike";
    var DRIVE = ".drive";
    var DIRECT = ".direct";
    var DISTANCE = ".distance";
    var REVERSE = ".reverse-ctrl";
    var UNDO = ".undo-ctrl";
    var CLEAR = '.clear-ctrl';
    var FINDME = '.findme';
    var WARNING = '.warning';
    var SHOW_LINK = "#show-link";
    var EMAIL = "#emailLink";
    var EMBED = "#embedLink";
    var TINYURL = "#short-url";
    var TINYURL_LABEL = "#short-url-label";
    var TOGGLE_MODAL = ".toggle-modal";
    var FIT = ".fit";
    var GEOCODE = ".geocode";
    var BUTTONS_NEEDING_ROUTE = ".needs-route.btn";
    var SHOW_ELEVATION = ".show-elevation";
    var FEEDBACK_IFRAME = "iframe[name=feedback]";
    var FEEDBACK_FORM_URL = 'https://docs.google.com/forms/d/1_vbxe2YK2ECW5WUOzculeGRQTMzQneXf_yOI14WZ8ZI/viewform?embedded=true';

    var haveParams = !!URL.params.context;

    // trigger click event on the "this" argument
    var click = function () {
      $(this).trigger('click');
      return false;
    };

    // remove temporary control container from DOM and add each control
    // container to the control position specified in the "data-pos" attribute
    // of the div
    var renderControls = function (gmap, $controls) {
      var controls = gmap.controls;
      $controls.children().each(function () {
        var $this = $(this);
        var destText = $this.data("pos");
        // display if no query string or if this control has special class
        // indicating it should show up in this special context
        if (!haveParams && !$this.hasClass("not-context-edit") ||
          haveParams && $this.hasClass("context-" + URL.params.context)) {
          if (destText === "OFF_CANVAS") {
            $("#outer-wrap").append($this);
          } else {
            var destEnum = positions[destText];
            var destArray = controls[destEnum];
            if (!destEnum) {
              throw new Error("Bad data-pos: " + destText);
            }
            $this.children().each(function () {
              var idx = $(this).data("idx");
              this.index = idx;
              destArray.push(this);
            });
          }
        }
      });
    };

    // add some touch/mouse handlers so that the elements behave better on touch screens
    var setupModalForTouch = function () {
      var $body = $('body');
      // hide modal dialog on any touch on the backdrop
      $(document).on('touchstart mousedown', '.modal-backdrop', click);
      // block any right click/select/drag events from elements on the map
      $body.on('selectstart contextmenu', function (e) {
        return $(e.target).is('.selectable, .selectable *');
      });
      // iOS scroll hacks for modals
      $(document).on('touchmove', function(e){
        e.preventDefault();
      });
      $body.on('touchstart', '.scrollable', function(e) {
        var tgt = e.currentTarget;
        if (tgt.scrollTop === 0) {
          tgt.scrollTop = 1;
        } else if (tgt.scrollHeight === tgt.scrollTop + tgt.offsetHeight) {
          tgt.scrollTop -= 1;
        }
      });
      $body.on('touchmove', '.scrollable', function(e) {
        var tgt = e.currentTarget;
        e.stopPropagation();
        if (tgt.scrollHeight === tgt.offsetHeight) {
          e.preventDefault();
        }
      });
      // TODO iframe scroll still elastic-scrolls the parent
    };

    // helper to find the max z-index of any element on the page
    var maxZindex = function (selector) {
      var max = 10000003;
      selector = selector || "#map_canvas *";
      $(selector).each(function() {
        var current = parseInt($(this).css("zIndex"), 10);
        if (current > max) {
          max = current;
        }
      });
      return max;
    };

    // set up proper z index values for controls
    var fixZIndexes = function () {
      // make warning text clickable
      var onTop = {
        zIndex: maxZindex() + 1
      };
      $(".warning").css(onTop);
      // make dropdown menu appear on top of buttons
      $(".dropdown-menu").css(onTop).parent().css(onTop);
    };

    // when the "warning" button clicked and the about dialog is shown, make it
    // scroll to the bottom so that the warnings are in view
    var makeWarningButtonScrollToBottom = function () {
      var $modalBody = $(".modal-body");
      var $modal = $("#aboutModal");
      $(WARNING).on('click', function () {
        $modal.one('shown', function () {
          $modalBody.scrollTop($modalBody[0].scrollHeight);
        });
      });
    };

    var bindRouteDependent = function (route, $controls) {
      var $route = $(route);
      var $needRoute = $controls.find(BUTTONS_NEEDING_ROUTE);
      var update = function () {
        if (route.lastPoint) {
          $needRoute.removeClass("disabled");
        } else {
          $needRoute.addClass("disabled");
        }
      };
      $route.on('change', update);
      update();
    };

    var openFeedbackIframeOnDropdwonShow = function () {
      $("#feedbackModal").on('show', function () {
        $(FEEDBACK_IFRAME).attr('src', FEEDBACK_FORM_URL);
      });
    };


    // add controls to the map and setup bindings
    var build = function (map, model, route, html) {
      var $controls = $(html);
      var $all = $controls.add($(document.body));
      var $$ = function (arg) { return $all.find(arg); };
      setupModalForTouch();
      renderControls(map, $controls);

      bindDistance(model, $$(DISTANCE));
      bindLegal(model, $$(WARNING), DROPDOWN_SPEED);
      bindBicycleLayer(map, model, $$(BIKE_LAYER));
      bindRouteDependent(route, $controls);
      bindElevation(model, map, $$(SHOW_ELEVATION));

      $$(FINDME).fastClick(function () {
        events.trigger(map, 'togglefollow');
      });

      $$(FIT).fastClick(function () {
        var bounds = route.getBounds();
        if (bounds) {
          map.fitBounds(bounds);
        }
      });

      if (URL.params.context === "embed") {
        var shareUrl = window.location.href.replace("context=embed", "context=share");
        $$(".embed-to-share a").attr("href", shareUrl);
      }

      if (!haveParams) {
        bindDropdowns($$);

        bindMapType(map, $$(ROAD), MapTypeId.ROADMAP);
        bindMapType(map, $$(SATELLITE), MapTypeId.HYBRID);
        bindMapType(map, $$(TERRAIN), MapTypeId.TERRAIN);

        bindRouteType(model, $$(BIKE), TravelMode.BICYCLING);
        bindRouteType(model, $$(DRIVE), TravelMode.DRIVING);
        bindRouteType(model, $$(WALK), TravelMode.WALKING);
        bindRouteType(model, $$(DIRECT), null);

        $$(UNDO).fastClick(function () {
          route.undoLast();
          return false;
        });

        $$(CLEAR).fastClick(function () {
          route.clear();
        });

        $$(TOGGLE_MODAL).fastClick(function () {
          var targetSelector = $(this).data('target');
          var $target = $(targetSelector);
          $target.modal('show');
        });

        $(window).on("shake", function() {
          if (route.lastPoint) {
            var clear = window.confirm("Clear route?");
            if (clear) { route.clear(); }
            LOG('Shake', clear ? "yes" : "no");
          }
        });

        $$(REVERSE).fastClick(function () {
          route.reverse();
        });

        bindShowLink(map, model, route, {
          SHOW_LINK: $$(SHOW_LINK),
          EMBED: $$(EMBED),
          EMAIL: $$(EMAIL),
          TINYURL_LABEL: $$(TINYURL_LABEL),
          TINYURL: $$(TINYURL)
        });

        bindGeocode(map, $$(GEOCODE));
      }

      return $controls;
    };

    // updates to perform after the map is finished rendering and in its first
    // "idle" state
    var postRenderUpdate = function (route, map) {
      fixZIndexes();
      makeWarningButtonScrollToBottom();
      $(route).trigger("change");
      openFeedbackIframeOnDropdwonShow();
      setupSocialButtons(map);
    };

    // API
    return {
      build: build,
      postRenderUpdate: postRenderUpdate
    };
  }
);