define([
    'jquery',
    'log'
  ],

  function (
    $,
    LOG) {

    "use strict";
    // imports
    var dirs = new google.maps.DirectionsService();
    var Status = google.maps.DirectionsStatus;
    var OK = Status.OK;
    var THROTTLED = Status.OVER_QUERY_LIMIT;
    var REQUEST_DENIED = Status.REQUEST_DENIED;
    var UNKNOWN_ERROR = Status.UNKNOWN_ERROR;
    var ZERO_RESULTS = Status.ZERO_RESULTS;
    var PATH_ERROR = "error";

    // allow burst of 5 requests in the first second
    var BURST_LENGTH = 1000;
    var BURST_REQESTS = 5;

    // after that, limit to 1 request per second until no requests for 5 seconds
    var RATE_AFTER_BURST = 1000;
    var TIME_TO_CLEAR_LIMITING = 5000;

    // rate limiting global state
    var timestamps = [];
    var q = [];
    var throttlingEnabled = true;
    var overQueryLimitUntil;
    var inRateLimit = false;

    // extract the copyright and warning strings as well as all of the points
    // from a direction response from google, then pass to the callback
    var extractPoints = function (result, callback, handlea, handleb) {
      var route = result && result.routes && result.routes[0];
      var getLeg = function (idx) {
        return route && route.legs && route.legs[idx];
      };
      var getLegSteps = function (idx) {
        var leg = getLeg(idx);
        return (leg && leg.steps) || [];
      };
      var getLegDistance = function (idx) {
        var leg = getLeg(idx);
        return (leg && leg.distance && leg.distance.value) || undefined;
      };
      var getLegPath = function (legNum) {
        var path;
        var distance = getLegDistance(legNum);
        $.each(getLegSteps(legNum), function (idx, step) {
          var points = step.path || [];
          if (path) {
            path = path.concat(points);
          } else {
            path = points;
          }
        });

        return distance ? {
          distance: distance,
          points: path || []
        } : {
          points: path || []
        };
      };
      var first = getLegPath(0);
      var second = getLegPath(1);
      var third = getLeg(2) ? getLegPath(2) : PATH_ERROR;
      var fourth = getLeg(3) ? getLegPath(3) : PATH_ERROR;

      // finally, invoke the callback
      callback(
        (route && route.warnings) || [],
        (route && route.copyrights) || null,
        first,
        second,
        handlea,
        third || PATH_ERROR,
        fourth || PATH_ERROR,
        handleb
      );
    };

    // utility to get the current time in milliseconds
    var now = function () {
      return new Date().getTime();
    };

    // schedule a function to be invoked at a specified date in the future
    var retryAt = function (operation, time) {
      setTimeout(operation, time - now());
    };

    // checks the queue of requests to see if any are ready to be run now, and
    // if so then runs them
    var checkQueue = function () {
      var time = now();
      var firstTimestamp = timestamps[0];
      var lastTimestamp = timestamps[timestamps.length - 1];

      // clear rate limit if last request was long enough ago
      if (inRateLimit && lastTimestamp < (time - TIME_TO_CLEAR_LIMITING)) {
        inRateLimit = false;
      }

      if (overQueryLimitUntil && time < overQueryLimitUntil) {
        // less than 2 seconds since receiving an OVER_QUERY_LIMIT response,
        // wait until 2 seconds is up and try again
        retryAt(checkQueue, overQueryLimitUntil);
      } else if (inRateLimit && time < (lastTimestamp + RATE_AFTER_BURST)) {
        // the "burst limit" has been reached and requests throttled back to 1
        // per second and has been less than one second after last request -
        // reschedule this event for when the 1 second is up
        retryAt(checkQueue, lastTimestamp + RATE_AFTER_BURST);
      } else if (throttlingEnabled && timestamps.length === BURST_REQESTS && firstTimestamp > (time - BURST_LENGTH)) {
        // in the middle of a burst exceeding the burst limit, reschedule this
        // event for the end of the burst blackout
        inRateLimit = true;
        retryAt(checkQueue, firstTimestamp + BURST_LENGTH);
      } else {

        // execute the event!
        if (throttlingEnabled) {
          timestamps.push(time);
          if (timestamps.length > BURST_REQESTS) {
            timestamps.shift();
          }
        }

        // pop the first request from the queue and invoke it
        while (q.length > 0 && q.shift()()) {
          // keep emptying queue while responses are true - indicating request ignored
        }
      }
    };

    // encapsulate all of the throttling and retry logic for making requests to
    // the google directions service as well as handling errors in the response
    var doWork = function (options, backup, callback, err, handlea, handleb) {
      var lastTryHitRateLimit = false;
      var retryUnknown = false;
      var tryBackup = false;
      var canceled = function () { return handlea.canceled && (!handleb || handleb.canceled); };

      // handle making the request to the google directions service and
      // processing the response
      var makeRequest = function () {
        if (canceled()) {
          // don't send request
        } else {
          dirs.route(tryBackup ? backup : options, function (result, status) {
            LOG('Direction', options.travelMode, status);
            if (canceled()) {
              // ignore response
            } else if (status === OK) {
              // successfule response, strip out the data we care about
              extractPoints(result, callback, handlea, handleb);
            } else if (status === ZERO_RESULTS && !tryBackup) {
              // directions failed for round trip, retry one-way
              tryBackup = true;
              q.unshift(makeRequest);
              checkQueue();
            } else if (status === UNKNOWN_ERROR && !retryUnknown) {
              // may pass on second try
              retryUnknown = true;
              q.unshift(makeRequest);
              checkQueue();
            } else if (status === THROTTLED && !lastTryHitRateLimit) {
              lastTryHitRateLimit = true;
              overQueryLimitUntil = now() + 2000;
              // push this to head of queue
              q.unshift(makeRequest);
              retryAt(checkQueue, overQueryLimitUntil);
            } else {
              if (status === THROTTLED) {
                LOG.error("Direction quota");
              } else if (status === REQUEST_DENIED) {
                LOG.error("Direction request denied");
              }
              err();
            }
          });
        }
        return canceled();
      };

      q.push(makeRequest);

      checkQueue();
    };

    // request directions from A to B
    var getDirections = function (start, end, routetype, callback, err) {
      var handlea = {canceled: false, reversed: false};
      doWork({
        origin: start,
        destination: start,
        waypoints: [{
          location: end
        }],
        travelMode: routetype
      },
      // backup - just go forward
      {
        origin: start,
        destination: end,
        travelMode: routetype
      }, callback, err, handlea);
      return handlea;
    };

    // request directions from A to B to C, useful for when a midpoint between
    // to follow-road segments is dragged
    var getDirectionsPair = function (start, mid, end, routetype, callback, err) {
      var handlea = {canceled: false, reversed: false};
      var handleb = {canceled: false, reversed: false};
      doWork({
        origin: start,
        destination: start,
        waypoints: [{
          location: mid
        }, {
          location: end
        }, {
          location: mid
        }],
        travelMode: routetype
      },
      // backup - just go forward
      {
        origin: start,
        destination: end,
        waypoints: [{
          location: mid
        }],
        travelMode: routetype
      }, callback, err, handlea, handleb);
      return [handlea, handleb];
    };

    // used for tests to disable/enable throtting mechanism
    var setThrottlingEnabled = function (enable) {
      throttlingEnabled = enable;
    };

    // used for tests to reset the state of the Direction module
    var reset = function () {
      timestamps = [];
      overQueryLimitUntil = null;
      inRateLimit = false;
    };

    // API
    return {
      get: getDirections,
      getPair: getDirectionsPair,
      throttle: setThrottlingEnabled,
      reset: reset,
      PATH_ERROR: PATH_ERROR
    };
  }
);