define([
    'jquery'
  ],

  function ($) {

    "use strict";

    var FT_PER_M = 3.28084;
    var M_PER_KM = 1000;
    var M_PER_MI = 1609.344;
    var CIRCLE = google.maps.SymbolPath.CIRCLE;
    var Marker = google.maps.Marker;
    var LatLng = google.maps.LatLng;
    var computeDistanceBetween = google.maps.geometry.spherical.computeDistanceBetween;
    var interpolatePoints = google.maps.geometry.spherical.interpolate;

    // marker style
    var MARKER = {
      draggable: false,
      clickable: false,
      zIndex: 1000,
      icon: {
        path: CIRCLE,
        scale: 5,
        strokeWeight: 1,
        strokeOpacity: 0.5,
        fillColor: "#ff7f0e",
        fillOpacity: 1
      }
    };

    var attachTo = function (selector, model, map) {
      /* global d3 */
      var margin = {top: 10, right: 20, bottom: 20, left: 40},
          statSpacing = 10,
          percent = d3.format("%"),
          distFmt = d3.format(".1f"),
          eleFmt = d3.format(",.0f"),
          data,
          indexedPath,
          lastLength,
          overallGrade = 0,
          x,
          verticalGain = {mi: 0, km: 0},
          verticalLoss = {mi: 0, km: 0},
          $container = $(selector),
          bisect = d3.bisector(function (d) { return d['distance' + model.getUnit()]; }).left,
          marker = new Marker(MARKER),
          last;

      var stats = function (container) {
        var unit = model.getUnit();
        var elevationUnit = unit === 'mi' ? 'ft' : 'm';
        var distanceText = model.getDistanceText().split(" ");

        container.append("tspan")
            .attr("font-weight", "bold")
            .text(distanceText[0] + " " + distanceText[1]);

        container.append("tspan")
            .attr("dx", statSpacing)
            .attr("font-weight", "bold")
            .text(eleFmt(verticalGain[unit]) + " " + elevationUnit);
        container.append("tspan")
            .attr("fill", "#555")
            .text(" gain");

        container.append("tspan")
            .attr("dx", statSpacing)
            .attr("font-weight", "bold")
            .text(eleFmt(verticalLoss[unit]) + " " + elevationUnit);
        container.append("tspan")
            .attr("fill", "#555")
            .text(" loss");

        container.append("tspan")
            .attr("dx", statSpacing)
            .attr("font-weight", "bold")
            .text(percent(overallGrade));
        container.append("tspan")
            .attr("fill", "#555")
            .text(" grade");
      };

      var left = function (text) {
        text
          .attr("text-anchor", "end")
          .attr("dx", -5);
      };

      var right = function (text) {
        text
          .attr("text-anchor", "start")
          .attr("dx", 5);
      };

      var update = function () {
        // re-render the chart
        var origWidth = $container.width(),
            origHeight = $container.height(),
            width = origWidth - margin.left - margin.right,
            height = origHeight - margin.top - margin.bottom,
            elevationTop = 10,
            elevationBottom = height,
            elevationKey = 'elevation' + model.getUnit(),
            distanceKey = 'distance' + model.getUnit();
        marker.setMap(null);

        var preInterpolate = function (array, dist) {
          function single(idx) {
            return {
              before: array[idx],
              after: array[idx],
              ratio: 0
            };
          }
          var after = bisect(array, dist);
          if (after <= 0) { return single(0); }
          if (after >= array.length) { return single(array.length - 1); }
          var before = after - 1;
          var dataBefore = array[before];
          var dataAfter = array[after];
          var ratio = (dist - dataBefore[distanceKey]) / (dataAfter[distanceKey] - dataBefore[distanceKey]);
          return {
            before: dataBefore,
            after: dataAfter,
            ratio: ratio
          };
        };

        var interpolateByDistance = function (dist) {
          var datum = preInterpolate(data, dist);
          var interpolated = d3.interpolateObject(datum.before, datum.after)(datum.ratio);
          // special case since grade is 1st derivative
          interpolated.grade = datum.after.grade;
          // special case for location, need to sample along the actual path
          var location = preInterpolate(indexedPath, dist);
          var mixed = interpolatePoints(location.before.point, location.after.point, location.ratio);
          interpolated.lat = mixed.lat();
          interpolated.lng = mixed.lng();
          return interpolated;
        };

        var show = function () {
          line.style('display', null);
          elevationText.style('display', null);
          gradeText.style('display', null);
          distanceText.style('display', null);
          if (!marker.getMap()) {
            marker.setMap(map);
          }
        };

        var displayInterpolated = function (interpolated, ycoord) {
          if (interpolated === null) {
            hide();
          } else {
            var coord = x(interpolated[distanceKey]);
            if (coord <= 0 || Math.round(coord) >= width || (typeof ycoord === 'number' && ycoord < 0)) {
              hide();
            } else {
              show();
              var lineHeight = +elevationText.style("font-size").replace("px", "");
              line
                .attr("x1", coord)
                .attr("x2", coord)
                .attr("y1", -10)
                .attr("y2", 2*height);
              elevationText
                .attr("dy", -lineHeight)
                .text(Math.round(interpolated[elevationKey]) + " " + (model.getUnit() === 'mi' ? 'ft' : 'm'));
              gradeText
                .text(percent(interpolated.grade));
              distanceText
                .attr("dy", -2 * lineHeight)
                .text(distFmt(interpolated[distanceKey]) + " " + model.getUnit());

              var y = yElevation(interpolated[elevationKey]);
              var padding = 2.25 * lineHeight;

              d3.selectAll(".moving-stat")
                .attr("x", coord)
                .attr("y", y < (10 + padding) ? (10 + padding) : (y > height - 5 ? height - 5 : y))
                .call(coord < width - 40 ? right : left);
              var pos = new LatLng(interpolated.lat, interpolated.lng);
              marker.setPosition(pos);
            }
          }
        };

        var mouseOverChartAtDist = function (dist) {
          if (data) {
            dist = model.getUnit() === 'km' ? dist / M_PER_KM : dist / M_PER_MI;
            var interpolated = interpolateByDistance(dist);
            displayInterpolated(interpolated);
          }
        };

        var mouseOverChartAtX = function (coord, y) {
          coord = Math.round(coord - margin.left);
          var dist = x.invert(coord);
          if (data) {
            var interpolated = interpolateByDistance(dist);
            displayInterpolated(interpolated, y);
          } else {
            elevationText.text("");
            gradeText.text("");
            distanceText.text("");
            if (coord < 0 || coord > width || y < 0) {
              hide();
            } else {
              show();
              line
                .attr("x1", coord)
                .attr("x2", coord)
                .attr("y1", -10)
                .attr("y2", 2 * height);
            }
            marker.setMap(null);
          }
        };

        var handleTouch = function () {
          var touches = d3.touches(this);
          var coord = d3.touches(this)[0][0];
          mouseOverChartAtX(coord, touches[0][1]);
          // prevent mouseover handler
          d3.event.preventDefault();
        };

        var handleMouseOver = function () {
          var coord = d3.mouse(this)[0];
          mouseOverChartAtX(coord);
        };

        var oldSvg = d3.select(selector).selectAll("svg");
        var outer = d3.select(selector)
          .append("svg")
            .attr("width", "100%")
            .attr("height", "100%");

        // TODO handle devices with mouse and touch screen?
        outer
          .on('touchmove', handleTouch)
          .on('touchstart', handleTouch)
          .on('touchcancel', hide)
          .on('mousemove', handleMouseOver)
          .on('mouseover', handleMouseOver)
          .on('mouseout', hide);

        outer.append("g")
            .attr("transform", "translate(" + margin.left + ",0)")
            .attr("class", "stats")
          .append("text")
            .attr("xml:space", "preserve")
            .attr("x", width / 2)
            .attr("y", 12)
            .attr("text-anchor", "middle")
            .call(stats);

        var svg = outer.append("g")
            .attr("width", width)
            .attr("height", height)
            .attr("transform", "translate(" + margin.left + "," + margin.top + ")");

        x = d3.scale.linear()
          .range([0, width]);

        var yElevation = d3.scale.linear()
          .range([elevationBottom, elevationTop]);

        var xAxis = d3.svg.axis()
          .tickSize(1, 0.1, 0)
          .scale(x)
          .orient("bottom");

        var elevationAxis = d3.svg.axis()
          .scale(yElevation)
          .ticks(Math.round(height / 40))
          .orient("left");

        x.domain(!data ? [0, 1] : d3.extent(data, function(d) { return d[distanceKey]; }));
        yElevation.domain(!data ? [0, 1] : d3.extent(data, function(d) { return d[elevationKey]; }));

        var elevationLine = d3.svg.area()
          .interpolate('monotone')
          .x(function(d) { return x(d[distanceKey]); })
          .y0(height)
          .y1(function(d) { return yElevation(d[elevationKey]); });

        svg.append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0," + height + ")")
            .call(xAxis);

        svg.append("g")
            .attr("class", "y axis")
            .call(elevationAxis)
          .append("text")
            .attr("y", 2)
            .style("text-anchor", "end")
            .text("Ele (" + (model.getUnit() === 'mi' ? 'ft' : 'm') + ")");

        svg.append("path")
            .datum(data)
            .attr("class", "area elevation")
            .attr("d", data ? elevationLine : "M 0 0");

        var line = svg.append("line")
          .attr("class", "brush")
          .style("display", "none");

        var elevationText = svg.append("text")
          .attr("class", "moving-stat")
          .attr("y", 0)
          .attr("x", -10);

        var distanceText = svg.append("text")
          .attr("class", "moving-stat")
          .attr("y", 0)
          .attr("x", -10);

        var gradeText = svg.append("text")
          .attr("class", "moving-stat")
          .attr("y", 0)
          .attr("x", -10);

        oldSvg.remove();

        last = {
          mouseOverChartAtX: mouseOverChartAtX,
          hide: hide,
          show: show,
          interpolateByDistance: interpolateByDistance,
          mouseOverChartAtDist: mouseOverChartAtDist
        };

        return last;
      };


      var hide = function () {
        d3.selectAll('line.brush').style('display', 'none');
        d3.selectAll('.moving-stat').style('display', 'none');
        marker.setMap(null);
      };

      var indexPath = function (path) {
        var dist = 0;
        var last = null;
        return path.map(function (elem) {
          if (last) {
            dist += computeDistanceBetween(last, elem);
          }
          last = elem;
          return {
            distance: dist,
            distancemi: dist / M_PER_MI,
            distancekm: dist / M_PER_KM,
            point: elem
          };
        });
      };

      var onData = function (length, newData, newPath) {
        data = newData;
        indexedPath = indexPath(newPath);
        formatData(length, data);
        update();
        return last;
      };

      var onEmpty = function () {
        overallGrade = 0;
        data = null;
        verticalGain = {mi: 0, km: 0};
        verticalLoss = {mi: 0, km: 0};
        update();
      };

      var onLoad = onEmpty;

      var onResize = function () {
        var oldLine = d3.select("line.brush");
        var oldCoord = oldLine.style("display") !== 'none' && oldLine.attr("x1");
        if (oldCoord) {
          oldCoord = x.invert(+oldCoord);
        }
        update();
        if (oldCoord) {
          last.mouseOverChartAtX(x(oldCoord) + margin.left);
          marker.setMap(map);
        }
      };

      var onUnit = function () {
        var oldLine = d3.select("line.brush");
        var oldCoord = oldLine.style("display") !== 'none' && oldLine.attr("x1");
        update();
        if (oldCoord) {
          last.mouseOverChartAtX(+oldCoord + margin.left);
          marker.setMap(map);
        }
      };

      var onHide = function () {
        marker.setMap(null);
      };

      var mouseoverRoute = function (data) {
        var dist = data.distance;
        if (last) {
          last.mouseOverChartAtDist(dist);
        }
      };

      var disableMouseover = function () {
        data = null;
      };

      var mouseoutRoute = hide;

      var formatData = function (length, data) {
        var spacing = length / (data.length - 1);
        var dist = 0;
        var last = null;
        var gain = 0;
        var loss = 0;
        lastLength = length;
        data.forEach(function(d) {
          d.distance = dist;
          d.distancemi = dist / M_PER_MI;
          d.distancekm = dist / M_PER_KM;
          if (last !== null) {
            var deltaElevation = d.elevation - last;
            if (deltaElevation > 0) { gain += deltaElevation; }
            else { loss -= deltaElevation; }
            d.grade = deltaElevation / spacing;
          } else {
            d.grade = 0;
          }
          last = d.elevation;
          d.elevation = d.elevation;
          d.elevationmi = d.elevation * FT_PER_M;
          d.elevationkm = d.elevation;
          d.lat = d.location.lat();
          d.lng = d.location.lng();
          dist += spacing;
        });
        verticalGain = {
          mi: gain * FT_PER_M,
          km: gain
        };
        verticalLoss = {
          mi: loss * FT_PER_M,
          km: loss
        };
        overallGrade = (data[data.length-1].elevation - data[0].elevation) / length;

        return {
          verticalGain: verticalGain,
          verticalLoss: verticalLoss,
          overallGrade: overallGrade
        };
      };

      return {
        data: onData,
        empty: onEmpty,
        hide: onHide,
        resize: onResize,
        loading: onLoad,
        unit: onUnit,
        formatData: formatData,
        mouseoverRoute: mouseoverRoute,
        mouseoutRoute: mouseoutRoute,
        disableMouseover: disableMouseover
      };
    };

    return {
      attachTo: attachTo
    };
  }
);