// Abstraction layer for google maps elevation service
define([
    'jquery',
    'rdp',
    'segment',
    'log',
    'store',
    'sample'
  ],

  function (
    $,
    RDP,
    Segment,
    LOG,
    Store,
    sample) {

    "use strict";

    var TIME_BETWEEN_REQUESTS = 1000;
    var CACHED_KEYS = "elevation.keys1";
    var ELEVATION_CACHE = "elevation.cache1";
    var elevator = new google.maps.ElevationService();
    var NUM_REQ_POINTS = 100;
    var pending;
    var requestInProgress = false;
    var scheduled = null;
    var cache = {};
    var CACHE_SIZE = 20;
    var cachedKeys = [];
    var MAX_MAPQUEST_LENGTH = 150 * 1609;
    var MAPQUEST_OK = "MAPQUEST_OK";

    var serviceScheduledFirstRequest = function () {
      scheduled();
      scheduled = null;
    };
    var serviceNextRequest = function () {
      if (pending) {
        pending();
        pending = null;
      } else {
        requestInProgress = false;
      }
    };
    var reviver = function (key, val) {
      if (key === 'location') {
        val = new google.maps.LatLng(val.lat, val.lng);
      }
      return val;
    };
    var replacer = function (key, val) {
      if (key === 'location') {
        val = {
          lat: val.lat(),
          lng: val.lng()
        };
      }
      return val;
    };


    var getElevation = function (path, succ, error, points, length) {
      points = points || NUM_REQ_POINTS;

      var sampled = sample({
        path: path,
        numPoints: points
      });
      var key = JSON.stringify(sampled);

      var handleResponse = function (results, status, extraInfo) {
        if (status === google.maps.ElevationStatus.OK || status === MAPQUEST_OK) {
          succ(results, extraInfo);
          cache[key] = {
            data: results,
            copyright: extraInfo
          };
          cachedKeys.push(key);
          if (cachedKeys.length > CACHE_SIZE) {
            delete cache[cachedKeys.shift()];
          }
          Store.force.set(ELEVATION_CACHE, JSON.stringify(cache, replacer));
          Store.force.set(CACHED_KEYS, JSON.stringify(cachedKeys));
        } else {
          error(status);
          LOG.error('Elevation Service Error', status);
        }
        LOG('Elevation', status === MAPQUEST_OK ? 'mapquest' : 'google', status, points);
        setTimeout(serviceNextRequest, TIME_BETWEEN_REQUESTS);
      };

      var makeGoogleRequest = function () {
        elevator.getElevationForLocations({
          locations: sampled
        }, handleResponse);
      };

      var isMapquestResponseInvalid = function (response) {
        return !response ||
          !response.info ||
          response.info.statuscode !== 0 ||
          !response.elevationProfile ||
          (response.info && response.info.messages && response.info.messages.length > 0) ||
          response.elevationProfile.length !== sampled.length;
      };

      var makeMapquestRequest = function () {
        var done = false;
        var points = sampled.map(function (point) {
          return point.toUrlValue();
        }).join(",");
        // first try mapquest, if that fails use google maps elevation service
        $.getJSON("http://open.mapquestapi.com/elevation/v1/profile?key=Fmjtd%7Cluub256t2d%2Cbw%3Do5-9u8n5w&shapeFormat=raw&latLngCollection=" + points, function (d) {
          if (done) {
            // do nothing
          } else if (isMapquestResponseInvalid(d)) {
            // error with mapquest, use google maps
            makeGoogleRequest();
          } else {
            // mapquest OK, use its data
            var results = d.elevationProfile.map(function (d) { return {elevation: d.height}; });
            results.forEach(function (r, i) {
              r.location = sampled[i];
            });
            handleResponse(results, MAPQUEST_OK, d.info.copyright);
          }
          done = true;
        }).fail(function () {
          done = true;
          makeGoogleRequest();
        });
      };

      var doMakeRequest = function () {
        requestInProgress = true;
        var lowerLat = 90;
        var upperLat = -90;
        sampled.forEach(function (s) {
          var lat = s.lat();
          if (lat < lowerLat) { lowerLat = lat; }
          if (lat > upperLat) { upperLat = lat; }
        });
        if (length < MAX_MAPQUEST_LENGTH && lowerLat > -56 && upperLat < 60) {
          makeMapquestRequest();
        } else {
          makeGoogleRequest();
        }
      };

      if (cache[key]) {
        succ(cache[key].data, cache[key].copyright);
      } else if (requestInProgress) {
        // service this request 1s after response from outstanding
        pending = doMakeRequest;
      } else if (scheduled) {
        // have already scheduled an immediate request during this event loop, do this one instead
        scheduled = doMakeRequest;
      } else {
        // schedule this for execution on next event loop cycle
        scheduled = doMakeRequest;
        setTimeout(serviceScheduledFirstRequest, 0);
      }
    };

    var reset = function () {
      cache = {};
      cachedKeys = [];
      pending = null;
      requestInProgress = false;
      scheduled = null;
    };

    var reload = function () {
      try {
        cache = JSON.parse(Store.force.get(ELEVATION_CACHE), reviver) || {};
        cachedKeys = JSON.parse(Store.force.get(CACHED_KEYS)) || [];
      } catch (e) {
        reset();
      }
    };

    reload();

    return {
      getElevation: getElevation,
      reset: reset,
      reload: reload
    };
  }
);