// handle displaying marker that follows device location on the map
define([
    'jquery',
    'gps',
    'store',
    'Modernizr'
  ],

  function (
    $,
    GPS,
    Store,
    Modernizr) {

    "use strict";

    // imports
    var maps = google.maps;
    var Marker = maps.Marker;
    var Circle = maps.Circle;
    var LatLng = maps.LatLng;
    var events = maps.event;
    var CIRCLE = maps.SymbolPath.CIRCLE;

    // style for position markers
    var LOCATION_MARKER = {
      clickable: false,
      dragable: false,
      title: "your current position",
      icon: {
        path: CIRCLE,
        scale: 4,
        strokeWeight: 0,
        fillColor: "blue",
        fillOpacity: 1
      }
    };
    var CIRCLE_OPTIONS = {
      clickable: false,
      dragable: false,
      fillColor: "blue",
      fillOpacity: 0.1,
      strokeColor: 'blue',
      strokeOpacity: 0.2,
      strokeWeight: 1
    };

    var WAS_FOLLOWING = "show_pos";
    var RECENTER_ON_LOAD_THRESHOLD = 2000;

    var attachTo = function (map, haveCenter) {
      var iosapp = Modernizr.ioswebapp;
      var dot = new Marker(LOCATION_MARKER);
      var circle = new Circle(CIRCLE_OPTIONS);
      var updateByUs = false;

      // update the center and zoom of a map based on location and accuracy
      var updateCenter = function (lat, lng, accuracy, first) {
        var zoom = 12;
        if ($.isNumeric(lat) && $.isNumeric(lng)) {
          updateByUs = true;
          try {
            var pos = new LatLng(lat, lng);
            if (first) {
              if (accuracy <= 125) {
                zoom = 16;
              } else if (accuracy <= 250) {
                zoom = 15;
              } else if (accuracy <= 500) {
                zoom = 14;
              } else if (accuracy <= 1000) {
                zoom = 13;
              }
              // hard-zoom on load
              map.setCenter(pos);
              map.setZoom(zoom);
            } else {
              // pan on updates
              map.panTo(pos);
            }
          } finally {
            updateByUs = false;
          }
        }
      };

      // update the position marker based on location and accuracy (hide if > 2000m)
      var updatePosition = function (lat, lng, accuracy) {
        if (accuracy > 2000) {
          clearPosition();
        } else if ($.isNumeric(lat) && $.isNumeric(lng)) {
          Store.set(WAS_FOLLOWING, true);
          var pos = new LatLng(lat, lng);
          dot.setPosition(pos);
          if (!dot.getMap()) {
            dot.setMap(map);
          }
          if ($.isNumeric(accuracy)) {
            circle.setCenter(pos);
            circle.setRadius(accuracy);
            if (!circle.getMap()) {
              circle.setMap(map);
            }
          }
        }
      };

      // clear the position marker from the map
      var clearPosition = function () {
        dot.setMap(null);
        circle.setMap(null);
        Store.set(WAS_FOLLOWING, false);
      };

      // location watch state
      var centeredAtLocation = function () {
        return dot.getMap() && map.getBounds().contains(dot.getPosition());
      };
      var watchMarker = null;
      var token = 0;

      // handle a "togglefollow" map event:
      // - if no position displayed, display it and pan to it
      // - if position displayed but not centered, pan to it
      // - if position displayed and centered, stop showing position
      var toggleFollow = function (first, ignoreRecenter) {
        if (watchMarker === null && !centeredAtLocation()) {
          // no marker and not following, add watch for both
          token += 1;
          var myToken = token;
          GPS.findMe(function (lat, lng, acc) {
            if (watchMarker === null && !centeredAtLocation() && myToken === token && (!first || acc < RECENTER_ON_LOAD_THRESHOLD || !haveCenter)) {
              updatePosition(lat, lng, acc);
              if (!ignoreRecenter) {
                updateCenter(lat, lng, acc, true);
              }
              watchMarker = GPS.followMe(updatePosition);
            }
          });
        } else if (!centeredAtLocation()) {
          // displaying a marker, but not centering on current position - so start centering!
          var last = GPS.getCached();
          if (last) {
            var coords = last.coords;
            if (!ignoreRecenter) {
              updateCenter(coords.latitude, coords.longitude, coords.accuracy, true);
            }
          }
        } else {
          // not showing position
          GPS.cancel(watchMarker);
          clearPosition();
          watchMarker = null;
        }
      };

      // and listen for future ones
      events.addListener(map, 'togglefollow', toggleFollow);
      // ios app remembers if you were displaying position or not
      if (Store.getBoolean(WAS_FOLLOWING) && iosapp) {
        toggleFollow(true, true);
      }

      return {
        toggle: toggleFollow
      };
    };

    // API
    return {
      attachTo: attachTo,
      LOCATION_MARKER: LOCATION_MARKER
    };
  }
);