define([
    'store',
    'url'
  ],

  function (
    Store,
    URL) {

    "use strict";

    // imports
    var geolocation = function () {
      return (window.OTGM && window.OTGM.geolocation) || navigator.geolocation;
    };

    // geolocation request parameters
    var OPTIONS = {
      enableHighAccuracy: true,
      maximumAge: 60000,
      timeout: 60000
    };
    // localStorage key for remembering when a user has denied a geolocation
    // request, so we don't automatically ask them for it again until they have
    // initiated one and have OK'd it
    var DONT_DISTURB = "donotdisturb";
    // cache the last geolocation response
    var last = null;

    // wrap a success callback to capture the last location request and reset
    // the do not disturb flag
    var createSuccessNotifier = function (callback) {
      return function (pos) {
        var latlon = pos.coords;
        Store.set(DONT_DISTURB, false);
        last = pos;
        callback(latlon.latitude, latlon.longitude, latlon.accuracy);
      };
    };

    // wrap an error callback to capture a geolocation error response and set
    // the do not disturb flag if error code was "PERMISSION_DENIED"
    var createErrorNotifier = function (callback) {
      return function (err) {
        // catch permission denied by user, remember that so we don't ask them
        // again by default
        if (err && err.code === 1) {
          // permission denied
          Store.set(DONT_DISTURB, true);
        } else {
          Store.set(DONT_DISTURB, false);
        }
        if (callback) {
          callback();
        }
      };
    };

    // test if geolocation is supported on this browser
    var isGeolocationSupported = function () {
      return geolocation() ? true : false;
    };

    // make a request to get the current position
    var getPosition = function (succ, err) {
      if (isGeolocationSupported()) {
        if (succ) {
          geolocation().getCurrentPosition(
            createSuccessNotifier(succ),
            createErrorNotifier(err),
            OPTIONS
          );
        }
      } else if (err) {
        err();
      }
    };

    // make a request to follow the current position and receive updates any
    // time it changes
    var followPosition = function (succ, err) {
      var resultId;
      if (isGeolocationSupported()) {
        if (succ) {
          var successNotifier = createSuccessNotifier(succ);
          if (last) {
            successNotifier(last);
          }
          resultId = geolocation().watchPosition(
            successNotifier,
            createErrorNotifier(err),
            OPTIONS
          );
        }
      } else if (err) {
        err();
      }
      return resultId;
    };

    // cancel a follow position watch
    var cancel = function (id) {
      geolocation().clearWatch(id);
    };

    // returns true if the do not disturb flag is set so geolocation requests
    // aren't made by default
    var dontDisturb = function () {
      return 'true' === Store.get(DONT_DISTURB) || URL.params.nogeo;
    };

    // return the last cached location
    var getCached = function () {
      return last;
    };

    // API
    return {
      enabled: isGeolocationSupported,
      findMe: getPosition,
      followMe: followPosition,
      cancel: cancel,
      dontDisturb: dontDisturb,
      getCached: getCached
    };
  }
);