define([
    'jquery',
    'util',
    'store',
    'url'
  ],

  function (
    $,
    util,
    Store,
    URL) {

    "use strict";

    // imports
    var TravelMode = google.maps.TravelMode;

    // treat storing a "null" value into localStorage differently than just not
    // having anything stored
    var NULL = "persisted-null";
    var TRAVEL_MODE_INIT = TravelMode.WALKING;
    var VALID_TRAVEL_MODES = [
      TravelMode.WALKING,
      TravelMode.BICYCLING,
      TravelMode.DRIVING,
      NULL
    ];

    // create a unit from name and meters in that unit
    var Unit = function (name, meters) {
      return {
        name: name,
        meters: meters
      };
    };

    // the allowed units
    var UNITS = {
      km: new Unit("km", 1000),
      mi: new Unit("mi", 1609.34)
    };

    // the model class
    var Model = function () {
      var that = this;
      var $this = $(this);

      // utility to add a property with getter and a setter that triggers event
      // changed properties
      var addProperty = function (name, init) {
        var value = init;
        var firstUpper = name.charAt(0).toUpperCase() + name.slice(1);
        that['get' + firstUpper] = function () {
          return value;
        };
        that['set' + firstUpper] = function (newValue) {
          value = newValue;
          $this.trigger(name, newValue);
        };
      };

      // utility that also adds a property, but makes it backed by local storage
      // so it comes up with the previous value on reload
      var addPersistedProperty = function (name, alias, validValues, init) {
        var lastValue = Store.get(alias) || Store.get(name);
        if ($.isPlainObject(validValues)) {
          init = validValues.hasOwnProperty(lastValue) ? validValues[lastValue] : init;
        } else if (validValues.indexOf(lastValue) !== -1) {
          init = lastValue === NULL ? null : lastValue;
        }
        addProperty(name, init);

        $this.on(name, function (e, value) {
          Store.set(name, (typeof value === 'undefined' || value === null) ? NULL : value);
        });
      };

      // The properties
      addPersistedProperty('travelMode', 't', VALID_TRAVEL_MODES, TRAVEL_MODE_INIT);
      addPersistedProperty('unit', 'u', ["mi", "km"], "mi");
      addPersistedProperty('bikeLayer', 'b', {"true": true, "false": false}, false);
      addProperty('meters', 0);
      addProperty('distanceText');
      addProperty('legal', { warnings: [], copyrights: [] });
      addProperty('elevation', false);

      // utility to toggle between miles and kilometers
      this.toggleUnit = function () {
        var unit = that.getUnit();
        that.setUnit(unit === "km" ? "mi" : "km");
      };

      // derive the "distanceText" property from unit and raw distance properties
      var updateDistance = function () {
        var unit = UNITS[that.getUnit()] || UNITS.mi;
        var amount = that.getMeters() / unit.meters;
        var rounded = util.round(amount, 1);
        that.setDistanceText(rounded + " " + unit.name);
      };
      $this.on('unit', updateDistance);
      $this.on('meters', updateDistance);
      updateDistance();
    };

    // API
    return function () {
      var model = new Model();
      var params = URL.params;
      if (params) {
        if (params.w || params.w) {
          model.setLegal({
            warnings: params.w ? params.w : [],
            copyrights: params.c ? params.c : []
          });
        }
        if (params.e) {
          model.setElevation(true);
        }
        if (params.d) {
          model.setMeters(parseInt(params.d, 10));
        }
      }

      return model;
    };
  }
);