define([
    'jquery',
    'directions',
    'gmapsingleclick'
  ],

  function (
    $,
    Directions,
    setupSingleClick) {

    "use strict";

    // imports
    var events = google.maps.event;
    var CIRCLE = google.maps.SymbolPath.CIRCLE;
    var Marker = google.maps.Marker;
    var encode = google.maps.geometry.encoding.encodePath;
    var decode = google.maps.geometry.encoding.decodePath;

    // marker styles
    var MID_MARKER = {
      draggable: true,
      title: "waypoint",
      zIndex: 2,
      icon: {
        path: CIRCLE,
        scale: 5,
        strokeWeight: 1,
        strokeOpacity: 0.5,
        fillColor: "white",
        fillOpacity: 1
      }
    };
    var FIRST_MARKER = {
      draggable: true,
      title: "start",
      zIndex: 3,
      icon: {
        path: CIRCLE,
        scale: 5,
        strokeWeight: 1,
        strokeOpacity: 0.5,
        fillColor: "#2ca02c",
        fillOpacity: 1
      }
    };
    var LAST_MARKER = {
      draggable: true,
      zIndex: 4,
      title: "finish",
      icon: {
        path: CIRCLE,
        scale: 5,
        strokeWeight: 1,
        strokeOpacity: 0.5,
        fillColor: "#d62728",
        fillOpacity: 1
      }
    };
    var SPLITTER_MARKER = {
      draggable: true,
      zIndex: 1,
      title: "drag to split this segment",
      icon: {
        path: CIRCLE,
        scale: 5,
        strokeWeight: 1,
        strokeOpacity: 0.25,
        fillColor: "white",
        fillOpacity: 0.25
      }
    };

    // Set up values that differ for each point instance
    var Point = function (map, latLng, route, options) {
      // this is true if the point is getting deserialized from a string
      if (typeof latLng === 'string') {
        latLng = decode(latLng)[0];
      }
      var that = this;
      this.route = route;
      this.marker = new Marker(options || MID_MARKER);
      this.marker.setMap(map);
      this.marker.setPosition(latLng);

      // setup event handlers
      events.addListener(this.marker, 'dragstart', $.proxy(this, 'updateBothSides', 'startUpdate'));
      events.addListener(this.marker, 'drag', $.proxy(this, 'updateBothSides', 'update'));
      events.addListener(this.marker, 'dragend', $.proxy(this, 'updateBothSides', 'finishUpdate'));
      events.addListener(this.marker, 'singleclick', function () {
        if (route.addPoint(that.marker.getPosition())) {
          route.update();
        }
      });
      setupSingleClick(this.marker);
    };

    // export marker styles from the module
    Point.FIRST_MARKER = FIRST_MARKER;
    Point.LAST_MARKER = LAST_MARKER;
    Point.MID_MARKER = MID_MARKER;
    Point.SPLITTER_MARKER = SPLITTER_MARKER;

    // make a request to the google direction service for a path from A to B to
    // C for two adjacent segments and their midpoint
    var makeOptimizedDirectionRequest = function (entering, mid, leaving) {
      var success = function (warnings, copyrights, first, last, handlea, lastrev, firstrev, handleb) {
        // only recalculate distance on second update
        entering.handleDirectionResponse(warnings, copyrights, first, firstrev, handlea, true);
        leaving.handleDirectionResponse(warnings, copyrights, last, lastrev, handleb);
      };
      var error = function () {
        entering.handleDirectionError();
        leaving.handleDirectionError();
      };
      var handles = Directions.getPair(
        entering.start.marker.getPosition(),
        mid.marker.getPosition(),
        leaving.end.marker.getPosition(),
        entering.travelMode,
        success,
        error
      );
      if (entering.lastReq) { entering.lastReq.canceled = true; }
      if (leaving.lastReq) { leaving.lastReq.canceled = true; }
      entering.lastReq = handles[0];
      leaving.lastReq = handles[1];
    };

    // Methods shared between all point instances
    Point.prototype = {
      // when midpoint moved, adjust segments on both sides
      updateBothSides: function (name) {
        var entering = this.entering;
        var leaving = this.leaving;
        if (entering && leaving && entering[name] &&
          entering.travelMode === leaving.travelMode &&
          !entering.error && !leaving.error &&
          entering[name].canOptimize) {
          // This is the case of two adjacent requests from A to B and B to C
          // and the user finishes dragging B, the direction requests can be
          // batched together into a single request from A to C via B to reduce
          // the number of requests made that count against app quota
          makeOptimizedDirectionRequest(entering, this, leaving);
        } else {
          if (entering && entering[name]) {
            entering[name](!(leaving && leaving[name]));
          }
          if (leaving && leaving[name]) {
            leaving[name](true);
          }
        }
      },
      // change the marker style from green to white to red based on whether
      // this is the first, last, or middle marker in a route
      updateMarker: function () {
        this.marker.setOptions(
          this.entering ?
            this.leaving ?
              Point.MID_MARKER :
              Point.LAST_MARKER :
            Point.FIRST_MARKER
        );
      },
      // remove a point from the map & route
      remove: function () {
        this.marker.setMap(null);
        if (this.entering) {
          this.entering.end = null;
        }
        if (this.leaving) {
          this.leaving.start = null;
        }
      },
      // return the next point in the route
      next: function () {
        return this.leaving ? this.leaving.end : null;
      },
      // return the previous point in the route
      last: function () {
        return this.entering ? this.entering.start : null;
      },
      // serialize the point to a string
      serialize: function () {
        var position = this.marker.getPosition();
        return encode([position]);
      },
      makeStatic: function () {
        this.marker.setClickable(false);
        this.marker.setDraggable(false);
        return this;
      }
    };

    // API
    return Point;
  }
);