// Variation on the Ramer–Douglas–Peucker algorithm. 

// Normally this algorithm is used to remove points from a route that are
// under a certain distance threshold from what the line would be without them
// (fixed epsilon).  In our case, we have a fixed length string that the route
// can be encoded into so we want to choose the N most important points in the
// route such that the encoded route falls under our length limit.

// The RDP algorithm is used just to assign weights of importance to each
// point in the route, and an iterative algorithm given a weight function is
// used to find the most points with the most weight while that distance
// function is less than 1.

// See: http:en.wikipedia//.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm
define([
    'line'
  ],

  function (Line) {

    "use strict";

    // constants
    var METERS_PER_DEGREE = 111000; // approximate

    // private state
    var canceled = 0;

    // Find weights for all points in a route using the RDP algorithm then
    // invoke the callback with a datasctructure that can be used to query the N
    // most significant points.  This can be time consuming for long routes, so
    // don't do completely recursively, instead periodically put the recursive
    // call on the event queue so UI remains responsive and we don't blow the
    // call stack
    var characterize = function (points, callback, dontCancel) {
      // store the weights at indices corresponding to those in points
      var weights = [];
      // keep track of invocation
      var n = 0;
      var len = points.length;
      if (!dontCancel) {
        ++canceled;
      }
      var cancelLimit = canceled;
      var douglasPeucker = function (start, end, done) {
        if (!dontCancel && canceled > cancelLimit) {
          // do nothing
        } else if (end > start + 1) {
          var doWork = function () {
            // make a line from first to last point in this segment
            var line = new Line(points[start], points[end]);
            var maxDist = -1;
            var maxDistIndex = 0;
            var dist;
            // find point furthest off the line from start to end
            for (var i = start + 1; i < end; i += 1) {
              dist = line.distanceToSquared(points[i]);
              if (dist > maxDist) {
                maxDist = dist;
                maxDistIndex = i;
              }
            }
            // record the weight of this point
            weights[maxDistIndex] = maxDist;
            // split the segment at that furthest point, and recursively invoke
            // this method to assign weights to each point in the subsegments
            douglasPeucker(start, maxDistIndex, function () {
              // when done, invoke the callback for this invocation
              douglasPeucker(maxDistIndex, end, done);
            });
          };
          // every 50th invocation, break the call stack and put on the event
          // queue, letting UI events sneak in so it remains responsive
          if (n++ % 50 === 0) {
            setTimeout(doWork, 0);
          } else {
            doWork();
          }
        } else {
          done();
        }
      };

      // kick off the algorithm starting with the entire route
      douglasPeucker(0, len - 1, function () {
        // when done...
        // always use the first and last point
        weights[0] = Infinity;
        weights[len - 1] = Infinity;
        // create descending weight array so to get n most important weights,
        // just find all points with weights >= weightsDescending[n]
        var weightsDescending = weights.slice();
        weightsDescending.sort(function (a, b) {
          return b - a;
        });
        var threshold = function (n) {
          return n > 0 ? weightsDescending[n - 1] : Infinity;
        };

        // pass the datastructure to the original requestor
        callback({
          length: len,
          threshold: threshold,
          // get the N most important points from the route
          getFirstN: function (n) {
            // add all points with weights >= weightsDescending[n] to the result
            var tolerance = threshold(n);
            return this.getWithTolerance(tolerance);
          },
          getWithTolerance: function (tolerance) {
            var result = [];
            for (var i = 0; i < len; i++) {
              if (weights[i] >= tolerance) {
                result.push(points[i]);
              }
            }
            return result;
          }
        });
      });
    };

    // grab the maximum number of points from data while keeping weightFunc of
    // the encoding less than 1.  Use start as a starting point, and
    // increment/decrement by 1 until maximum number of points is found
    var maximize = function (data, start, weightFunc, encode) {
      // most points while staying weight < 1
      var maxSuccess = 0;
      var longestSuccess = [];
      var longestEncodedSuccess = encode([]);

      var normalizeStart = function () {
        start = start < 0 ? 0 : start;
        start = start >= data.length ? data.length - 1 : start;
      };
      // try encoding the N most important points and return the weight.  If
      // weight < 1 but this is the longest encoding so far, store it
      var tryOut = function (n) {
        var firstN = data.getFirstN(n);
        var encoded = encode(firstN);
        var weight = weightFunc(encoded);
        if (weight < 1 && n > maxSuccess) {
          maxSuccess = n;
          longestEncodedSuccess = encoded;
          longestSuccess = firstN;
        }
        return weight;
      };

      // initialize start
      normalizeStart();
      var startWeight = tryOut(start);
      // make starting guess better before iterating up or down
      if (startWeight) {
        start = Math.round(start / startWeight);
        normalizeStart();
        startWeight = tryOut(start);
      }

      // iterate up or down by 1 until the maximum number of points is found
      // such that weight is still < 1
      var idx = start;
      if (startWeight < 1) {
        // need to add points
        while (tryOut(++idx) < 1 && idx < data.length) {}
      } else {
        // need to cut points
        while (tryOut(--idx) >= 1 && idx > 0) {}
      }

      // calculate the furthest distance point from the line that had to be
      // thrown away, in meters
      var maxError = data.threshold ? Math.sqrt(data.threshold(maxSuccess)) * METERS_PER_DEGREE : 0;
      return {
        encoded: longestEncodedSuccess,
        points: longestSuccess,
        maxError: maxError,
        numRetained: maxSuccess,
        numOriginal: data.length
      };
    };

    // invoke the actual RDP algorithm to encode to points, shortening it to
    // the point where weightFunc(encoded) is the highest that it can be,
    // without going over 1, then invoke callback
    var shorten = function (points, weightFunc, encode, callback) {
      var initialEncode = encode(points);
      var initialWeight = weightFunc(initialEncode);
      if (initialWeight < 1) {
        // if # points is already short enough, don't bother characterizing
        callback({
          encoded: initialEncode,
          points: points,
          maxError: 0,
          numRetained: points.length,
          numOriginal: points.length
        });
      } else {
        // otherwise characterize using RDP, and apply maximization algorithm to
        // the results
        characterize(points, function (data) {
          var start = Math.round(points.length / initialWeight);
          var maximized = maximize(data, start, weightFunc, encode);
          callback(maximized);
        });
      }
    };

    var simplify = function (points, tolerance, callback) {
      var toleranceInDegrees = tolerance / METERS_PER_DEGREE;
      characterize(points, function (data) {
        callback(data.getWithTolerance(toleranceInDegrees * toleranceInDegrees));
      }, true);
    };

    // cancel in-progress calculations
    var cancel = function () { canceled++; };

    // API
    return {
      shorten: shorten,
      characterize: characterize,
      maximize: maximize,
      Line: Line,
      cancel: cancel,
      simplify: simplify
    };
  }
);
