define([
    'jquery',
    'segment',
    'point',
    'store',
    'url',
    'util',
    'rdp',
    'log'
  ],

  function (
    $,
    Segment,
    Point,
    Store,
    URL,
    Util,
    RDP,
    LOG) {

    "use strict";
    // imports
    var events = google.maps.event;
    var LatLngBounds = google.maps.LatLngBounds;

    // constants
    var ROUTE_KEY = "lastroute";

    // Set utility to maintain a unique list of elements
    var Set = function () {
      this.array = [];
      this.hash = {};
    };
    Set.prototype = {
      add: function (element) {
        if (element && !this.hash[element]) {
          this.hash[element] = true;
          this.array.push(element);
        }
      }
    };

    // the route class
    var Route = function (map, model) {
      var that = this;
      var $this = $(this);
      this.lastPoint = null;

      var changed = function () {
        $this.trigger('change');
      };

      // do the actual work of removing the last point in a route, without
      // invoking any events
      var doUndoLast = function () {
        var end = that.lastPoint;
        var segmentToRemove;
        that.lastPoint = null;
        if (end) {
          end.remove();
          segmentToRemove = end.entering;
          if (segmentToRemove) {
            segmentToRemove.remove();
            that.lastPoint = segmentToRemove.start;
          }
        }
      };

      // add a point to the route.  if latLng is a string, this point is being
      // deserialized and edge should be specified, otherwise this may result in
      // a direction request to obtain an edge
      var addPoint = function (latLng, edge) {
        var prev = that.lastPoint;
        // only add a point if the new location is different from last
        // ignore deserialization
        if (!prev || typeof latLng === 'string' || !prev.marker.getPosition().equals(latLng)) {
          var next = new Point(map, latLng, that);
          var segment;
          that.lastPoint = next;
          if (prev) {
            if (edge) {
              // restoring from serialized
              segment = new Segment(map, edge.t, prev, next, that, edge);
            } else {
              // user dropped a point
              segment = new Segment(map, model.getTravelMode(), prev, next, that, false);
              if (!model.getTravelMode()) {
                LOG('Direction', "direct", "(direct)");
              }
            }
            prev.leaving = segment;
            next.entering = segment;
            prev.updateMarker();
          }
          next.updateMarker();
        }
        return model.getTravelMode() === null;
      };

      // handler for add point events
      var addPointHandler = function (e) {
        var first = !that.lastPoint;
        addPoint(e.latLng);
        if (!model.getTravelMode() || first) {
          update();
        }
      };

      // utility to iterate through each point (will be reversed!)
      var forEachPoint = function (callback) {
        var point = that.lastPoint;
        while (point) {
          callback(point);
          point = point.entering && point.entering.start;
        }
      };

      // utility to iterate through each segment (will be reversed!)
      var forEachSegment = function (callback) {
        forEachPoint(function (point) {
          var entering = point.entering;
          if (entering) {
            callback(entering);
          }
        });
      };

      // recalculate the distance from each segment in the route and set it in
      // the model
      var updateDistance = function () {
        var distance = 0;
        forEachSegment(function (segment) {
          distance += segment.getDistance();
        });
        model.setMeters(distance);
      };

      // update the legal info in the model from warnings and copyrights in the
      // route
      var updateLegal = function () {
        var warnings = new Set();
        var copyrights = new Set();
        forEachSegment(function (segment) {
          if (segment.warnings) {
            $.each(segment.warnings, function (idx, warning) {
              warnings.add(warning);
            });
          }
          copyrights.add(segment.copyright);
        });
        model.setLegal({
          warnings: warnings.array,
          copyrights: copyrights.array
        });
      };

      // serialize the route to a string for persistent storage
      var serialize = function () {
        var state = [];
        forEachPoint(function (point) {
          state.push({
            e: point.entering && point.entering.serialize(),
            p: point.serialize()
          });
        });

        return JSON.stringify(state);
      };

      var getMergedPath = function () {
        var path = [];
        forEachPoint(function (point) {
          var entering = point.entering;
          if (entering && entering.polyline && entering.polyline.getMap()) {
            var thisPath = entering.polyline.getPath().getArray();
            var middlePoints = thisPath.slice(1, thisPath.length);
            path.unshift.apply(path, middlePoints);
          } else {
            path.unshift(point.marker.getPosition());
          }
        });
        return path;
      };

      // serialize the route for sharing (merge all segments)
      var getMergedEncodedPath = function (maxLength, weightFunc, callback) {
        var path = getMergedPath();
        if (arguments.length > 0) {
          RDP.shorten(path, weightFunc, Segment.encode, function (result) {
            callback(result);
          });
        } else {
          return Segment.encode(path);
        }
      };

      // deserialize the route from a string in persistent storage
      var deserialize = function (string) {
        var state = JSON.parse(string);
        var point;
        var i;
        for (i = state.length - 1; i >= 0; i -= 1) {
          point = state[i];
          addPoint(point.p, point.e);
        }
        updateDistance();
        updateLegal();
        changed();
      };

      // update all things in the model derived from the route
      var update = function () {
        updateDistance();
        updateLegal();
        updateStore();
        changed();
      };

      // remove the last point in the route and update the marker/model
      var undoLast = function () {
        doUndoLast();
        update();
        if (that.lastPoint) {
          that.lastPoint.updateMarker();
        }
      };

      // remove all points in the route then update the model
      var clear = function () {
        while (that.lastPoint) {
          doUndoLast();
        }
        update();
      };

      // serialize the route to a string and store in persistent storage
      var updateStore = function () {
        var serialized = serialize();
        Store.set(ROUTE_KEY, serialized);
      };

      // get the serialized route from persistent storage and load it up
      var restore = function () {
        var state = Store.get(ROUTE_KEY);
        if (state) {
          try {
            deserialize(state);
          } catch (e) {
            clear();
          }
        }
      };

      // return a LatLngBounds object that fits the entire route, or null if no points
      var getBounds = function () {
        var result = null;
        if (that.lastPoint) {
          var point = that.lastPoint.marker.getPosition();
          result = new LatLngBounds(point, point);
          forEachSegment(function (segment) {
            segment.polyline.getPath().forEach($.proxy(result, 'extend'));
          });
        }
        return result;
      };

      // reverse the route
      var reverse = function () {
        var point = that.lastPoint, lastPoint = point;
        while (point) {
          lastPoint = point;
          var prevEntering = point.entering;
          point.entering = point.leaving;
          point.leaving = prevEntering;
          point = prevEntering && prevEntering.start;
          if (prevEntering) {
            prevEntering.reverse();
          }
          lastPoint.updateMarker();
        }
        that.lastPoint = lastPoint;
        update();
        LOG('Reverse', "Reverse");
      };

      // create a new segment, start, and end from a polyline encoded in the
      // URL. Returns that LatLngBounds containing the entire route.
      var restoreMerged = function (encoded, needDistance) {
        var restored = Segment.restore(map, that, encoded);
        that.lastPoint = restored.end || restored.start;
        if (needDistance) {
          updateDistance();
        }
        changed();
        return getBounds();
      };


      // API
      $.extend(this, {
        addPoint: addPoint,
        clear: clear,
        undoLast: undoLast,
        update: update,
        restore: restore,
        getMergedEncodedPath: getMergedEncodedPath,
        restoreMerged: restoreMerged,
        getBounds: getBounds,
        reverse: reverse,
        getMergedPath: getMergedPath
      });

      // on single map clicks, add a new point (only if not a shared route)
      if (!URL.params.context) {
        events.addListener(map, 'singleclick', addPointHandler);
      }
    };

    // module API
    return Route;
  }
);