define([
    'jquery',
    'point',
    'directions',
    'bisectpolyline',
    'gmapsingleclick'
  ],

  function (
    $,
    Point,
    Directions,
    bisectPolyline,
    setupSingleClick) {

    "use strict";

    // imports
    var events = google.maps.event;
    var Polyline = google.maps.Polyline;
    var interpolate = google.maps.geometry.spherical.interpolate;
    var computeLength = google.maps.geometry.spherical.computeLength;
    var encode = google.maps.geometry.encoding.encodePath;
    var decode = google.maps.geometry.encoding.decodePath;

    // polyline styles
    var POLYLINE_STYLE = {
      strokeColor: '#0000ff',
      strokeOpacity: 0.6,
      clickable: false,
      geodesic: true
    };
    var TEMP_POLYLINE_STYLE = {
      strokeColor: '#0000ff',
      strokeOpacity: 0.2,
      clickable: false,
      geodesic: true
    };
    var ERROR_POLYLINE_STYLE = {
      strokeColor: '#ff0000',
      strokeOpacity: 0.6,
      clickable: false,
      geodesic: true
    };
    var MOUSEOVER_STYLE = {
      strokeOpacity: 0,
      strokeWeight: 10,
      clickable: true,
      draggable: true,
      geodesic: true
    };

    // internal implementation of a direct segment
    // all fields unique to each direct segment instance
    var DirectSegment = function (map, start, end, route, edge, doUpdate) {
      this.route = route;
      this.start = start;
      this.end = end;
      this.map = map;
      this.polyline = new Polyline(POLYLINE_STYLE);
      this.tempPolyline = new Polyline(TEMP_POLYLINE_STYLE);
      this.listenForMouseover();
      this.polyline.setMap(map);
      this.finishUpdate(doUpdate);
      if (edge) {
        this.distance = edge.d;
      }
    };

    // methods shared among all direct segment instances
    DirectSegment.prototype = {
      // return the midpoint of a direct segment
      getMidpoint: function () {
        return interpolate(
          this.start.marker.getPosition(),
          this.end.marker.getPosition(),
          0.5
        );
      },
      // when a user starts dragging an endpoint, begin to display the temp
      // polyline
      startUpdate: function () {
        this.tempPolyline.setMap(this.map);
        this.update();
      },
      // when the position of an endpoint changes during a drag, update the temp
      // polyline
      update: function () {
        this.tempPolyline.setPath([
          this.start.marker.getPosition(),
          this.end.marker.getPosition()
        ]);
      },
      // when the user finishes dragging an endpoint, stop displaying the temp
      // polyline and update the distance of the route and the splitter marker
      finishUpdate: function (updateRoute) {
        this.clearCached();
        this.tempPolyline.setMap(null);
        this.polyline.setMap(this.map);
        var startPoint = this.start.marker.getPosition();
        var endPoint = this.end.marker.getPosition();
        var points = [startPoint, endPoint];
        this.polyline.setPath(points);
        this.pathChanged();
        var lastSeg = this.start.entering && this.start.entering.pathrev;
        var nextSeg = this.end.leaving && this.end.leaving.pathrev;
        this.pathrev = {
          points: [
            this.end.leaving instanceof RoadSegment && nextSeg ? nextSeg.points[nextSeg.points.length - 1] : endPoint,
            this.start.entering instanceof RoadSegment && lastSeg ? lastSeg.points[0] : startPoint
          ]
        };
        this.path = {
          points: points
        };
        this.splitter.marker.setPosition(this.getMidpoint());
        if (updateRoute !== false) {
          this.route.update();
        }
      },
      // when an adjacent route finishes updating and the point shared with this
      // route is different, update this polyline and the distance
      fixStart: function () {
        this.finishUpdate();
      },
      fixEnd: function () {
        this.finishUpdate();
      },
      // when the user begins dragging the "splitter" midpoint marker, create 2
      // new polylines that split the route in two and display their live-
      // updating temp polylines - don't actually display the polylines until
      // the user finishes dragging
      split: function () {
        var oldPolyline = this.polyline;
        var oldMouseover = this.mouseoverpoly;
        var splitter = this.splitter;
        var newFirstHalf = new DirectSegment(
          null,
          this.start,
          splitter,
          this.route
        );
        var newSecondHalf = new DirectSegment(
          null,
          splitter,
          this.end,
          this.route
        );
        newFirstHalf.map = this.map;
        newSecondHalf.map = this.map;
        newFirstHalf.startUpdate();
        newSecondHalf.startUpdate();
        this.start.leaving = newFirstHalf;
        splitter.entering = newFirstHalf;
        this.end.entering = newSecondHalf;
        splitter.leaving = newSecondHalf;
        splitter.updateMarker();
        oldMouseover.setPath([]);
        events.addListener(splitter.marker, 'dragend', function () {
          oldPolyline.setMap(null);
          oldMouseover.setMap(null);
          events.clearInstanceListeners(oldMouseover);
        });
      },
      // serialize this segment to a string (just cache the distance)
      serialize: function () {
        return {
          d: this.distance
        };
      }
    };

    // internal implementation of a follow road segment
    // all fields unique to a road segment instance
    var RoadSegment = function (map, travelMode, start, end, route, edge, delay) {
      this.route = route;
      this.start = start;
      this.end = end;
      this.travelMode = travelMode;
      this.map = map;
      this.polyline = new Polyline(POLYLINE_STYLE);
      this.tempPolyline = new Polyline(TEMP_POLYLINE_STYLE);
      this.listenForMouseover();
      if (edge) {
        this.errorrev = edge.er;
        var rev =  edge.pr ? {
          distance: edge.dr,
          points: decode(edge.pr)
        } : null;
        // restored route
        if (edge.e || !edge.p) {
          this.handleDirectionError();
          this.pathrev = rev;
          this.errorrev = edge.er;
        } else {
          this.handleDirectionResponse(
            edge.w,
            edge.c,
            {
              distance: edge.d,
              points: decode(edge.p)
            }, edge.er ? Directions.PATH_ERROR : rev,
            {},
            true
          );
        }
      } else if (!delay) {
        // user-dragged route
        this.finishUpdate();
      }
    };

    // methods shared among all road segment instances
    RoadSegment.prototype = {
      // return the midpoint of a direct segment
      getMidpoint: function () {
        // TODO
      },
      // when a user starts dragging an endpoint, begin to display the temp
      // polyline
      startUpdate: function () {
        this.tempPolyline.setMap(this.map);
        // cancel pending request
        if (this.lastReq) {
          this.lastReq.canceled = true;
        }
      },
      // when the position of an endpoint changes during a drag, update the temp
      // polyline
      update: function () {
        this.tempPolyline.setPath([
          this.start.marker.getPosition(),
          this.end.marker.getPosition()
        ]);
      },
      // when the user finishes dragging an endpoint, stop displaying the temp
      // polyline and update the distance of the route and the splitter marker
      finishUpdate: function () {
        this.tempPolyline.setMap(null);
        this.clearCached();
        var from = this.start.marker.getPosition();
        var to = this.end.marker.getPosition();
        this.lastReq = Directions.get(from, to, this.travelMode,
          $.proxy(this, 'handleDirectionResponse'),
          $.proxy(this, 'handleDirectionError'));
      },
      // makes a straight line between the start and endpoint
      makeDirectLine: function () {
        this.polyline.setPath([
          this.start.marker.getPosition(),
          this.end.marker.getPosition()
        ]);
        this.pathChanged();
      },
      // when an adjacent route finishes updating the previous segment and the
      // route comes back different - prepend a segment to this polyline to
      // fudge the difference
      fixStart: function () {
        if (this.error) {
          this.makeDirectLine();
        } else {
          var startPos = this.start.marker.getPosition();
          var path = this.polyline.getPath();
          var temp = this.tempPolyline;
          var polyStart;
          if (this.polyline.getMap() !== null) {
            polyStart = path.getAt(0);
            if (polyStart && !polyStart.equals(startPos)) {
              // add segment to start
              path.insertAt(0, startPos);
              this.clearCached();
            }
          }
          if (temp && temp.getMap() !== null) {
            temp.getPath().setAt(0, startPos);
          }
        }
      },
      // when an adjacent route finishes updating the next segment and the
      // route comes back different - append a segment to this polyline to
      // fudge the difference
      fixEnd: function () {
        if (this.error) {
          this.makeDirectLine();
        } else {
          var endPos = this.end.marker.getPosition();
          var path = this.polyline.getPath();
          var temp = this.tempPolyline;
          var polyEnd;
          if (this.polyline.getMap() !== null) {
            polyEnd = path.getAt(path.getLength() - 1);
            if (polyEnd && !polyEnd.equals(endPos)) {
              // add segment to end
              path.push(endPos);
              this.clearCached();
            }
          }
          if (temp && temp.getMap() !== null) {
            temp.getPath().setAt(1, endPos);
          }
        }
      },
      // handle a response from google direction service
      handleDirectionResponse: function (warnings, copyright, path, pathrev, handle, disableUpdate) {
        if (handle.canceled) { return; }
        var errorrev = pathrev === Directions.PATH_ERROR;
        this.pathrev = errorrev ? null : pathrev;
        this.path = path;
        this.errorrev = errorrev;
        if (handle.reversed) {
          this.pathrev = path;
          this.errorrev = false;
          if (pathrev === null || errorrev) {
            this.handleDirectionError(true);
            return;
          }
          path = pathrev;
        }
        var points = path.points;
        this.polyline.setOptions(POLYLINE_STYLE);
        this.error = false;
        this.warnings = warnings;
        this.copyright = copyright;
        this.clearCached();
        this.distance = path.distance;
        this.polyline.setPath(points);
        this.pathChanged();
        this.polyline.setMap(this.map);
        this.tempPolyline.setMap(null);
        this.start.marker.setPosition(points[0]);
        this.end.marker.setPosition(points[points.length - 1]);
        // tell adjacent route segments to fix if endpoints don't match
        var last = this.start.entering;
        var next = this.end.leaving;
        if (last && last.fixEnd) {
          last.fixEnd();
        }
        if (next && next.fixStart) {
          next.fixStart();
        }
        // need to tell route to update here since call on 'dragend' of point
        // occurs before the response received
        if (disableUpdate !== true) {
          this.route.update();
        }
      },
      // return an encoded string for this polyline, this is memoized until
      // changed to avoid recalculating often
      getEncoded: function () {
        if (!this.encoded) {
          this.encoded = encode(this.polyline.getPath());
        }
        return this.encoded;
      },
      // serialize this route to a string
      serialize: function () {
        return {
          d: this.distance,
          p: this.getEncoded(),
          dr: this.pathrev && !this.errorrev ? this.pathrev.distance : null,
          pr: this.pathrev && !this.errorrev ? encode(this.pathrev.points) : null,
          w: this.warnings,
          c: this.copyright,
          t: this.travelMode,
          e: this.error,
          er: this.errorrev
        };
      },
      // when the user begins dragging the "splitter" midpoint marker, create 2
      // new polylines that split the route in two and display their live-
      // updating temp polylines - don't actually display the polylines until
      // the user finishes dragging
      split: function () {
        var oldPolyline = this.polyline;
        var oldMouseover = this.mouseoverpoly;
        var splitter = this.splitter;
        var newFirstHalf = new RoadSegment(
          this.map,
          this.travelMode,
          this.start,
          splitter,
          this.route,
          null,
          true
        );
        var newSecondHalf = new RoadSegment(
          this.map,
          this.travelMode,
          splitter,
          this.end,
          this.route,
          null,
          true
        );
        newFirstHalf.map = this.map;
        newSecondHalf.map = this.map;
        newFirstHalf.startUpdate();
        newSecondHalf.startUpdate();
        this.start.leaving = newFirstHalf;
        splitter.entering = newFirstHalf;
        this.end.entering = newSecondHalf;
        splitter.leaving = newSecondHalf;
        splitter.updateMarker();
        oldMouseover.setPath([]);
        events.addListener(splitter.marker, 'dragend', function () {
          oldPolyline.setMap(null);
          oldMouseover.setMap(null);
          events.clearInstanceListeners(oldMouseover);
        });
      }
    };

    // internal implementation of a restored segment
    var RestoredSegment = function (map, route, encoded) {
      var decoded = decode(encoded);
      var start = decoded[0];
      var end = decoded[decoded.length - 1];
      this.route = route;
      this.map = map;

      if (decoded.length > 0) {
        this.start = new Point(map, start, route, Point.FIRST_MARKER);
        this.start.makeStatic();
      }
      if (decoded.length > 1) {
        this.start.leaving = this;
        this.end = new Point(map, end, route, Point.LAST_MARKER).makeStatic();
        this.end.makeStatic();
        this.polyline = new Polyline(POLYLINE_STYLE);
        this.polyline.setMap(map);
        this.polyline.setPath(decoded);
        this.pathChanged();
        this.end.entering = this;
        this.listenForMouseover();
      }
    };

    // methods shared between road and direct segments
    var common = {
      // get the distance of this segement - memoized between calls until
      // changed
      getDistance: function () {
        if (!this.distance) {
          this.distance = computeLength(this.polyline.getPath());
        }
        return this.distance;
      },
      // clear all memoized values from this segment
      clearCached: function () {
        this.distance = null;
        this.encoded = null;
      },
      // remove this segment from the map and route
      remove: function () {
        this.polyline.setMap(null);
        this.mouseoverpoly.setMap(null);
        if (this.start) {
          this.start.leaving = null;
        }
        if (this.end) {
          this.end.entering = null;
        }
        if (this.lastReq) {
          this.lastReq.canceled = true;
        }
        if (this.splitter) {
          this.splitter.remove();
        }
      },
      // reverse start/end and polyline
      reverse: function () {
        var prevError = this.error;
        var prevErrorrev = this.errorrev;
        var prevStart = this.start;
        this.error = prevErrorrev;
        this.errorrev = prevError;
        this.start = this.end;
        this.end = prevStart;
        if (prevErrorrev) {
          this.pathrev = this.path;
          this.handleDirectionError(true);
        } else if (this.pathrev) {
          this.polyline.setOptions(POLYLINE_STYLE);
          var points = this.pathrev.points;
          this.polyline.setPath(points);
          this.pathChanged();
          this.distance = null;
          var oldpath = this.pathrev;
          this.pathrev = this.path;
          this.path = oldpath;
          this.start.marker.setPosition(points[0]);
          this.end.marker.setPosition(points[points.length - 1]);
        } else {
          var path = this.polyline.getPath().getArray();
          this.polyline.setPath(path.reverse());
          this.pathChanged();
        }
        if (this.tempPolyline) {
          var tempPath = this.tempPolyline.getPath().getArray();
          this.tempPolyline.setPath(tempPath.reverse());
        }
        this.encoded = null;
        if (this.lastReq) {
          this.lastReq.reversed = !this.lastReq.reversed;
        }
      },
      // Handle an error from google direction service
      handleDirectionError: function (ignoreRev) {
        this.error = true;
        if (!ignoreRev) {
          this.errorrev = true;
        }
        this.polyline.setOptions(ERROR_POLYLINE_STYLE);
        this.polyline.setPath([
          this.start.marker.getPosition(),
          this.end.marker.getPosition()
        ]);
        this.pathChanged();
        this.polyline.setMap(this.map);
        this.tempPolyline.setMap(null);
        this.route.update();
      },
      canSplit: function () { return !(this instanceof RestoredSegment); },
      createMouseoverPolyline: function () {
        var that = this;
        var mouseoverpoly = new Polyline(MOUSEOVER_STYLE);
        this.mouseoverpoly = mouseoverpoly;
        mouseoverpoly.setMap(this.map);
        mouseoverpoly.setPath(this.polyline.getPath().getArray());
        function update() {
          mouseoverpoly.setPath(that.polyline.getPath().getArray());
          mouseoverpoly.setMap(that.map);
        }
        events.addListener(this.polyline, 'pathchanged', update);
        return mouseoverpoly;
      },
      createSplitter: function () {
        this.splitter = new Point(
          this.map,
          this.start.marker.getPosition(),
          this.route,
          Point.SPLITTER_MARKER
        );
        this.splitter.marker.setMap(null);
        events.addListenerOnce(
          this.splitter.marker,
          'dragstart',
          $.proxy(this, 'split')
        );
        return this.splitter;
      },
      listenForMouseover: function () {
        var that = this;
        var mousedown = false;
        var dragging = false;
        var canSplit = this.canSplit();
        var mouseoverpoly = this.createMouseoverPolyline();
        var splitter;
        var route = this.route;

        // add listener on splitter to track drag event
        if (canSplit) {
          splitter = this.createSplitter();
          events.addListener(splitter.marker, 'mousedown', function () {
            mousedown = true;
          });
          events.addListener(splitter.marker, 'mouseup', function () {
            mousedown = false;
          });
        }

        // handle updating listeners with a "mouseover" @ distance X event when
        // mouse is over the wide polyline
        function handleMouseMove(evt) {
          var pos = evt.latLng;
          if (!dragging) {
            var base = 0;
            for (var seg = that.start.entering; seg; seg = seg.start.entering) {
              base += seg.getDistance();
            }
            var current = bisectPolyline(that.polyline.getPath().getArray(), pos).dist;
            $(route).trigger('mouseover', {
              position: [evt.latLng.lat(), evt.latLng.lng()],
              distance: base + current
            });
          }
          route.currentMouseover = that;
          if (canSplit) {
            if (!splitter.marker.getMap()) {
              splitter.marker.setMap(that.map);
            }
            splitter.marker.setPosition(evt.latLng);
          }
        }

        // handle updating listeners with a mouse out event when should cancel
        function handleMouseOut() {
          if (route.currentMouseover === that) {
            // no args tells listener mouseover is done
            $(route).trigger('mouseover');
          }
          if (canSplit && !mousedown) {
            splitter.marker.setMap(null);
          }
        }

        // fake drag events on splitter when it is actually the wide polyline
        // being dragged
        if (canSplit) {
          events.addListener(mouseoverpoly, 'dragstart', function (evt) {
            $(route).trigger('mouseover');
            dragging = true;
            events.trigger(splitter.marker, 'dragstart', evt);
          });
          events.addListener(mouseoverpoly, 'drag', function (evt) {
            events.trigger(splitter.marker, 'drag', evt);
          });
          events.addListener(mouseoverpoly, 'dragend', function (evt) {
            events.trigger(splitter.marker, 'dragend', evt);
          });
        }

        // bind event listeners
        events.addListener(mouseoverpoly, 'mouseover', handleMouseMove);
        events.addListener(mouseoverpoly, 'mousemove', handleMouseMove);
        events.addListener(mouseoverpoly, 'mousedown', handleMouseMove);
        events.addListener(mouseoverpoly, 'mouseout', handleMouseOut);

        // allow a mouseclick on the route to add a segment looping back to that point
        if (canSplit) {
          setupSingleClick(mouseoverpoly);
          events.addListener(mouseoverpoly, 'singleclick', function (evt) {
            handleMouseOut();
            if (route.addPoint(evt.latLng)) {
              route.update();
            }
          });
        }
      },
      pathChanged: function () {
        events.trigger(this.polyline, 'pathchanged');
      }
    };
    $.extend(RoadSegment.prototype, common);
    $.extend(DirectSegment.prototype, common);
    $.extend(RestoredSegment.prototype, common);

    // updates to adjacent road segments can be optimized into a single request
    RoadSegment.prototype.finishUpdate.canOptimize = true;

    var Segment = function (map, travelMode, start, end, route, edge) {
      // choose the correct internal implementation based on if there is a
      // travel mode or not
      return travelMode ?
        new RoadSegment(map, travelMode, start, end, route, edge) :
        new DirectSegment(map, start, end, route, edge, edge);
    };

    Segment.restore = function (map, route, encoded) {
      return new RestoredSegment(map, route, encoded);
    };

    // Segment API
    Segment.POLYLINE_STYLE = POLYLINE_STYLE;
    Segment.TEMP_POLYLINE_STYLE = TEMP_POLYLINE_STYLE;
    Segment.ERROR_POLYLINE_STYLE = ERROR_POLYLINE_STYLE;
    Segment.MOUSEOVER_STYLE = MOUSEOVER_STYLE;
    Segment.encode = encode;

    // module API
    return Segment;
  }
);