define([
    'jquery',
    'util',
    'log',
    'compressroute'
  ],

  function (
    $,
    Util,
    LOG,
    CompressRoute) {

    "use strict";
    var IFRAME_WIDTH = 425;
    var IFRAME_HEIGHT = 350;
    var MAX_URL_LENGTH = 1900;
    var MapTypeId = google.maps.MapTypeId;
    var EXPAND_VALUES = {
      r: MapTypeId.ROADMAP,
      s: MapTypeId.HYBRID,
      t: MapTypeId.TERRAIN
    };
    var SHORTEN_VALUES = Util.invert(EXPAND_VALUES);

    // create iframe code from a URL, and encode ampersands
    var createIframeCode = function (url) {
      return [
        '<iframe width="',
        IFRAME_WIDTH,
        '" height="',
        IFRAME_HEIGHT,
        '" frameborder="0" scrolling="no" marginheight="0" marginwidth="0" src="',
        $("<div>").html(url).html(),
        '"></iframe><br/><small><a href="http://onthegomap.com" style="color:#0000FF;text-align:left" target="_blank">On The Go Map</a></small>'
      ].join("");
    };

    var encodeValue = function (value) {
      value = SHORTEN_VALUES[value] || value;
      return encodeURIComponent(value)
        .replace(/[!'()]/g, escape)
        .replace(/\*/g, "%2A")
        .replace(/%20/g, "+");
    };

    // encode a JavaScript object into a query string
    // ie {a: 1, b: 2} -> "a=1&b=2"
    var encode = function (obj) {
      var parts = [];
      $.each(obj, function (key, value) {
        if (value === true) {
          parts.push(key);
        } else if ($.isArray(value)) {
          if (value.length > 0) {
            var mapped = $.map(value, function (toMap) {
              return encodeValue(toMap);
            });
            parts.push([
              encodeURIComponent(key + "[]"), mapped.join(",")
            ].join("="));
          }
        } else if (!!value) {
          parts.push([
            key, encodeValue(value)
          ].join("="));
        }
      });
      return parts.join("&");
    };

    var decodeValue = function (value) {
      value = EXPAND_VALUES[value] || value;
      return decodeURIComponent(value.replace(/\+/g, " "));
    };

    // decode a JavaScript object from a query string
    // ie "a=1&b=2" -> {a: 1, b: 2}
    var decode = function (str) {
      var result = {};
      // first parse the route out of the URI
      var routeStart = str.indexOf("&r=");
      if (routeStart > -1) {
        result.r = decodeURI(str.substr(routeStart + 3));
        str = str.substr(0, routeStart);
      } else if ((routeStart = str.indexOf("&r2=")) > -1) {
        result.r = CompressRoute.decompress(str.substr(routeStart + 3));
        str = str.substr(0, routeStart);
      }
      // then parse the rest of the URI
      var parts = str.split("&");
      for (var i = 0; i < parts.length; i += 1) {
        var thisPart = parts[i].split('=');
        if (thisPart.length === 1 && thisPart[0].length > 0) {
          // length 1 means true
          result[thisPart[0]] = "true";
        } else if (thisPart.length === 2) {
          var key = decodeURIComponent(thisPart[0]);
          var value = thisPart[1];
          if (key.substr(key.length-2) === "[]") {
            result[key.replace("[]", "")] = $.map(value.split(","), decodeValue);
          } else {
            result[key] = decodeValue(value);
          }
        }
      }
      return result;
    };

    // build a query object to be encoded from the map, model, and route
    // provided
    var buildQuery = function (map, model, route) {
      var center = map.getCenter();
      // only store viewport when no route
      var obj = {
        m: map.getMapTypeId(),
        u: model.getUnit(),
        b: model.getBikeLayer(),
        w: model.getLegal().warnings,
        c: model.getLegal().copyrights,
        d: Math.round(model.getMeters())
      };

      if (!route.lastPoint) {
        $.extend(obj, {
          lat: Util.round(center.lat(), 6),
          lng: Util.round(center.lng(), 6),
          zoom: map.getZoom()
        });
      }
      return obj;
    };

  // build a URL from the query object provided
    var createUrlFromQueryObject = function (queryObject) {
      var encoded = encode(queryObject);
      return ["http://onthegomap.com/", encoded].join("?");
    };

    // top-level function to build a link URL and iframe snippet from the map,
    // model, route, and current URL
    var build = function (map, model, route, callback) {
      var queryObject = buildQuery(map, model, route);
      var linkUrl = createUrlFromQueryObject(queryObject);
      var embedUrl = linkUrl;
      linkUrl += "&context=share";
      embedUrl += "&context=embed";
      var allowedForRoute = MAX_URL_LENGTH - linkUrl.length - 2;
      var stopProfile = LOG.profile('CPU', 'Encode Route', model.getDistanceText());
      route.getMergedEncodedPath(allowedForRoute, function (routeStr) {
        var fullURL = linkUrl + "&r=" +  CompressRoute.compress(routeStr);
        return encodeURIComponent(fullURL).length / MAX_URL_LENGTH;
      }, function (result) {
        var encodedRoute = result.encoded;
        var maxError = result.maxError || 0;
        // also:
        // result.numRetained
        // result.numOriginal
        if (encodedRoute) {
          var encoded = CompressRoute.compress(encodedRoute);
          linkUrl += "&r2=" + encoded;
          embedUrl += "&r2=" + encoded;
        }
        callback(linkUrl, createIframeCode(embedUrl), result.maxError);
        // log max error
        LOG('Share Link', 'Create Link', model.getDistanceText(), Math.round(maxError));
        stopProfile();
      });
    };

    // parse the search string out of a URL and the parameters out of that
    var parse = function (url) {
      var href = $("<a>")
        .attr("href", url)[0];
      var params = href.search;
      return decode(params.substring(1));
    };

    // API
    return {
      encode: encode,
      decode: decode,
      build: build,
      buildQuery: buildQuery,
      parse: parse,
      iframe: createIframeCode,
      params: parse(window.location.href)
    };
  }
);