define([
    "jquery",
    "directions"
  ],

  function (
    $,
    Directions) {

    /*jshint camelcase:false*/
    "use strict";

    // imports
    var DIR = google.maps.DirectionsService;
    var TravelMode = google.maps.TravelMode;
    var THROTTLE = google.maps.DirectionsStatus.OVER_QUERY_LIMIT;
    var OK = google.maps.DirectionsStatus.OK;
    var ERR = google.maps.DirectionsStatus.NOT_FOUND;

    // constants
    var A = "A";
    var B = "B";
    var C = "C";
    var D = "D";

    // test state
    var error = 0;
    var requests = 0;
    var called = 0;
    var callbacksInvoked = 0;
    var clock;
    var warning, copyright, points, points2, pointsrev, pointsrev2;

    // test utilities
    var fakeDirections = function () {
      var captured = [];
      callbacksInvoked = 0;
      error = 0;
      warning = null;
      copyright = null;
      called = 0;
      points = null;
      points2 = null;
      pointsrev = null;
      pointsrev2 = null;
      DIR.mockImpl = function () {
        captured.push($.makeArray(arguments));
      };
      return function () {
        var args = arguments;
        var options = [];
        var snapshot = captured.slice();
        captured = [];
        $.each(snapshot, function (idx, capturedArgs) {
          capturedArgs[1].apply(null, args);
          options.push(capturedArgs[0]);
        });
        return options;
      };
    };
    var recordArguments = function (a, b, c, d, e, f, g) {
      callbacksInvoked++;
      called++;
      warning = a;
      copyright = b;
      points = c;
      points2 = d;
      pointsrev = f;
      pointsrev2 = g;
    };
    var recordError = function () {
      called++;
      error++;
    };
    var checkPoints = function (c) {
      deepEqual(points, c, "expected points");
    };
    var checkPoints2 = function (c) {
      deepEqual(points2, c, "expected points 2");
    };
    var argumentsShouldBe = function (a, b, c, d) {
      ok(!error, "error should not have occurred");
      deepEqual(warning, a, "expected warning");
      strictEqual(copyright, b, "expected copyright");
      checkPoints(c);
      if (d) {
        checkPoints2(d);
      }
    };
    var recordRequests = function () {
      DIR.mockImpl = function (options, callback) {
        requests++;
        callback({}, OK);
      };
    };
    var waitToRespond = function () {
      var queuedUp = [];
      DIR.mockImpl = function (options, callback) {
        requests++;
        queuedUp.push(callback);
      };
      return function (status) {
        queuedUp[0]({}, status);
        queuedUp.shift();
      };
    };
    var makeRequests = function (num) {
      var requests = [];
      for (var i = 0; i < num; i += 1) {
        requests.push(Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError));
      }
      return function () {
        requests[0].canceled = true;
        requests.shift();
      };
    };
    var assertRequestsMade = function (num, description) {
      strictEqual(requests, num, "num requests made " + description);
      requests = 0;
    };
    var assertErrorCallbacksInvoked = function (num, description) {
      strictEqual(error, num, "num error callbacks invoked " + description);
      error = 0;
    };
    var assertCallbacksInvoked = function (num, description) {
      strictEqual(callbacksInvoked, num, "num callbacks invoked " + description);
      callbacksInvoked = 0;
    };




    ////////////////////////////////////////////////////////////////////

    module("Directions tests", {
      setup: function () {
        DIR.mockImpl = null;
        Directions.throttle(false);
        window.ga = sinon.stub();
      },
      teardown: function () {
        Directions.throttle(true);
      }
    });

    test("test basic happy path response 2 points", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          copyrights: "copyright string",
          warnings: ["warning a", "warning b"],
          legs: [{
            start_location: A,
            end_location: B,
            distance: {
              value: 12
            },
            steps: [{
              start_location: A,
              end_location: B,
              path: [A, B]
            }]
          }]
        }]
      }, OK);

      argumentsShouldBe(
        ["warning a", "warning b"],
        "copyright string",
        {
          points: [A, B],
          distance: 12
        }
      );
    });

    test("test basic happy path response 3 points", function () {
      var respond = fakeDirections();
      Directions.getPair(A, B, C, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          copyrights: "copyright string",
          warnings: ["warning a", "warning b"],
          legs: [{
            distance: {
              value: 12
            },
            steps: [{
              path: [A, B]
            }]
          }, {
            distance: {
              value: 13
            },
            steps: [{
              path: [B, C]
            }]
          }]
        }]
      }, OK);

      argumentsShouldBe(
        ["warning a", "warning b"],
        "copyright string",
        {
          points: [A, B],
          distance: 12
        },
        {
          points: [B, C],
          distance: 13
        }
      );
    });

    test("test extract more than 2 points", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          legs: [{
            steps: [{
              path: [A, B, C, D]
            }]
          }]
        }]
      }, OK);
      checkPoints({points: [A, B, C, D]});
      strictEqual(pointsrev, "error");
      strictEqual(pointsrev2, "error");
    });

    test("test extract points from more than 2 steps", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          legs: [{
            steps: [{
              path: [A, B]
            }, {
              path: [C, D]
            }]
          }]
        }]
      }, OK);
      checkPoints({points: [A, B, C, D]});
    });

    // error handling

    test("error response - 1 point", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      var options = respond({}, ERR);
      ok(called, "callback should be invoked");
      ok(error, "error handler invoked");
      deepEqual(options[0], {
        origin: A,
        destination: A,
        waypoints: [{
          location: B
        }],
        travelMode: TravelMode.WALKING
      }, "options passed into directions service");
    });

    test("error response - midpoint", function () {
      var respond = fakeDirections();
      Directions.getPair(A, B, C, TravelMode.WALKING, recordArguments, recordError);
      var options = respond({}, ERR);
      ok(called, "callback should be invoked");
      ok(error, "error handler invoked");
      deepEqual(options[0], {
        origin: A,
        destination: A,
        travelMode: TravelMode.WALKING,
        waypoints: [{
          location: B
        }, {
          location: C
        }, {
          location: B
        }]
      }, "options passed into directions service");
    });

    test("ZERO_RESULTS on round-trip retries forward only", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      var options = respond({}, google.maps.DirectionsStatus.ZERO_RESULTS);
      options = respond({}, google.maps.DirectionsStatus.OK);
      ok(called, "callback should be invoked");
      ok(!error, "error handler invoked");
      deepEqual(options[0], {
        origin: A,
        destination: B,
        travelMode: TravelMode.WALKING
      }, "options passed into directions service on 2nd try");
    });

    test("ZERO_RESULTS on round-trip getPair retries forward only", function () {
      var respond = fakeDirections();
      Directions.getPair(A, B, C, TravelMode.WALKING, recordArguments, recordError);
      var options = respond({}, google.maps.DirectionsStatus.ZERO_RESULTS);
      options = respond({}, google.maps.DirectionsStatus.OK);
      ok(called, "callback should be invoked");
      ok(!error, "error handler invoked");
      deepEqual(options[0], {
        origin: A,
        destination: C,
        travelMode: TravelMode.WALKING,
        waypoints: [{
          location: B
        }]
      }, "options passed into directions service on 2nd try");
    });

    test("ZERO_RESULTS twice is failure", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      var options = respond({}, google.maps.DirectionsStatus.ZERO_RESULTS);
      options = respond({}, google.maps.DirectionsStatus.ZERO_RESULTS);
      ok(!callbacksInvoked, "callback should not be invoked");
      ok(error, "error handler should be invoked");
    });

    test("ZERO_RESULTS twice on getPair is failure", function () {
      var respond = fakeDirections();
      Directions.getPair(A, B, C, TravelMode.WALKING, recordArguments, recordError);
      var options = respond({}, google.maps.DirectionsStatus.ZERO_RESULTS);
      options = respond({}, google.maps.DirectionsStatus.ZERO_RESULTS);
      ok(!callbacksInvoked, "callback should not be invoked");
      ok(error, "error handler should be invoked");
    });

    // request 2-step directions

    test("test request 2-step direction happy-path", function () {
      var respond = fakeDirections();
      Directions.getPair(A, B, C, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          legs: [{
            distance: {
              value: 1
            },
            steps: [{
              path: [A, B]
            }]
          }, {
            distance: {
              value: 2
            },
            steps: [{
              path: [B, C]
            }]
          }]
        }]
      }, OK);
      checkPoints({distance: 1, points: [A, B]});
      checkPoints2({distance: 2, points: [B, C]});
    });

    test("test request 2-step direction empty second path", function () {
      var respond = fakeDirections();
      Directions.getPair(A, B, C, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          legs: [{
            steps: [{
              path: [A, B]
            }]
          }]
        }]
      }, OK);
      checkPoints({points: [A, B]});
      checkPoints2({points: []});
    });


    /////////////////// handle funny responses

    test("test second leg ignored", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          legs: [{
            steps: [{
              path: [A, B]
            }]
          }, {
            steps: [{
              path: [C, D]
            }]
          }]
        }]
      }, OK);
      checkPoints({points: [A, B]});
    });

    test("test response with no steps", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          copyrights: "copyright string",
          warnings: ["warning a", "warning b"],
          legs: [{
            start_location: A,
            end_location: B
          }]
        }]
      }, OK);

      argumentsShouldBe(
        ["warning a", "warning b"],
        "copyright string",
        {points: []}
      );
    });

    test("test response with empty step list", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          copyrights: "copyright string",
          warnings: ["warning a", "warning b"],
          legs: [{
            start_location: A,
            end_location: B,
            steps: []
          }]
        }]
      }, OK);

      argumentsShouldBe(
        ["warning a", "warning b"],
        "copyright string",
        {points: []}
      );
    });
    test("test response with no legs", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          copyrights: "copyright string",
          warnings: ["warning a", "warning b"]
        }]
      }, OK);

      argumentsShouldBe(
        ["warning a", "warning b"],
        "copyright string",
        {points: []}
      );
    });

    test("test response with empty leg list and copyright/warning", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({
        routes: [{
          legs: []
        }]
      }, OK);

      argumentsShouldBe(
        [],
        null,
        {points: []}
      );
    });

    test("test response with no routes", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({}, OK);

      argumentsShouldBe(
        [],
        null,
        {points: []}
      );
    });

    test("test response with empty route list", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({routes: []}, OK);

      argumentsShouldBe(
        [],
        null,
        {
          points: []
        }
      );
    });

    test("test error logged", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({}, ERR);
      deepEqual(window.ga.args, [['send', 'event', 'Direction', TravelMode.WALKING, ERR]]);
    });

    test("test request denied error serious", function () {
      var respond = fakeDirections();
      Directions.get(A, B, TravelMode.WALKING, recordArguments, recordError);
      respond({}, google.maps.DirectionsStatus.REQUEST_DENIED);
      deepEqual(window.ga.args[1], ['send', 'event', 'error', 'Direction request denied']);
    });



    //////////////////////////////////////////////////////////////////


    module("rate throttling", {
      setup: function () {
        callbacksInvoked = 0;
        recordRequests();
        requests = 0;
        error = 0;
        clock = sinon.useFakeTimers();
        Directions.throttle(true);
        Directions.reset();
      },
      teardown: function () {
        clock.tick(100000);
        clock.restore();
      }
    });

    test("can make up to 5 requests in a second", function () {
      makeRequests(5);
      assertRequestsMade(5);
    });

    test("6th request gets delayed 1 second", function () {
      makeRequests(6);
      assertRequestsMade(5);
      clock.tick(1000);
      assertRequestsMade(1);
    });

    test("6th and 7th request gets delayed 1 second", function () {
      makeRequests(7);
      assertRequestsMade(5);

      clock.tick(999);
      assertRequestsMade(0);
      clock.tick(1);
      assertRequestsMade(1);

      clock.tick(999);
      assertRequestsMade(0);
      clock.tick(1);
      assertRequestsMade(1);


      clock.tick(5000);
      assertRequestsMade(0);
    });

    test("rate limit cleared after 5 seconds", function () {
      makeRequests(6);
      assertRequestsMade(5);

      clock.tick(999);
      assertRequestsMade(0);
      clock.tick(1);
      assertRequestsMade(1);

      clock.tick(5001);
      makeRequests(7);
      assertRequestsMade(5);

      clock.tick(999);
      assertRequestsMade(0);
      clock.tick(1);
      assertRequestsMade(1);
    });

    test("rate limit not cleared before 5 seconds", function () {
      makeRequests(6);
      assertRequestsMade(5);

      clock.tick(999);
      assertRequestsMade(0);
      clock.tick(1);
      assertRequestsMade(1);

      clock.tick(3000);
      makeRequests(2);
      assertRequestsMade(1, "during rate limit");

      clock.tick(999);
      assertRequestsMade(0, "<1s later");
      clock.tick(1);
      assertRequestsMade(1, "after 1s during rate limit");
    });

    test("Wait 2 seconds to send next request on OVER_QUERY_LIMIT error code", function () {
      var respond = waitToRespond();
      makeRequests(1);
      assertRequestsMade(1, "initially");
      respond(THROTTLE);
      assertErrorCallbacksInvoked(0);
      makeRequests(1);
      assertRequestsMade(0, "on send after received OQL");

      clock.tick(1999);
      assertRequestsMade(0, "<2s after OQL");
      clock.tick(1);
      assertRequestsMade(2, "2s after OQL");
    });

    test("On 2 consecutive OQL's, assume daily limit succeeded and respond error", function () {
      var respond = waitToRespond();
      makeRequests(1);
      assertRequestsMade(1, "initially");
      respond(THROTTLE);
      assertErrorCallbacksInvoked(0);

      clock.tick(1999);
      assertRequestsMade(0, "<2s after OQL");
      clock.tick(1);
      assertRequestsMade(1, "2s after OQL");
      assertErrorCallbacksInvoked(0);

      respond(THROTTLE);
      assertErrorCallbacksInvoked(1, "after throttled twice in a row");
      clock.tick(5000);
      assertRequestsMade(0, "after getting shut off");
      deepEqual(window.ga.args[window.ga.args.length - 1], ['send', 'event', 'error', 'Direction quota']);
    });

    test("test usage exceeded twice in a row", function () {
      var respond = waitToRespond();
      makeRequests(5);
      respond(OK);
      respond(OK);
      respond(OK);
      respond(OK);
      respond(THROTTLE);
      assertRequestsMade(5, "initially");
      assertErrorCallbacksInvoked(0);

      clock.tick(1000);
      makeRequests(5);
      assertRequestsMade(0, "during blackout");
      assertErrorCallbacksInvoked(0);

      clock.tick(1000);
      assertRequestsMade(5, "after blackout expires");
      respond(OK);
      respond(OK);
      respond(OK);
      respond(OK);
      respond(THROTTLE);
      assertErrorCallbacksInvoked(0);

      clock.tick(1999);
      assertRequestsMade(0, "during second blackout");

      clock.tick(1);
      assertRequestsMade(1, "after second blackout");
      respond(THROTTLE);
      assertErrorCallbacksInvoked(1);
    });

    // canceling a request

    test("test cancel a request", function () {
      var respond = waitToRespond();
      var cancel = makeRequests(1);
      cancel();
      assertCallbacksInvoked(0, "when request canceled");
      respond(OK);
      assertCallbacksInvoked(0, "when request canceled");
    });

    test("test cancel a request while queued up", function () {
      var respond = waitToRespond();
      makeRequests(5);
      var cancel = makeRequests(1);
      respond(OK);
      respond(OK);
      respond(OK);
      respond(OK);
      respond(OK);
      assertRequestsMade(5, "initially");
      assertCallbacksInvoked(5, "intially");
      cancel();
      clock.tick(5000);
      assertRequestsMade(0, "after wait with request canceled");
      assertCallbacksInvoked(0, "after wait with request canceled");
    });

    test("test many requests queued up, all but last canceled", function () {
      var respond = waitToRespond();
      var cancel = makeRequests(100);
      for (var i = 0; i < 100; i++) {
        cancel();
      }
      makeRequests(1);
      respond(OK);
      respond(OK);
      respond(OK);
      respond(OK);
      respond(OK);
      assertRequestsMade(5, "initially");
      assertCallbacksInvoked(0, "intially");
      clock.tick(1000);
      assertRequestsMade(1, "after wait with prior requests canceled");
      respond(OK);
      assertCallbacksInvoked(1, "after wait with prior requests canceled");
      clock.tick(10000);
      assertRequestsMade(0);
    });



    test("unknown error retries once", function () {
      var respond = waitToRespond();
      makeRequests(1);
      respond(google.maps.DirectionsStatus.UNKNOWN_ERROR);
      respond(google.maps.DirectionsStatus.UNKNOWN_ERROR);
      assertRequestsMade(2);
      assertCallbacksInvoked(0);
      assertErrorCallbacksInvoked(1);
    });

    test("unknown error then success", function () {
      var respond = waitToRespond();
      makeRequests(1);
      respond(google.maps.DirectionsStatus.UNKNOWN_ERROR);
      respond(google.maps.DirectionsStatus.OK);
      assertRequestsMade(2);
      assertCallbacksInvoked(1);
      assertErrorCallbacksInvoked(0);
    });

    test("other error doesn't retry", function () {
      var respond = waitToRespond();
      makeRequests(1);
      respond(ERR);
      assertRequestsMade(1);
      assertCallbacksInvoked(0);
      assertErrorCallbacksInvoked(1);
    });
  }
);