define([
    'jquery',
    'elevation',
    'elevationprovider',
    'elevationdisplay',
    'model'
  ],

  function (
    $,
    Elevation,
    ElevationProvider,
    ElevationDisplay,
    Model) {

    "use strict";

    var map,
        model,
        route,
        mergedPath,
        display,
        MIN = Elevation.MIN,
        MAX = Elevation.MAX,
        RES = Elevation.RES;

    module("Elevation tests", {
      setup: function () {
        map = {};
        model = new Model();
        route = {
          getMergedPath: function () {
            return mergedPath;
          }
        };
        display = Elevation.bind(map, model, route);
        sinon.stub(ElevationProvider, "getElevation");
        sinon.stub(display, "hide");
        sinon.stub(display, "empty");
        sinon.stub(display, "data");
        sinon.stub(display, "resize");
        sinon.stub(display, "loading");
        sinon.stub(display, "unit");
        sinon.stub(display, "mouseoutRoute");
        sinon.stub(display, "mouseoverRoute");
      },
      teardown: function () {
        ElevationProvider.getElevation.restore();
        $(window).off('resize');
      }
    });

    test("getPointsForMeters", function () {
      strictEqual(Elevation.getPointsForMeters(0), MIN, "points for 0 meters");
      strictEqual(Elevation.getPointsForMeters(1), MIN, "points for 1 meter");
      strictEqual(Elevation.getPointsForMeters(MIN * RES - 1), MIN, "points for 1 less than min * res");
      strictEqual(Elevation.getPointsForMeters(MIN * RES), MIN, "points for min * res");
      strictEqual(Elevation.getPointsForMeters(MIN * RES + 1), MIN + 1, "points for 1 over min * res");
      strictEqual(Elevation.getPointsForMeters(MIN * RES + RES), MIN + 1, "min * res + res");
      strictEqual(Elevation.getPointsForMeters(Infinity), MAX, "large number");
    });

    test("getElevation when elevation not set", function () {
      mergedPath = [];
      model.setElevation(false);
      $(route).trigger('change');
      ok(!ElevationProvider.getElevation.called);
      ok(!display.empty.called, "empty called");
    });

    test("getElevation when no route", function () {
      mergedPath = [];
      model.setElevation(true);
      ok(!ElevationProvider.getElevation.called);
      ok(display.empty.called, "empty called");
    });

    test("getElevation when 1 point", function () {
      mergedPath = [new google.maps.LatLng(0, 0)];
      model.setElevation(true);
      ok(!ElevationProvider.getElevation.called);
      ok(!display.loading.called);
      ok(display.empty.called, "empty called");
    });

    test("getElevation when 2 points", function () {
      model.setMeters(1);
      mergedPath = [new google.maps.LatLng(0, 0), new google.maps.LatLng(0, 0)];
      model.setElevation(true);
      strictEqual(ElevationProvider.getElevation.args[0][0], mergedPath);
      strictEqual(ElevationProvider.getElevation.args[0][3], MIN);
      deepEqual(display.loading.args[0], [1]);

      ElevationProvider.getElevation.args[0][1]("data", "copyright");
      deepEqual(display.data.args[0], [1, "data", mergedPath, "copyright"]);
    });

    test("getElevation triggered from route change", function () {
      model.setMeters(0);
      mergedPath = [new google.maps.LatLng(0, 0), new google.maps.LatLng(0, 0)];
      model.setElevation(true);
      ElevationProvider.getElevation.reset();
      display.loading.reset();
      $(route).trigger('change');
      strictEqual(ElevationProvider.getElevation.args[0][0], mergedPath);
      strictEqual(ElevationProvider.getElevation.args[0][3], MIN);
      ok(!display.loading.called);
    });

    test("resize triggered when elevation set", function () {
      model.setElevation(true);
      $(window).trigger('resize');
      ok(display.resize.called);
    });


    test("resize not triggered when elevation not set", function () {
      model.setElevation(false);
      $(window).trigger('resize');
      ok(!display.resize.called);
    });

    test("onunit triggered when elevation set", function () {
      model.setElevation(true);
      model.toggleUnit();
      ok(display.unit.called);
    });


    test("onunit not triggered when elevation not set", function () {
      model.setElevation(false);
      model.toggleUnit();
      ok(!display.unit.called);
    });

    test("Second request not made for same data ", function () {
      model.setMeters(1);
      mergedPath = [new google.maps.LatLng(0, 0), new google.maps.LatLng(0, 0)];
      model.setElevation(true);
      ElevationProvider.getElevation.reset();
      display.loading.reset();
      $(route).trigger('change');
      strictEqual(ElevationProvider.getElevation.args[0][0], mergedPath);
      strictEqual(ElevationProvider.getElevation.args[0][3], MIN);
      ElevationProvider.getElevation.args[0][1]("data");

      display.data.reset();
      ElevationProvider.getElevation.reset();
      model.setElevation(false);
      model.setElevation(true);


      ok(!ElevationProvider.getElevation.called, "elevation provider should not be called second time");
      deepEqual(display.data.args[0], [1, "data", mergedPath, undefined]);
      ok(!display.loading.called);
    });

    test("clear marker on hide elevation", function () {
      model.setMeters(1);
      mergedPath = [new google.maps.LatLng(0, 0), new google.maps.LatLng(0, 0)];
      model.setElevation(true);
      display.hide.reset();
      model.setElevation(false);
      ok(display.hide.called, "empty should be called");
    });

    test("clear marker on mouse out route", function () {
      model.setMeters(1);
      mergedPath = [new google.maps.LatLng(0, 0), new google.maps.LatLng(0, 0)];
      model.setElevation(true);
      display.mouseoutRoute.reset();
      $(route).trigger('mouseover');
      ok(display.mouseoutRoute.called);
    });

    test("show marker on mouse over route", function () {
      model.setMeters(1);
      mergedPath = [new google.maps.LatLng(0, 0), new google.maps.LatLng(0, 0)];
      model.setElevation(true);
      display.mouseoverRoute.reset();
      $(route).trigger('mouseover', {distance: 10});
      ok(display.mouseoverRoute.called);
    });
  }
);