define([
    'elevationdisplay',
    'model'
  ],

  function (
    ElevationDisplay,
    Model) {

    "use strict";

    var display;
    var TOLERANCE = 1e-3;
    var fuzzyEqual = function (a, b) {
      ok(Math.abs(a - b) < TOLERANCE, "Got " + a + " expected " + b + " +/-" + TOLERANCE);
    };
    var assertSubsetEquals = function (actual, expectedSubset, message) {
      var actualView = {};
      var expectedView = {};
      for (var attr in expectedSubset) {
        if (expectedSubset.hasOwnProperty(attr)) {
          actualView[attr] = actual[attr];
          expectedView[attr] = expectedSubset[attr];
        }
      }
      deepEqual(actualView, expectedView, message);
    };
    var model;

    module("Elevation display tests", {
      setup: function () {
        model = new Model();
        display = ElevationDisplay.attachTo("#qunit-fixture", model);
      }
    });

    test("empty, resize, data", function () {
      expect(0);
      display.empty();
      display.resize();
      display.data(10, [{
        location: new google.maps.LatLng(0, 0),
        elevation: 10
      }, {
        location: new google.maps.LatLng(0, 1),
        elevation: 100
      }], [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 1)
      ]);
    });

    test("loading, resize, data", function () {
      expect(0);
      display.loading(10);
      display.resize();
      display.data(10, [{
        location: new google.maps.LatLng(0, 0),
        elevation: 10
      }, {
        location: new google.maps.LatLng(0, 1),
        elevation: 100
      }], [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 1)
      ]);
    });

    test("onData, resize", function () {
      expect(0);
      display.data(10, [{
        location: new google.maps.LatLng(0, 0),
        elevation: 10
      }, {
        location: new google.maps.LatLng(0, 1),
        elevation: 100
      }], [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 1)
      ]);
      display.resize();
    });

    test("onData, unit", function () {
      expect(0);
      display.data(10, [{
        location: new google.maps.LatLng(0, 0),
        elevation: 10
      }, {
        location: new google.maps.LatLng(0, 1),
        elevation: 100
      }], [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 1)
      ]);
      display.unit();
    });

    test("onData, empty, resize", function () {
      expect(0);
      display.data(10, [{
        location: new google.maps.LatLng(0, 0),
        elevation: 10
      }, {
        location: new google.maps.LatLng(0, 1),
        elevation: 100
      }], [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 1)
      ]);
      display.empty();
      display.resize();
    });

    test("data format", function () {
      var data = [{
        location: new google.maps.LatLng(0, 0),
        elevation: 0
      }, {
        location: new google.maps.LatLng(0, 1),
        elevation: 1000
      }, {
        location: new google.maps.LatLng(0, 2),
        elevation: 500
      }];

      var path = [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 0.5),
        new google.maps.LatLng(0, 1),
        new google.maps.LatLng(0, 1.5),
        new google.maps.LatLng(0, 2)
      ];

      var result = display.formatData(2000, data, path);

      assertSubsetEquals(data[0], {
        distance: 0,
        distancemi: 0,
        distancekm: 0,
        grade: 0,
        elevation: 0,
        elevationmi: 0,
        elevationkm: 0
      }, "first data point");

      assertSubsetEquals(data[1], {
        distance: 1000,
        distancekm: 1,
        distancemi: 1 / 1.609344,
        grade: 1,
        elevation: 1000,
        elevationmi: 1000 * 3.28084,
        elevationkm: 1000
      }, "second data point");

      deepEqual(result, {
        verticalGain: {
          mi: 1000 * 3.28084,
          km: 1000
        },
        verticalLoss: {
          mi: 500 * 3.28084,
          km: 500
        },
        overallGrade: 500 / 2000
      }, "overall results");
    });

    test("mouse/touch handlers happy-path", function () {
      expect(0);
      var data = [{
        location: new google.maps.LatLng(0, 0),
        elevation: 0
      }, {
        location: new google.maps.LatLng(0, 1),
        elevation: 1000
      }, {
        location: new google.maps.LatLng(0, 2),
        elevation: 500
      }];

      var methods = display.data(2000, data, data.map(function (p) { return p.location; }));

      methods.show();
      methods.mouseOverChartAtX(100);
      methods.hide();

      display.empty();

      methods.show();
      methods.mouseOverChartAtX(100);
      methods.hide();
    });

    test("Interpolate begin, middle, end of path", function () {
      var path = [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(1, 1)
      ];
      var len = google.maps.geometry.spherical.computeLength(path);
      var data = [{
        location: new google.maps.LatLng(0, 0),
        elevation: 0
      }, {
        location: new google.maps.LatLng(1, 1),
        elevation: 1000
      }];

      model.setUnit("km");

      var methods = display.data(len, data, data.map(function (p) { return p.location; }));
      methods.show();
      var interpolated = methods.interpolateByDistance(0);
      fuzzyEqual(interpolated.lat, 0, "beginning lat");
      fuzzyEqual(interpolated.lng, 0, "beginning lng");

      interpolated = methods.interpolateByDistance(len / 1000 / 2);
      fuzzyEqual(interpolated.lat, 0.5, "middle lat");
      fuzzyEqual(interpolated.lng, 0.5, "middle lng");

      interpolated = methods.interpolateByDistance(len / 1000);
      fuzzyEqual(interpolated.lat, 1, "end lat");
      fuzzyEqual(interpolated.lng, 1, "end lng");
    });

    test("Interpolate at corner", function () {
      var path = [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 1),
        new google.maps.LatLng(1, 1)
      ];
      var len = google.maps.geometry.spherical.computeLength(path);
      var data = [{
        location: new google.maps.LatLng(0, 0),
        elevation: 0,
        idx: 0,
        dist: 0
      }, {
        location: new google.maps.LatLng(1, 1),
        elevation: 2000,
        idx: 2,
        dist: len
      }];

      model.setUnit("km");

      var methods = display.data(len, data, path);
      methods.show();
      var interpolated = methods.interpolateByDistance(0);
      fuzzyEqual(interpolated.lat, 0, "beginning lat");
      fuzzyEqual(interpolated.lng, 0, "beginning lng");

      interpolated = methods.interpolateByDistance(len / 1000 / 2);
      fuzzyEqual(interpolated.lat, 0, "middle lat");
      fuzzyEqual(interpolated.lng, 1, "middle lng");

      interpolated = methods.interpolateByDistance(len / 1000);
      fuzzyEqual(interpolated.lat, 1, "end lat");
      fuzzyEqual(interpolated.lng, 1, "end lng");
    });

    test("Interpolate length 4", function () {
      var path = [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 1),
        // middle is 0.5, 1
        new google.maps.LatLng(1, 1),
        new google.maps.LatLng(1, 0)
      ];
      var len = google.maps.geometry.spherical.computeLength(path);
      var data = [{
        location: new google.maps.LatLng(0, 0),
        elevation: 0,
        idx: 0,
        dist: 0
      }, {
        location: new google.maps.LatLng(1, 0),
        elevation: 2000,
        idx: 2,
        dist: len
      }];

      model.setUnit("km");

      var methods = display.data(len, data, path);
      methods.show();

      var interpolated = methods.interpolateByDistance(len / 1000 / 2);
      fuzzyEqual(interpolated.lat, 0.5, "middle lat");
      fuzzyEqual(interpolated.lng, 1, "middle lng");
    });

    test("Interpolate same point", function () {
      var path = [
        new google.maps.LatLng(0, 0),
        new google.maps.LatLng(0, 0)
      ];
      var len = google.maps.geometry.spherical.computeLength(path);
      var data = [{
        location: new google.maps.LatLng(0, 0),
        elevation: 0
      }, {
        location: new google.maps.LatLng(0, 0),
        elevation: 2000
      }];

      model.setUnit("km");

      var methods = display.data(len, data, path);
      methods.show();

      // all requests, inside and outside range should equal lone point
      var interpolated = methods.interpolateByDistance(10);
      fuzzyEqual(interpolated.lat, 0, "lat");
      fuzzyEqual(interpolated.lng, 0, "lng");

      interpolated = methods.interpolateByDistance(-10);
      fuzzyEqual(interpolated.lat, 0, "lat");
      fuzzyEqual(interpolated.lng, 0, "lng");
    });
  }
);