define([
    'jquery',
    'elevationprovider'
  ],

  function (
    $,
    ElevationProvider) {

    "use strict";

    var succ, error;
    var OK = google.maps.ElevationStatus.OK;
    var ERR = google.maps.ElevationStatus.ERR;
    var clock;
    var jsonpFailMethod;

    var ll = function (lat, lng) {
      return new google.maps.LatLng(lat, lng);
    };

    var sendAndReplyWith = function (points, result, response) {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, points);
      clock.tick(1);
      var args = google.maps.ElevationService.getElevationForLocations.args[0];
      args[1](result, response);
    };

    module("Elevation provider tests", {
      setup: function () {
        succ = sinon.stub();
        error = sinon.stub();
        google.maps.ElevationService.getElevationForLocations.reset();
        clock = sinon.useFakeTimers();
        window.ga = sinon.stub();
        ElevationProvider.reset();
        jsonpFailMethod = sinon.stub();
        sinon.stub($, 'getJSON', function () {
          return {
            fail: jsonpFailMethod
          };
        });
      },
      teardown: function () {
        clock.tick(1);
        var args = google.maps.ElevationService.getElevationForLocations.args[0];
        if (args && args[1]) {
          args[1](OK, OK);
        }
        if ($.getJSON.args && $.getJSON.args[0]) {
          $.getJSON.args[0][1](OK, OK);
        }
        $.getJSON.restore();
        clock.tick(10000);
        clock.restore();
      }
    });

    test("elevation request sends correct parameters", function () {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      var args = google.maps.ElevationService.getElevationForLocations.args[0];
      deepEqual(args[0].locations.length, 10);
    });

    test("elevation ERR response", function () {
      sendAndReplyWith(20, null, ERR);
      ok(!succ.called, "success called");
      ok(error.called, "error not called");
      deepEqual(window.ga.args[1], ['send', 'event', 'Elevation', 'google', ERR, 20]);
      deepEqual(window.ga.args[0], ['send', 'event', 'error', 'Elevation Service Error', ERR]);
    });

    test("elevation OK response", function () {
      sendAndReplyWith(30, "result", OK);
      ok(succ.called, "success not called");
      ok(!error.called, "error called");
      strictEqual(succ.args[0][0], "result");
      deepEqual(window.ga.args, [['send', 'event', 'Elevation', 'google', OK, 30]]);
    });


    test("requests rate-limited", function () {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      var stub = google.maps.ElevationService.getElevationForLocations;

      // only first request should have gone out
      strictEqual(stub.args.length, 1);

      // even after 1 second
      clock.tick(1000);
      strictEqual(stub.args.length, 1);

      // and after first request gets handled
      stub.args[0][1](1, OK);

      // but then 1s later, the pending request gets sent
      clock.tick(999);
      strictEqual(stub.args.length, 1);
      clock.tick(1);
      strictEqual(stub.args.length, 2);

      // and there is no 3rd request, since it got canceled by first
      stub.args[1][1](1, OK);
      clock.tick(10000);
      strictEqual(stub.args.length, 2);
    });


    test("Only last request in a batch is serviced", function () {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, 10);
      ElevationProvider.getElevation(list, succ, error, 10);
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      var stub = google.maps.ElevationService.getElevationForLocations;

      // only first request should have gone out
      strictEqual(stub.args.length, 1);
      stub.args[0][1](1, OK);
      clock.tick(5000);
      strictEqual(stub.args.length, 1);
    });


    test("Responses cached", function () {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      var stub = google.maps.ElevationService.getElevationForLocations;

      strictEqual(stub.args.length, 1);
      stub.args[0][1](1, OK);

      var succ2 = sinon.stub();
      ElevationProvider.getElevation(list, succ2, error, 10);
      ok(succ2.called);
      strictEqual(succ2.args[0][0], 1);

      strictEqual(stub.args.length, 1);
    });


    test("Responses cached", function () {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      var stub = google.maps.ElevationService.getElevationForLocations;

      strictEqual(stub.args.length, 1);
      stub.args[0][1](1, OK);

      var succ2 = sinon.stub();
      ElevationProvider.getElevation(list, succ2, error, 10);
      ok(succ2.called);
      strictEqual(succ2.args[0][0], 1);

      strictEqual(stub.args.length, 1);
    });


    test("Responses cached persistently", function () {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, 10);
      clock.tick(1);
      var stub = google.maps.ElevationService.getElevationForLocations;

      strictEqual(stub.args.length, 1);
      stub.args[0][1]([{
        location: ll(1, 2)
      }], OK);

      // reload cached state
      ElevationProvider.reload();

      var succ2 = sinon.stub();
      ElevationProvider.getElevation(list, succ2, error, 10);
      ok(succ2.called, "callback invoked");
      strictEqual(succ2.args[0][0][0].location.lat(), 1, "lat revived properly");
      strictEqual(succ2.args[0][0][0].location.lng(), 2, "lng revived properly");
      strictEqual(stub.args.length, 1, "reload did not hit elevation API again");
    });

    function shouldUseGoogleApi() {
      clock.tick(1);
      ok(google.maps.ElevationService.getElevationForLocations.called, "google elevation API not called");
      ok(!$.getJSON.called, "mapquest elevation API called");
    }

    function shouldUseMapquestApi() {
      clock.tick(1);
      ok(!google.maps.ElevationService.getElevationForLocations.called, "google elevation api called");
      ok($.getJSON.called, "mapquest elevation API not called");
    }

    test("Should use google maps when over 150mi", function () {
      var list = [ll(0, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, list, 151 * 1609);
      shouldUseGoogleApi();
    });
    test("Should use google maps when < 56 south", function () {
      var list = [ll(-56, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, list, 100);
      shouldUseGoogleApi();
    });
    test("Should use google maps when > 60 north", function () {
      var list = [ll(61, 0), ll(0, 1)];
      ElevationProvider.getElevation(list, succ, error, list, 100);
      shouldUseGoogleApi();
    });

    test("Should use mapquest when under 150mi and between -56 and 60 latitude", function () {
      var list = [ll(59, 0), ll(-55, 1)];
      ElevationProvider.getElevation(list, succ, error, list, 149 * 1609);
      clock.tick(1);
      shouldUseMapquestApi();
      strictEqual(
        $.getJSON.args[0][0],
        'http://open.mapquestapi.com/elevation/v1/profile?key=Fmjtd%7Cluub256t2d%2Cbw%3Do5-9u8n5w&shapeFormat=raw&latLngCollection=59,0,-55,1'
      );
    });

    var goodResponseData = {
      elevationProfile: [
        {height: 0},
        {height: 10}
      ],
      info: {
        copyright: {
          text: "copyright",
          imageURL: "url",
          imageAltText: "alttext"
        },
        statuscode: 0,
        messages: []
      }
    };

    test("successful mapquest response should be used", function () {
      var list = [ll(59, 0), ll(-55, 1)];
      ElevationProvider.getElevation(list, succ, error, 2, 149 * 1609);
      clock.tick(1);
      shouldUseMapquestApi();
      $.getJSON.args[0][1](goodResponseData);
      ok(succ.called, "callback invoked");
      deepEqual(succ.args[0][0], [{
        elevation: 0,
        location: list[0]
      }, {
        elevation: 10,
        location: list[1]
      }]);
      deepEqual(succ.args[0][1], {
        text: "copyright",
        imageURL: "url",
        imageAltText: "alttext"
      });
      deepEqual(window.ga.args, [['send', 'event', 'Elevation', 'mapquest', "MAPQUEST_OK", 2]]);
    });

    function testBadMapquestResponse(data) {
      var list = [ll(59, 0), ll(-55, 1)];
      ElevationProvider.getElevation(list, succ, error, list, 149 * 1609);
      clock.tick(1);
      shouldUseMapquestApi();
      $.getJSON.args[0][1](data);
      ok(!succ.called, "callback invoked");
      ok(google.maps.ElevationService.getElevationForLocations.called, "google elevation API should have been called");
    }

    test("null mapquest response thrown away", function () {
      testBadMapquestResponse(null);
    });

    test("empty mapquest response thrown away", function () {
      testBadMapquestResponse({});
    });

    test("mapquest response with no info thrown away", function () {
      testBadMapquestResponse({
        elevationProfile: [
          {height: 0},
          {height: 10}
        ]
      });
    });

    test("mapquest response with no elevation data thrown away", function () {
      testBadMapquestResponse({
        info: {
          copyright: {
            text: "copyright",
            imageURL: "url",
            imageAltText: "alttext"
          },
          statuscode: 0,
          messages: []
        }
      });
    });

    test("mapquest response with wrong # elevation points thrown away", function () {
      testBadMapquestResponse({
        elevationProfile: [
          {height: 0}
        ],
        info: {
          copyright: {
            text: "copyright",
            imageURL: "url",
            imageAltText: "alttext"
          },
          statuscode: 0,
          messages: []
        }
      });
    });

    test("mapquest response with bad status code thrown away", function () {
      testBadMapquestResponse({
        elevationProfile: [
          {height: 0},
          {height: 10}
        ],
        info: {
          copyright: {
            text: "copyright",
            imageURL: "url",
            imageAltText: "alttext"
          },
          statuscode: 1,
          messages: []
        }
      });
    });

    test("mapquest response with messages thrown away", function () {
      testBadMapquestResponse({
        elevationProfile: [
          {height: 0},
          {height: 10}
        ],
        info: {
          copyright: {
            text: "copyright",
            imageURL: "url",
            imageAltText: "alttext"
          },
          statuscode: 0,
          messages: ["bad"]
        }
      });
    });

    test("mapquest JSONP fails, call google", function () {
      var list = [ll(59, 0), ll(-55, 1)];
      ElevationProvider.getElevation(list, succ, error, list, 149 * 1609);
      clock.tick(1);
      shouldUseMapquestApi();
      jsonpFailMethod.args[0][0]();
      $.getJSON.args[0][1](goodResponseData);
      ok(!succ.called, "response should be ignored and callback not invoked");
      ok(google.maps.ElevationService.getElevationForLocations.called, "google elevation API should have been called");
    });
  }
);