define([
    'geocode'
  ],

  function (geocode) {

    "use strict";
    var geocodeStub;
    var clock;

    module("Geocode tests", {
      setup: function () {
        geocodeStub = sinon.stub();
        google.maps.Geocoder.mockImpl = geocodeStub;
        window.ga = sinon.stub();
        clock = sinon.useFakeTimers();
        geocode.reset();
      },
      teardown: function () {
        geocode.reset();
        clock.tick(100000);
        clock.restore();
      }
    });

    test("Test geocode service invoked when called", function () {
      var stub = sinon.stub();
      geocode("addr", stub);
      deepEqual(geocodeStub.args[0][0], {
        address: "addr"
      }, "geocoder service invoked with correct address");
      ok(!stub.called, "callback not invoked");
    });

    test("Test geocode service responds success - logged and callback invoked", function () {
      var bounds = {};
      var stub = sinon.stub();
      geocode("addr", stub);
      geocodeStub.args[0][1]([{
        geometry: {
          bounds: bounds
        }
      }], google.maps.GeocoderStatus.OK);
      deepEqual(stub.args[0][0], bounds, "callback invoked with args");
      deepEqual(window.ga.args, [['send', 'event', 'Geocode', 'Response', google.maps.GeocoderStatus.OK]], "result logged");
    });

    test("Test geocode service responds success but has viewport instead of bounds", function () {
      var bounds = {};
      var stub = sinon.stub();
      geocode("addr", stub);
      geocodeStub.args[0][1]([{
        geometry: {
          viewport: bounds
        }
      }], google.maps.GeocoderStatus.OK);
      deepEqual(stub.args[0][0], bounds, "callback invoked with args");
    });

    test("Test geocode service responds error - error callback invoked", function () {
      var succ = sinon.stub();
      var err = sinon.stub();
      geocode("addr", succ, err);
      geocodeStub.args[0][1]([{}], google.maps.GeocoderStatus.ERROR);
      ok(err.called, "error callback invoked");
      ok(!succ.called, "succ callback not invoked");
    });

    test("Test geocode service responds success but response malformed", function () {
      var succ = sinon.stub();
      var err = sinon.stub();
      geocode("addr", succ, err);
      geocodeStub.args[0][1]([{}], google.maps.GeocoderStatus.OK);
      ok(err.called, "error callback invoked");
      ok(!succ.called, "succ callback not invoked");
      deepEqual(window.ga.args[1], ['send', 'event', 'error', 'Geocode Response Malformed', "OK:[{}]"], "malformed response logged");
    });

    test("Test geocode service responds success but response malformed - has geom, no parts", function () {
      var succ = sinon.stub();
      var err = sinon.stub();
      geocode("addr", succ, err);
      geocodeStub.args[0][1]([{
        geometry: {}
      }], google.maps.GeocoderStatus.OK);
      ok(err.called, "error callback invoked");
      ok(!succ.called, "succ callback not invoked");
    });


    test("Test rate limiting", function () {
      var succ = sinon.stub();
      var err = sinon.stub();
      geocode("addr", succ, err);
      deepEqual(1, geocodeStub.args.length, "first call @0ms maded");
      clock.tick(499);
      geocode("addr", succ, err);
      deepEqual(1, geocodeStub.args.length, "second call @499ms blocked");
      clock.tick(2);
      deepEqual(1, geocodeStub.args.length, "still blocked @501ms");
      geocode("addr", succ, err);
      deepEqual(2, geocodeStub.args.length, "third call @501ms allowed");
    });
  }
);