/*global sinon*/
define([
    'jquery',
    'map',
    'store',
    'util',
    'url',
    'Modernizr'
  ],

  function (
    $,
    Map,
    Store,
    Util,
    URL,
    Modernizr) {

    "use strict";

    // importsf
    var LatLng = google.maps.LatLng;
    var MapTypeId = google.maps.MapTypeId;
    var events = google.maps.event;

    // constants
    var DEFAULT_LAT = 42.2297;
    var DEFAULT_LNG = -71.5181;
    var DEFAULT_ZOOM = 12;
    var DEFAULT_MAP_TYPE = MapTypeId.ROADMAP;

    // utilities
    var centerShouldBe = function (gmap, lat, lng) {
      var center = gmap.getCenter();
      ok(Util.pointsAreClose(center, new LatLng(lat, lng)),
        "center expected " + lat + ", " + lng + " was " + center);
    };
    var centerShouldBeDefault = function (gmap) {
      centerShouldBe(gmap, DEFAULT_LAT, DEFAULT_LNG);
    };
    var shouldShowPositionAt = function (gmap, lat, lng, accuracy) {
      var expected = new LatLng(lat, lng);
      strictEqual(gmap.locationDot.length, 1, "# location dots");
      ok(Util.pointsAreClose(gmap.locationDot[0].getPosition(), expected),
        "location dot expected " + lat + ", " + lng + " was " + gmap.locationDot[0].getPosition());
      strictEqual(gmap.circles.length, 1, "# circles");
      ok(Util.pointsAreClose(gmap.circles[0].getCenter(), expected),
        "location circle center expected " + lat + ", " + lng + " was " + gmap.circles[0].getCenter());
      strictEqual(gmap.circles[0].getRadius(), accuracy,
        "radius of circle should equal accuracy");
    };
    var positionShouldNotBeDisplayed = function (gmap) {
      strictEqual(gmap.locationDot.length, 0, "should be no location dots");
      strictEqual(gmap.circles.length, 0, "should be no location circles");
    };



    module("Map tests", {
      setup: function () {
        Modernizr.ioswebapp = false;
        Store.remove('lat');
        Store.clear();
        window.OTGM.geolocation.watches = [];
      },
      teardown: function () {
        URL.params = {};
      }
    });

    (function latLngTests() {
      // lat lng test utilities
      var testLatLngIgnored = function (lat, lng) {
        test(lat + " " + lng + " should be ignored", function () {
          Store.set("lat", lat);
          Store.set("lng", lng);
          var gmap = Map.attachTo("#qunit-fixture");
          centerShouldBeDefault(gmap);
        });
      };
      var testLatLngArePersisted = function (lat, lng) {
        test("center " + lat + ", " + lng + " persisted across sessions", function () {
          var gmap = Map.attachTo("#qunit-fixture");
          gmap.setCenter(new LatLng(lat, lng));
          var gmap2 = Map.attachTo("#qunit-fixture");
          centerShouldBe(gmap2, lat, lng);
        });
      };
      testLatLngIgnored(null, null);
      testLatLngIgnored(10, null);
      testLatLngIgnored(null, 10);
      testLatLngIgnored(10, "bad");
      testLatLngIgnored("bad", 10);
      testLatLngIgnored(85.1, 0);
      testLatLngIgnored(-85.1, 0);
      testLatLngIgnored(0, 181);
      testLatLngIgnored(0, -181);

      testLatLngArePersisted(10, 10);
      testLatLngArePersisted(0.5, 0.5);
      testLatLngArePersisted(0, 180);
      testLatLngArePersisted(0, -180);
      testLatLngArePersisted(85, 0);
      testLatLngArePersisted(-85, 0);
    }());


    test("geolocation overrides last location in non-iOS webapp", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      centerShouldBeDefault(gmap);
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 10
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 1, 2, 10);
    });


    test("geolocation overrides last location on first load of iOS webapp", function () {
      Modernizr.ioswebapp = true;
      window.OTGM.geolocation.lastSucc = null;
      var gmap = Map.attachTo("#qunit-fixture");
      centerShouldBeDefault(gmap);
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 10
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 1, 2, 10);
    });


    test("geolocation does not override last location on subsequent load of iOS webapp but does show position", function () {
      Modernizr.ioswebapp = true;
      Map.attachTo("#qunit-fixture").setCenter(new LatLng(3, 3));
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 3,
          longitude: 3,
          accuracy: 1
        }
      });
      window.OTGM.geolocation.lastSucc = null;
      var gmap = Map.attachTo("#qunit-fixture");
      ok(window.OTGM.geolocation.lastSucc,
        "request should have been made for current position");
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 10
        }
      });
      centerShouldBe(gmap, 3, 3);
      shouldShowPositionAt(gmap, 1, 2, 10);
    });


    test("does not ask for geolocation when nogeo parameter provided", function () {
      window.OTGM.geolocation.lastSucc = null;
      URL.params = {
        nogeo: true
      };
      Map.attachTo("#qunit-fixture");
      ok(!window.OTGM.geolocation.lastSucc,
        "request should not have been made for current position");
    });


    test("geolocation does not override last location on subsequent load of iOS webapp but does show position", function () {
      Modernizr.ioswebapp = true;
      Map.attachTo("#qunit-fixture").setCenter(new LatLng(3, 3));
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 3,
          longitude: 3,
          accuracy: 1
        }
      });
      window.OTGM.geolocation.lastSucc = null;
      var gmap = Map.attachTo("#qunit-fixture");
      ok(window.OTGM.geolocation.lastSucc,
        "request should have been made for current position");
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 10
        }
      });
      centerShouldBe(gmap, 3, 3);
      shouldShowPositionAt(gmap, 1, 2, 10);
    });


    test("geolocation does not override last location on subsequent load of iOS webapp after disabled", function () {
      Modernizr.ioswebapp = true;
      var map = Map.attachTo("#qunit-fixture");
      map.setCenter(new LatLng(3, 3));
      events.trigger(map, 'toggleFollow');
      window.OTGM.geolocation.lastSucc = null;
      var gmap = Map.attachTo("#qunit-fixture");
      ok(!window.OTGM.geolocation.lastSucc,
        "no request should have been made for current position");
      centerShouldBe(gmap, 3, 3);
      positionShouldNotBeDisplayed(gmap);
    });


    test("does ask for geolocation on subsequent load of non-iOS webapp", function () {
      Map.attachTo("#qunit-fixture").setCenter(new LatLng(3, 3));
      window.OTGM.geolocation.lastSucc = null;
      var gmap = Map.attachTo("#qunit-fixture");
      ok(window.OTGM.geolocation.lastSucc,
        "request should have been made for current position");
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 10
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 1, 2, 10);
    });


    test("does not recenter when have lat/lon zoom and current location is too inaccurate", function () {
      Map.attachTo("#qunit-fixture").setCenter(new LatLng(3, 3));
      window.OTGM.geolocation.lastSucc = null;
      var gmap = Map.attachTo("#qunit-fixture");
      ok(window.OTGM.geolocation.lastSucc,
        "request should have been made for current position");
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 3000
        }
      });
      centerShouldBe(gmap, 3, 3);
    });


    test("does recenter when don't have previous center even though current location is too inaccurate", function () {
      window.OTGM.geolocation.lastSucc = null;
      var gmap = Map.attachTo("#qunit-fixture");
      ok(window.OTGM.geolocation.lastSucc,
        "request should have been made for current position");
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 3000
        }
      });
      centerShouldBe(gmap, 1, 2);
    });


    test("does not ask for geolocation when map is embedded and centers at URL location//", function () {
      Map.attachTo("#qunit-fixture").setCenter(new LatLng(3, 3));
      window.OTGM.geolocation.lastSucc = null;
      URL.params = {
        context: "embed",
        lat: "1",
        lng: "2",
        zoom: "15"
      };
      var gmap = Map.attachTo("#qunit-fixture");
      ok(!window.OTGM.geolocation.lastSucc,
        "no request should have been made for current position");
      centerShouldBe(gmap, 1, 2);
      strictEqual(gmap.getZoom(), 15);
      positionShouldNotBeDisplayed(gmap);
    });


    /////////////////// toggle follow (find me) button
    var toggleFollow = function (gmap) {
      window.OTGM.geolocation.lastSucc = null;
      events.trigger(gmap, 'togglefollow');
    };

    test("toggleFollow event", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      toggleFollow(gmap);
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 200
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 1, 2, 200);
    });

    test("toggleFollow event - too inaccurate is OK", function () {
      Map.attachTo("#qunit-fixture").setCenter(new LatLng(3, 3));
      var gmap = Map.attachTo("#qunit-fixture");
      toggleFollow(gmap);
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 5000
        }
      });
      centerShouldBe(gmap, 1, 2);
    });

    // toggle, update -> changes center
    test("toggle then update should change center and location dot", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      toggleFollow(gmap);

      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 200
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 1, 2, 200);

      window.OTGM.geolocation.update({
        coords: {
          latitude: 3,
          longitude: 4,
          accuracy: 100
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 3, 4, 100);
    });

    // toggle, untoggle, update -> makes no change
    test("toggle, untoggle, update should make no change", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      toggleFollow(gmap);
      toggleFollow(gmap);
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 200
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 1, 2, 200);

      toggleFollow(gmap);

      // should have no effect
      window.OTGM.geolocation.update({
        coords: {
          latitude: 3,
          longitude: 4,
          accuracy: 100
        }
      });
      centerShouldBe(gmap, 1, 2);
      positionShouldNotBeDisplayed(gmap);
    });

    // toggle, move, update -> updates position, not center
    $.each(['center_changed', 'zoom_changed'], function (idx, evt) {
      test("toggle, " + evt + ", update should update should update position marker, not center", function () {
        var gmap = Map.attachTo("#qunit-fixture");
        toggleFollow(gmap);
        window.OTGM.geolocation.lastSucc({
          coords: {
            latitude: 1,
            longitude: 2,
            accuracy: 200
          }
        });
        centerShouldBe(gmap, 1, 2);
        shouldShowPositionAt(gmap, 1, 2, 200);

        events.trigger(gmap, evt);

        window.OTGM.geolocation.update({
          coords: {
            latitude: 3,
            longitude: 4,
            accuracy: 100
          }
        });
        // update after drag changes marker, not position
        centerShouldBe(gmap, 1, 2);
        shouldShowPositionAt(gmap, 3, 4, 100);
      });
    });

    // toggle, move, toggle -> still shows position, back @ center
    $.each(['zoom_changed', 'center_changed'], function (idx, evname) {
      test("toggle, move (" + evname + "), toggle should still show position, but go back to center", function () {
        var gmap = Map.attachTo("#qunit-fixture");
        toggleFollow(gmap);
        window.OTGM.geolocation.lastSucc({
          coords: {
            latitude: 1,
            longitude: 2,
            accuracy: 200
          }
        });
        centerShouldBe(gmap, 1, 2);
        shouldShowPositionAt(gmap, 1, 2, 200);

        events.trigger(gmap, evname);
        gmap.setCenter(new LatLng(3, 4));
        gmap.setZoom(5);
        centerShouldBe(gmap, 3, 4);
        strictEqual(gmap.getZoom(), 5, "zoom");

        toggleFollow(gmap);

        // toggle after drag changes position, not marker
        centerShouldBe(gmap, 1, 2);
        shouldShowPositionAt(gmap, 1, 2, 200);
        strictEqual(gmap.getZoom(), 15, "zoom");
      });
    });

    // toggle, move, toggle, toggle -> doesn't show, doesn't request
    test("toggle, move, toggle, toggle should stop showing position and cancel listeners", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      toggleFollow(gmap);
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 200
        }
      });
      centerShouldBe(gmap, 1, 2);
      shouldShowPositionAt(gmap, 1, 2, 200);

      gmap.setCenter(new LatLng(3, 4));
      toggleFollow(gmap);
      toggleFollow(gmap);

      ok(!window.OTGM.geolocation.lastSucc);
      strictEqual(window.OTGM.geolocation.watches.length, 0);
      positionShouldNotBeDisplayed(gmap);
    });

    // user denies -> next time don't ask on startup
    test("user denies position request, stop bothering them!", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      window.OTGM.geolocation.lastErr({
        // permission denied
        code: 1
      });
      positionShouldNotBeDisplayed(gmap);
      window.OTGM.geolocation.lastSucc = null;
      gmap = Map.attachTo("#qunit-fixture");
      ok(!window.OTGM.geolocation.lastSucc,
        "after denying request, don't ask again");
      toggleFollow(gmap);
      ok(window.OTGM.geolocation.lastSucc,
        "on force follow, does ask even though request denied");
      window.OTGM.geolocation.lastSucc({
        coords: {
          latitude: 1,
          longitude: 2,
          accuracy: 200
        }
      });
      window.OTGM.geolocation.lastSucc = null;
      Map.attachTo("#qunit-fixture");
      ok(window.OTGM.geolocation.lastSucc,
        "on map reload after user allowed geolocation request, do ask again");
    });

    $.each([
      { startZoom: 5, accuracy: 125, endZoom: 16 },
      { startZoom: 5, accuracy: 2000, endZoom: 12 },
      { startZoom: 5, accuracy: 2001, endZoom: 12, noPosition: true },
      { startZoom: 18, accuracy: 125, endZoom: 16 },
      { startZoom: 16, accuracy: 250, endZoom: 15 },
      { startZoom: 16, accuracy: 500, endZoom: 14 },
      { startZoom: 18, accuracy: 1000, endZoom: 13 },
      { startZoom: 18, accuracy: 2000, endZoom: 12 },
      { startZoom: 18, accuracy: 2001, endZoom: 12, noPosition: true }
    ], function (idx, scenario) {
      test("accuracy determines zoom: start " + scenario.startZoom +
        " accuracy: " + scenario.accuracy +
        " end zoom: " + scenario.endZoom +
        " no position: " + scenario.noPosition, function () {
          Store.set("zoom", scenario.startZoom);
          var gmap = Map.attachTo("#qunit-fixture");
          window.OTGM.geolocation.lastSucc({
            coords: {
              latitude: 0,
              longitude: 0,
              accuracy: scenario.accuracy
            }
          });
          centerShouldBe(gmap, 0, 0);
          strictEqual(gmap.getZoom(), scenario.endZoom, "end zoom");
          if (scenario.noPosition) {
            positionShouldNotBeDisplayed(gmap);
          } else {
            shouldShowPositionAt(gmap, 0, 0, scenario.accuracy);
          }

          window.OTGM.geolocation.update({
            coords: {
              latitude: 1,
              longitude: 1,
              accuracy: scenario.accuracy
            }
          });
          centerShouldBe(gmap, 0, 0);

          window.OTGM.geolocation.update({
            coords: {
              latitude: 1,
              longitude: 1,
              accuracy: scenario.accuracy * 2
            }
          });
          strictEqual(gmap.getZoom(), scenario.endZoom, "zoom after update");
        });
    });

    (function mapTypeTests() {
      // maptype test utilities
      var testMapTypeIgnored = function (mapType) {
        test(mapType + " should be ignored", function () {
          Store.set("maptype", mapType);
          var gmap = Map.attachTo("#qunit-fixture");
          strictEqual(gmap.getMapTypeId(), DEFAULT_MAP_TYPE);
        });
      };
      var testMapTypePersisted = function (mapType) {
        test(mapType + " persisted across sessions", function () {
          var gmap = Map.attachTo("#qunit-fixture");
          gmap.setMapTypeId(mapType);
          var gmap2 = Map.attachTo("#qunit-fixture");
          strictEqual(gmap2.getMapTypeId(), mapType);
        });
      };
      testMapTypeIgnored(null);
      testMapTypeIgnored(MapTypeId.SATELLITE);
      testMapTypeIgnored(123);

      testMapTypePersisted(MapTypeId.HYBRID);
      testMapTypePersisted(MapTypeId.TERRAIN);
      testMapTypePersisted(MapTypeId.ROADMAP);
    }());


    test("default zoom", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      strictEqual(gmap.getZoom(), DEFAULT_ZOOM);
    });
    test("non-numeric zoom ignored", function () {
      Store.set('zoom', "string");
      var gmap = Map.attachTo("#qunit-fixture");
      strictEqual(gmap.getZoom(), DEFAULT_ZOOM);
    });
    test("negative zoom ignored", function () {
      Store.set('zoom', -1);
      var gmap = Map.attachTo("#qunit-fixture");
      strictEqual(gmap.getZoom(), DEFAULT_ZOOM);
    });
    test("floating point zoom truncated", function () {
      Store.set('zoom', 1.5);
      var gmap = Map.attachTo("#qunit-fixture");
      strictEqual(gmap.getZoom(), 1);
    });
    test("zoom change persisted across sessions", function () {
      var gmap = Map.attachTo("#qunit-fixture");
      gmap.setZoom(20);
      var gmap2 = Map.attachTo("#qunit-fixture");
      strictEqual(gmap2.getZoom(), 20);
    });

    // ensure that data stored in client browsers remains compatible
    test("snapshot stored data behavior", function () {
      Store.set("lat", "1");
      Store.set("lng", "2");
      Store.set("maptype", MapTypeId.HYBRID);
      Store.set("zoom", "3");
      var gmap = Map.attachTo("#qunit-fixture");
      strictEqual(gmap.getMapTypeId(), MapTypeId.HYBRID);
      deepEqual(gmap.getCenter(), new LatLng(1, 2));
      deepEqual(gmap.getZoom(), 3);
    });

    // ensure that single click event listener set up
    test("single click is set up", function () {
      expect(1);
      var clock = sinon.useFakeTimers();
      try {
        var gmap = Map.attachTo("#qunit-fixture");
        events.addListener(gmap, 'singleclick', function () {
          ok(true, "slick click called");
        });
        events.trigger(gmap, 'click');
        clock.tick(251);
      } finally {
        clock.restore();
      }
    });
  }
);