define([
    'jquery',
    'model',
    'map',
    'route',
    'point',
    'directions',
    'util',
    'store',
    'url'
  ],

  function (
    $,
    Model,
    Map,
    Route,
    Point,
    Directions,
    util,
    Store,
    URL) {

    "use strict";

    // imports
    var LatLng = google.maps.LatLng;
    var LatLngBounds = google.maps.LatLngBounds;
    var TravelMode = google.maps.TravelMode;
    var OTGM = window.OTGM || {};

    // test state
    var gmap;
    var route;
    var model;

    // helper class
    var Multiset = function (elements) {
      for (var i = 0; i < elements.length; i += 1) {
        var element = elements[i];
        if (element.length !== 0) {
          this[element] = this[element] ? this[element] + 1 : 1;
        }
      }
    };

    // utilities
    var walk = function () {
      model.setTravelMode(TravelMode.WALKING);
    };
    var drive = function () {
      model.setTravelMode(TravelMode.DRIVING);
    };
    var bike = function () {
      model.setTravelMode(TravelMode.BICYCLING);
    };
    var assertContainSameElements = function (a, b, msg) {
      var aset = new Multiset(a);
      var bset = new Multiset(b);
      deepEqual(aset, bset, msg);
    };
    var direct = function () {
      model.setTravelMode(null);
    };
    var markersShouldBe = function (markers) {
      var markerLocations = $.map(gmap.markers, function (marker) {
        var pos = marker.getPosition();
        return [[
          util.round(pos.lat(), 3),
          util.round(pos.lng(), 3)
        ]];
      });
      assertContainSameElements(
        markerLocations,
        markers,
        "markers displayed on the map"
      );
    };
    var splittersShouldBe = function (markers) {
      var splitterLocations = $.map(gmap.splitters, function (marker) {
        var pos = marker.getPosition();
        return [[
          util.round(pos.lat(), 3),
          util.round(pos.lng(), 3)
        ]];
      });

      $.each(gmap.splitters, function (idx, marker) {
        ok(marker.options === Point.SPLITTER_MARKER,
          "splitter " + idx + " should be splitter marker");
      });
      assertContainSameElements(splitterLocations, markers,
        "splitters displayed on the map");
    };
    var checkPolylines = function (actual, expected, message) {
      var polylinesOnMap = $.map(actual, function (polyline) {
        var path = polyline.getPath();
        var result = [];
        path.forEach(function (point) {
          result.push([
            util.round(point.lat(), 3),
            util.round(point.lng(), 3)
          ]);
        });
        return [result];
      });
      assertContainSameElements(polylinesOnMap, expected,
        message);
    };
    var tempPolylinesShouldBe = function (expected) {
      checkPolylines(gmap.tempPolylines, expected,
        "temp polylines displayed on the map");
    };
    var polylinesShouldBe = function (expected) {
      checkPolylines(gmap.polylines, expected,
        "polylines displayed on the map");
    };
    var errorPolylinesShouldBe = function (expected) {
      checkPolylines(gmap.errorPolylines, expected,
        "error polylines displayed on the map");
    };
    var mouseoverPolylinesShouldBe = function (expected) {
      checkPolylines(gmap.mouseoverPolylines, expected,
        "mouseover polylines displayed on the map");
    };
    var distShouldBe = function (text) {
      strictEqual(model.getDistanceText(), text, "distance display");
    };
    var checkPointAndSegmentLinks = function (numPolys) {
      var point = route.lastPoint;
      var prev = point;
      var num = 0;
      if (point) {
        ok(!point.leaving);
        while (point) {
          if (!point.last()) {
            ok(point.marker.options === Point.FIRST_MARKER,
              "should be first marker");
          } else if (!point.next()) {
            ok(point.marker.options === Point.LAST_MARKER,
              "should be last marker");
          } else {
            ok(point.marker.options === Point.MID_MARKER,
              "marker should be mid marker");
          }
          num += 1;
          point = point.last();
          if (point) {
            // have a pair, check correct links
            ok(point.next() === prev);
            ok(point === prev.last());
            ok(point.leaving === prev.entering);
          } else {
            // first point
            ok(!prev.entering);
          }
          prev = point;
        }
        strictEqual(gmap.polylines.length, numPolys,
          "num polylines");
      }
      strictEqual(gmap.markers.length, num,
        "num markers");
    };
    var stateShouldBe = function (state) {
      state = state || {};
      var markers = state.markers || [];
      var polylines = state.polylines || [];
      var errorPolylines = state.errorPolylines || [];
      var mouseoverPolylines = state.mouseoverPolylines || polylines.concat(errorPolylines);
      var dist = state.dist;
      markersShouldBe(markers);
      polylinesShouldBe(polylines);
      checkPointAndSegmentLinks(polylines.length);
      errorPolylinesShouldBe(errorPolylines);
      mouseoverPolylinesShouldBe(mouseoverPolylines);

      if (dist) {
        distShouldBe(dist);
      }
      if (state.splitters) {
        splittersShouldBe(state.splitters);
      }
      if (state.tempPolylines) {
        tempPolylinesShouldBe(state.tempPolylines);
      }
    };
    var ll = function (lat, lng) {
      return new LatLng(lat, lng);
    };
    var mi = function (num) {
      return 1609.34 * num;
    };
    var catchGetRequests = function () {
      var last = null;
      var handle = {};
      Directions.get = function (start, end, routetype, callback, err) {
        last = {};
        last.start = [start.lat(), start.lng()];
        last.end = [end.lat(), end.lng()];
        last.routetype = routetype;
        last.callback = callback;
        last.err = err;
        return handle;
      };
      return {
        assertCalledWith: function (start, end, routetype) {
          deepEqual(last.start, start, "start");
          deepEqual(last.end, end, "end");
          strictEqual(last.routetype, routetype, "route type");
        },
        assertCanceled: function (message) {
          ok(handle.canceled, "request should have been canceled " + message);
        },
        respondWithPoints: function () {
          var points = $.makeArray(arguments);
          this.respondWithRoundTripPoints(points, points.slice().reverse());
        },
        respondWithRoundTripPoints: function (out, back) {
          var points = [], pointsrev = [];
          $.each(out, function (idx, point) {
            points.push(ll(point[0], point[1]));
          });
          if (back === null) {
            pointsrev = null;
          } else {
            $.each(back, function (idx, point) {
              pointsrev.push(ll(point[0], point[1]));
            });
          }
          if (!handle.canceled) {
            last.callback(
              ["warning"],
              "copyright",
              {
                points: points,
                distance: google.maps.geometry.spherical.computeLength(points)
              }, pointsrev === null ? Directions.PATH_ERROR :
              {
                points: pointsrev,
                distance: google.maps.geometry.spherical.computeLength(pointsrev)
              },
              handle
            );
          }
        },
        respondWithArgs: function (a, b, c, d, e) {
          if (!handle.canceled) {
            last.callback(a, b, c, d, e);
          }
        },
        assertCalled: function () {
          ok(last, "should have been called but was not");
        },
        assertNotCalled: function () {
          ok(!last, "should not have been called but was");
        },
        error: function () {
          last.err();
        }
      };
    };
    var catchGetPairRequests = function () {
      var last = null;
      var handlea = {};
      var handleb = {};
      Directions.getPair = function (start, mid, end, routetype, callback, err) {
        last = {};
        last.start = [start.lat(), start.lng()];
        last.mid = [mid.lat(), mid.lng()];
        last.end = [end.lat(), end.lng()];
        last.routetype = routetype;
        last.callback = callback;
        last.err = err;
        return [handlea, handleb];
      };
      return {
        assertCalledWith: function (start, mid, end, routetype) {
          deepEqual(last.start, start, "start");
          deepEqual(last.mid, mid, "mid");
          deepEqual(last.end, end, "end");
          strictEqual(last.routetype, routetype, "route type");
        },
        assertCanceled: function (message) {
          ok(handlea.canceled && (!handleb || handleb.canceled),
            "request should have been canceled " + message);
        },
        respondWithRoundTripPoints: function (first, second, lastrev, firstrev) {
          var a = [], ar = [], b = [], br = [];
          $.each(first, function (idx, point) {
            a.push(ll(point[0], point[1]));
          });
          $.each(second, function (idx, point) {
            b.push(ll(point[0], point[1]));
          });
          if (firstrev === null) {
            ar = null;
          } else {
            $.each(firstrev, function (idx, point) {
              ar.push(ll(point[0], point[1]));
            });
          }
          if (lastrev === null) {
            br = null;
          } else {
            $.each(lastrev, function (idx, point) {
              br.push(ll(point[0], point[1]));
            });
          }
          if (!(handlea.canceled && (!handleb || handleb.canceled))) {
            last.callback(
              ["warning"],
              "copyright",
              {
                points: a,
                distance: google.maps.geometry.spherical.computeLength(a)
              },
              {
                points: b,
                distance: google.maps.geometry.spherical.computeLength(b)
              },
              handlea, lastrev === null ? Directions.PATH_ERROR :
              {
                points: br,
                distance: google.maps.geometry.spherical.computeLength(br)
              }, firstrev === null ? Directions.PATH_ERROR :
              {
                points: ar,
                distance: google.maps.geometry.spherical.computeLength(ar)
              },
              handleb
            );
          }
        },
        respondWithPoints: function (a, b, c, d) {
          this.respondWithRoundTripPoints(
            [a, b],
            [c, d],
            [d, c],
            [b, a]
          );
        },
        respondWithArgs: function (a, b, c, d, e, f, g, h) {
          if (!(handlea.canceled && (!handleb || handleb.canceled))) {
            last.callback(a, b, c, d, e, f, g, h);
          }
        },
        assertNotCalled: function () {
          ok(!last, "should not have been called but was");
        },
        error: function () {
          last.err();
        }
      };
    };
    var originalGet;
    var originalGetPair;
    var changeEvents;
    var resetMap = function () {
      gmap = Map.attachTo("#qunit-fixture");
      model = new Model();
      route = new Route(gmap, model);
      changeEvents = 0;
      $(route).on('change', function () {
        changeEvents++;
      });
    };
    var setupTeardown = {
      setup: function () {
        Store.clear();
        OTGM.standalone = false;
        resetMap();
        originalGet = Directions.get;
        originalGetPair = Directions.getPair;
        window.ga = sinon.stub();
      },
      teardown: function () {
        Directions.get = originalGet;
        Directions.getPair = originalGetPair;
      }
    };
    var warningsShouldBe = function () {
      assertContainSameElements(model.getLegal().warnings, $.makeArray(arguments), "warnings");
    };
    var copyrightsShouldBe = function () {
      assertContainSameElements(model.getLegal().copyrights, $.makeArray(arguments), "copyrights");
    };



    /////////////////// DIRECT ROUTING TESTS ///////////////////

    module("Direct routing", setupTeardown);

    test("starting state", function () {
      stateShouldBe();
    });

    test("drop a point", function () {
      direct();
      gmap.click(0, 0);
      stateShouldBe({
        markers: [
          [0, 0]
        ],
        splitters: []
      });
      strictEqual(changeEvents, 1);
    });

    test("drop two points at same location", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 0);
      stateShouldBe({
        markers: [
          [0, 0]
        ],
        splitters: []
      });
      deepEqual(window.ga.args, []);
      strictEqual(changeEvents, 2);
    });

    test("drop 2 points, direct routing", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]]
        ],
        dist: "69.2 mi"
      });
      deepEqual(window.ga.args, [['send', 'event', 'Direction', 'direct', '(direct)']]);
    });

    test("drop 3 points, direct routing", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ],
        dist: "138.3 mi"
      });
    });

    test("move beginning point", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);
      changeEvents = 0;
      gmap.markers[0].dragStart(0, 3);
      stateShouldBe({
        markers: [
          [0, 3],
          [0, 1],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ],
        tempPolylines: [
          [[0, 3], [0, 1]]
        ]
      });

      strictEqual(changeEvents, 0);

      gmap.markers[0].dragEnd(0, 3);
      stateShouldBe({
        markers: [
          [0, 3],
          [0, 1],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 3], [0, 1]],
          [[0, 1], [0, 2]]
        ],
        dist: "207.5 mi"
      });
      strictEqual(changeEvents, 1);
    });

    test("move middle point", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);

      changeEvents = 0;
      gmap.markers[1].dragStart(0, 3);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 3],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ],
        tempPolylines: [
          [[0, 0], [0, 3]],
          [[0, 3], [0, 2]]
        ]
      });

      strictEqual(changeEvents, 0);

      gmap.markers[1].dragEnd(0, 3);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 3],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 3]],
          [[0, 3], [0, 2]]
        ]
      });
      strictEqual(changeEvents, 1);
    });

    test("move end point", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);
      gmap.markers[2].dragStart(0, 3);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 3]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ],
        tempPolylines: [
          [[0, 1], [0, 3]]
        ]
      });

      gmap.markers[2].dragEnd(0, 3);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 3]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 3]]
        ]
      });
    });

    test("undo", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      changeEvents = 0;
      route.undoLast();
      strictEqual(changeEvents, 1);
      stateShouldBe({
        markers: [
          [0, 0]
        ],
        splitters: [],
        dist: "0 mi"
      });
      gmap.click(0, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]]
        ]
      });
    });

    test("clear", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      changeEvents = 0;
      route.clear();
      strictEqual(changeEvents, 1);
      stateShouldBe({
        dist: "0 mi",
        splitters: []
      });
      gmap.click(0, 0);
      gmap.click(0, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1]]
        ]
      });
    });

    test("mouseover a segment displays its midpoint", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 2);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 2]]
        ]
      });

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseout', {
        latLng: ll(0, 1)
      });

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ]
      });
    });

    test("mouseover a road segment displays its midpoint", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 2);
      req.respondWithPoints([0, 0], [0, 1], [0, 2]);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1], [0, 2]]
        ]
      });

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseout', {
        latLng: ll(0, 1)
      });

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1], [0, 2]]
        ]
      });
    });

    test("mouseover a restored segment does not display its midpoint", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 2);
      req.respondWithPoints([0, 0], [0, 1], [0, 2]);

      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1], [0, 2]]
        ]
      });

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseout', {
        latLng: ll(0, 1)
      });

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 1], [0, 2]]
        ]
      });
    });

    test("split a line segment by dragging its midpoint - direct", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 2);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });

      // starting state
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 2]]
        ]
      });

      var splitter = gmap.splitters[0];
      google.maps.event.trigger(splitter, 'mousedown', {
        latLng: ll(0, 1)
      });
      var oldPoly = gmap.polylines[0];
      google.maps.event.trigger(oldPoly, 'mouseout', {
        latLng: ll(0, 1)
      });
      splitter.dragStart();
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [0, 1],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ],
        mouseoverPolylines: [[]],
        tempPolylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ]
      });

      splitter.drag(1, 1);
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ],
        mouseoverPolylines: [[]],
        tempPolylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ]
      });

      splitter.dragEnd(1, 1);
      google.maps.event.trigger(splitter, 'mouseup');
      google.maps.event.trigger(oldPoly, 'mouseout');
      stateShouldBe({
        dist: "195.6 mi",
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ]
      });
    });

    test("split a line segment by dragging its polyline - direct", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 2);
      var poly = gmap.mouseoverPolylines[0];
      google.maps.event.trigger(poly, 'mouseover', {
        latLng: ll(0, 1)
      });

      // starting state
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 2]]
        ]
      });
      google.maps.event.trigger(poly, 'dragstart', {
        latLng: ll(0, 1)
      });
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [0, 1],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ],
        mouseoverPolylines: [[]],
        tempPolylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ]
      });

      google.maps.event.trigger(poly, 'mousedown', {
        latLng: ll(1, 1)
      });
      google.maps.event.trigger(poly, 'mousemove', {
        latLng: ll(1, 1)
      });
      google.maps.event.trigger(poly, 'drag', {
        latLng: ll(1, 1)
      });
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ],
        mouseoverPolylines: [[]],
        tempPolylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ]
      });

      google.maps.event.trigger(poly, 'dragend', {
        latLng: ll(0, 2)
      });
      stateShouldBe({
        dist: "195.6 mi",
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ]
      });
    });

    test("split a line segment by dragging its midpoint - road", function () {
      walk();
      gmap.click(0, 0);
      var req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 0], [0, 2]);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });

      // starting state
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 2]]
        ]
      });

      req = catchGetRequests();
      var reqPair = catchGetPairRequests();

      var splitter = gmap.splitters[0];
      google.maps.event.trigger(splitter, 'mousedown', {
        latLng: ll(0, 1)
      });
      var oldPoly = gmap.polylines[0];
      google.maps.event.trigger(oldPoly, 'mouseout', {
        latLng: ll(0, 1)
      });
      splitter.dragStart();

      req.assertNotCalled();

      splitter.drag(1, 1);
      stateShouldBe({
        dist: "138.3 mi",
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        mouseoverPolylines: [[], [], []],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ],
        tempPolylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ]
      });

      reqPair.assertNotCalled();

      splitter.dragEnd(1, 1);
      google.maps.event.trigger(splitter, 'mouseup');
      google.maps.event.trigger(oldPoly, 'mouseout');

      req.assertNotCalled();
      reqPair.assertCalledWith([0, 0], [1, 1], [0, 2], TravelMode.WALKING);

      reqPair.respondWithPoints([0, 0], [1, 1], [1, 1], [0, 2]);

      stateShouldBe({
        dist: "195.6 mi",
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ]
      });
    });


    test("can't split a line segment by dragging its midpoint - restored", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 2);

      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });

      // mouseover doesn't show up
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ]
      });

      var mouseoverPoly = gmap.mouseoverPolylines[0];
      [
        'mousedown',
        'mousemove',
        'dragstart',
        'drag',
        'mousemove',
        'dragend',
        'mouseout'
      ].forEach(function (evt) {
        google.maps.event.trigger(mouseoverPoly, evt, {
          latLng: ll(0, 1)
        });
      });

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 2]
        ],
        splitters: [],
        polylines: [
          [[0, 0], [0, 2]]
        ]
      });
    });



    test("when start dragging the mouseover polyline, mouseover events stop", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 2);

      var listener = sinon.stub();
      $(route).on('mouseover', listener);

      var mouseoverPoly = gmap.mouseoverPolylines[0];
      google.maps.event.trigger(mouseoverPoly, 'dragstart', {
        latLng: ll(0, 1)
      });
      strictEqual(listener.args.length, 1, "dragstart should invoke mouseout event");
      strictEqual(listener.args[0].length, 1, "should have no args");
      google.maps.event.trigger(mouseoverPoly, 'mouseover', {
        latLng: ll(0, 1)
      });
      strictEqual(listener.args.length, 1, "mousemoves should not create additional events");
      google.maps.event.trigger(mouseoverPoly, 'drag', {
        latLng: ll(0, 1)
      });
      deepEqual(listener.args.length, 1, "drags should not create additional events");
    });

    /////////////////// ROAD ROUTING TESTS ///////////////////

    module("Road routing", setupTeardown);

    test("drop a point", function () {
      walk();
      gmap.click(0, 0);
      stateShouldBe({
        markers: [
          [0, 0]
        ]
      });
      strictEqual(changeEvents, 1);
    });

    test("drop 2 points - walking", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      changeEvents = 0;
      gmap.click(1, 1);
      strictEqual(changeEvents, 0);
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1]
        ],
        polylines: []
      });

      // google responds...
      req.assertCalledWith([0, 0], [1, 1], TravelMode.WALKING);
      req.respondWithArgs(
        ["warning"],
        "copyright",
        {
          points: [ll(0, 0), ll(1, 1)],
          distance: mi(2)
        },
        {
          points: [ll(1, 1), ll(0, 0)],
          distance: mi(2)
        },
        {}
      );

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1]
        ],
        polylines: [
          [[0, 0], [1, 1]]
        ],
        dist: "2 mi"
      });
      strictEqual(changeEvents, 1);
    });

    test("change route segment while request pending cancels previous", function () {
      walk();
      var req1 = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1]
        ],
        polylines: []
      });
      req1.assertCalledWith([0, 0], [1, 1], TravelMode.WALKING);

      // before google response, move the endpoint from 1, 1 to 2, 2
      var req2 = catchGetRequests();

      gmap.markers[1].dragStart();
      req1.assertCanceled("after drag starts");

      gmap.markers[1].dragEnd(2, 2);
      stateShouldBe({
        markers: [
          [0, 0],
          [2, 2]
        ],
        polylines: []
      });
      req2.assertCalledWith([0, 0], [2, 2], TravelMode.WALKING);

      changeEvents = 0;
      // req1 comes back and is ignored since canceled
      req1.respondWithArgs(
        ["warning"],
        'copyright',
        {
          points: [ll(0, 0), ll(1, 1)],
          distance: mi(2)
        }
      );
      stateShouldBe({
        markers: [
          [0, 0],
          [2, 2]
        ],
        polylines: []
      });

      // req2 comes back and is handled
      req2.respondWithArgs(
        ["warning"],
        'copyright',
        {
          points: [ll(0, 0), ll(2, 2)],
          distance: mi(2)
        },
        {
          points: [ll(2, 2), ll(0, 0)],
          distance: mi(2)
        },
        {}
      );
      stateShouldBe({
        markers: [
          [0, 0],
          [2, 2]
        ],
        polylines: [
          [[0, 0], [2, 2]]
        ]
      });
      strictEqual(changeEvents, 1);
    });


    test("test move start and end point before first response comes back", function () {
      drive();
      var req1 = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1]
        ],
        polylines: []
      });
      req1.assertCalledWith([0, 0], [1, 1], TravelMode.DRIVING);

      walk();
      changeEvents = 0;

      // before google response, move the endpoint from 1, 1 to 2, 2 and start from 0, 0 to 1, 1
      var req2 = catchGetRequests();
      gmap.markers[1].dragTo(2, 2);
      var req3 = catchGetRequests();
      gmap.markers[0].dragTo(1, 1);
      stateShouldBe({
        markers: [
          [1, 1],
          [2, 2]
        ],
        polylines: []
      });
      req1.assertCanceled();
      req2.assertCanceled();
      req3.assertCalledWith([1, 1], [2, 2], TravelMode.DRIVING);

      // req1 comes back and is ignored since canceled
      req3.respondWithArgs(
        ["warning"],
        'copyright',
        {
          points: [ll(1, 1), ll(2, 2)],
          distance: mi(3)
        },
        {
          points: [ll(2, 2), ll(1, 1)],
          distance: mi(2)
        },
        {}
      );
      stateShouldBe({
        markers: [
          [1, 1],
          [2, 2]
        ],
        polylines: [
          [[1, 1], [2, 2]]
        ]
      });
      strictEqual(changeEvents, 1);
    });

    test("drag a point", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 1);

      // google responds...
      req.respondWithArgs(
        ["warning"],
        'copyright',
        {
          points: [ll(0, 0), ll(1, 1)],
          distance: mi(2)
        },
        {
          points: [ll(1, 1), ll(0, 0)],
          distance: mi(2)
        },
        {}
      );

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1]
        ],
        polylines: [
          [[0, 0], [1, 1]]
        ],
        dist: "2 mi"
      });

      // drag doesn't make request
      req = catchGetRequests();
      gmap.markers[1].set('position', ll(2, 2));
      gmap.markers[1].dragStart();
      gmap.markers[1].drag();
      req.assertNotCalled();

      stateShouldBe({
        markers: [
          [0, 0],
          [2, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]]
        ],
        tempPolylines: [
          [[0, 0], [2, 2]]
        ]
      });

      // drag END DOES make request
      req = catchGetRequests();
      gmap.markers[1].dragEnd(2, 2);
      req.assertCalledWith([0, 0], [2, 2], TravelMode.WALKING);
      req.respondWithArgs(
        ["warning"],
        'copyright',
        {
          points: [ll(0, 0), ll(2, 2)],
          distance: mi(2)
        },
        {
          points: [ll(2, 2), ll(0, 0)],
          distance: mi(2)
        },
        {}
      );

      stateShouldBe({
        markers: [
          [0, 0],
          [2, 2]
        ],
        polylines: [
          [[0, 0], [2, 2]]
        ],
        tempPolylines: []
      });
    });


    test("drop 2 points but start and end point come back different - biking", function () {
      bike();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1]
        ],
        polylines: []
      });

      // google responds...
      req.assertCalledWith([0, 0], [1, 1], TravelMode.BICYCLING);
      req.respondWithArgs(
        ["warning"],
        'copyright',
        {
          points: [ll(0.1, 0.2), ll(1.1, 1.2)],
          distance: mi(2)
        },
        {
          points: [ll(1.1, 1.2), ll(0.1, 0.2)],
          distance: mi(2)
        },
        {}
      );

      stateShouldBe({
        markers: [
          [0.1, 0.2],
          [1.1, 1.2]
        ],
        polylines: [
          [[0.1, 0.2], [1.1, 1.2]]
        ],
        dist: "2 mi"
      });
    });


    test("direct then routing but route comes back with different start changes direct", function () {
      direct();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1]
        ],
        polylines: [
          [[0, 0], [1, 1]]
        ]
      });

      req.assertNotCalled();

      walk();

      gmap.click(2, 2);
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1],
          [2, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]]
        ]
      });

      // google responds...
      req.respondWithArgs(
        ["warning"],
        'copyright',
        {
          points: [ll(1.1, 1.2), ll(2, 2)],
          distance: mi(2)
        },
        {
          points: [ll(2, 2), ll(1.1, 1.2)],
          distance: mi(2)
        },
        {}
      );

      stateShouldBe({
        markers: [
          [0, 0],
          [1.1, 1.2],
          [2, 2]
        ],
        polylines: [
          [[0, 0], [1.1, 1.2]],
          [[1.1, 1.2], [2, 2]]
        ],
        splitters: []
      });
    });

    test("polyline updated on correction from directions recalculates distance", function () {
      direct();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      walk();
      gmap.click(0, 2);
      distShouldBe("69.2 mi");

      // google responds...
      req.respondWithArgs(
        ["warning"],
        'copyright',
        {
          // response makes direct polyline 0 miles long
          points: [ll(0, 0), ll(0, 2)],
          // and has a distance of 2 miles
          distance: mi(2)
        },
        {
          points: [ll(0, 2), ll(0, 0)],
          distance: mi(2)
        },
        {}
      );

      distShouldBe("2 mi");
    });

    test("updating an adjacent polyline while direction request pending", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 1);
      direct();
      gmap.click(2, 2);

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1],
          [2, 2]
        ],
        polylines: [
          [[1, 1], [2, 2]]
        ]
      });

      var endOfPoly = gmap.markers[2];
      endOfPoly.dragStart();
      endOfPoly.drag(3, 3);

      // google responds...
      req.respondWithPoints([0, 0], [1.1, 1.1]);

      stateShouldBe({
        markers: [
          [0, 0],
          [1.1, 1.1],
          [3, 3]
        ],
        polylines: [
          [[0, 0], [1.1, 1.1]],
          [[1.1, 1.1], [3, 3]]
        ]
      });

      endOfPoly.dragEnd();

      stateShouldBe({
        markers: [
          [0, 0],
          [1.1, 1.1],
          [3, 3]
        ],
        polylines: [
          [[0, 0], [1.1, 1.1]],
          [[1.1, 1.1], [3, 3]]
        ]
      });
    });

    test("two adjacent follow road segments, second comes back with different start than end of first", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      bike();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      secondReq.respondWithPoints([0, 1.1], [0, 2]);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1.1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [0, 1], [0, 1.1]],
          [[0, 1.1], [0, 2]]
        ],
        dist: "138.3 mi"
      });
    });

    test("two adjacent follow road segments, first comes back with different end than start of second", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      bike();
      gmap.click(0, 2);

      secondReq.respondWithPoints([0, 1], [0, 2]);
      firstReq.respondWithPoints([0, 0], [0, 0.9]);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 0.9],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [0, 0.9]],
          [[0, 0.9], [0, 1], [0, 2]]
        ],
        dist: "138.3 mi"
      });
    });


    test("two adjacent follow road segments, second comes back with different start than end of first WHILE DRAGGING FIRST BY START", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      bike();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      gmap.markers[0].dragStart();
      gmap.markers[0].drag(0, 0.1);
      secondReq.respondWithPoints([0, 1.1], [0, 2]);
      tempPolylinesShouldBe([[[0, 0.1], [0, 1.1]]]);
    });

    test("drag midpoint shows correct polyline", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      firstReq.respondWithPoints([0, 0.1], [0, 1]);
      var secondReq = catchGetRequests();
      bike();
      gmap.click(0, 2);
      secondReq.respondWithPoints([0, 1.1], [0, 2.1]);

      gmap.markers[1].drag(1, 1);
      gmap.markers[1].dragStart();
      tempPolylinesShouldBe([[[0, 0.1], [1, 1]], [[1, 1], [0, 2.1]]]);
    });


    test("two adjacent follow road segments, first comes back with different end than start of second WHILE DRAGGING SECOND BY END", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      bike();
      gmap.click(0, 2);

      secondReq.respondWithPoints([0, 1], [0, 2]);
      gmap.markers[2].dragStart();
      gmap.markers[2].drag(0, 3);
      firstReq.respondWithPoints([0, 0], [0, 0.9]);
      tempPolylinesShouldBe([[[0, 0.9], [0, 3]]]);
    });


    test("clicking a point adds a segment to it", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      google.maps.event.trigger(gmap.markers[0], 'singleclick');
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 0]]
        ]
      });
    });


    test("clicking a segment adds a segment to it", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mousedown', {
        latLng: ll(0, 0.5)
      });
      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'singleclick', {
        latLng: ll(0, 0.5)
      });
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 0.5]
        ],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 0.5]]
        ],
        splitters: []
      });
    });


    test("clicking a segment doesn't return to it when restored", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);
      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'singleclick', {
        latLng: ll(0, 0.5)
      });
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ]
      });
    });


    test("double clicking a point adds a segment to it - walk", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      walk();
      changeEvents = 0;
      var req = catchGetRequests();
      google.maps.event.trigger(gmap.markers[0], 'singleclick');
      strictEqual(changeEvents, 0);
      req.respondWithPoints([0, 1], [0, 0]);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 0]]
        ]
      });
      strictEqual(changeEvents, 2);
    });


    test("clicking last point does not add path to it", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      google.maps.event.trigger(gmap.markers[1], 'singleclick');
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ]
      });
    });


    test("clicking middle point does add path to it", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);
      changeEvents = 0;
      google.maps.event.trigger(gmap.markers[1], 'singleclick');
      strictEqual(changeEvents, 1);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 2],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]],
          [[0, 2], [0, 1]]
        ]
      });
    });


    test("dragging a midpoint of two adjacent road segments batches requests together", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      secondReq.respondWithPoints([0, 1], [0, 2]);

      changeEvents = 0;
      var dragReq = catchGetRequests();
      var batchDragReq = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);
      dragReq.assertNotCalled();
      batchDragReq.assertCalledWith([0, 0], [1, 1], [0, 2], TravelMode.WALKING);

      batchDragReq.respondWithArgs(
        ['warning'],
        'copyright',
        {
          points: [ll(0, 0), ll(1, 1)],
          distance: mi(1)
        },
        {
          points: [ll(1, 1), ll(0, 2)],
          distance: mi(2)
        },
        {},
        {
          points: [ll(0, 2), ll(1, 1)],
          distance: mi(2)
        },
        {
          points: [ll(1, 1), ll(0, 0)],
          distance: mi(1)
        },
        {}
      );

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ],
        dist: '3 mi'
      });
      strictEqual(changeEvents, 1);
    });



    test("dragging a midpoint of two adjacent road segments but end of 1st not same as start of 2nd", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      secondReq.respondWithPoints([0, 1], [0, 2]);

      var dragReq = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);

      dragReq.respondWithPoints([0, 0], [1, 1], [1.1, 1.1], [0, 2]);

      stateShouldBe({
        markers: [
          [0, 0],
          [1.1, 1.1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [1, 1], [1.1, 1.1]],
          [[1.1, 1.1], [0, 2]]
        ]
      });
    });



    test("optimized getPair request is canceled by subsequent segment drag", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      secondReq.respondWithPoints([0, 1], [0, 2]);

      var dragReq = catchGetPairRequests();
      var secondDragReq = catchGetRequests();
      gmap.markers[1].dragTo(1, 1);
      gmap.markers[2].dragTo(1, 2);

      secondDragReq.respondWithPoints([1, 1], [1, 2]);
      dragReq.respondWithPoints([0, 0], [1, 1], [1, 1], [0, 2]);

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1],
          [1, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [1, 2]]
        ]
      });
    });

    test("optimized getPair request is canceled by previous segment drag", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      secondReq.respondWithPoints([0, 1], [0, 2]);

      var dragReq = catchGetPairRequests();
      var secondDragReq = catchGetRequests();
      gmap.markers[1].dragTo(1, 1);
      gmap.markers[0].dragTo(1, 0);

      secondDragReq.respondWithPoints([1, 0], [1, 1]);
      dragReq.respondWithPoints([0, 0], [1, 1], [1, 1], [1, 2]);

      stateShouldBe({
        markers: [
          [1, 0],
          [1, 1],
          [1, 2]
        ],
        polylines: [
          [[1, 0], [1, 1]],
          [[1, 1], [1, 2]]
        ]
      });
    });

    test("regular drag canceled by optimized getPair request", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      secondReq.respondWithPoints([0, 1], [0, 2]);

      var dragReq = catchGetPairRequests();
      var secondDragReq = catchGetRequests();
      gmap.markers[0].dragTo(1, 0);
      gmap.markers[1].dragTo(1, 1);

      dragReq.respondWithPoints([0, 0], [1, 1], [1, 1], [1, 2]);
      secondDragReq.respondWithPoints([1, 0], [1, 1]);

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1],
          [1, 2]
        ],
        polylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [1, 2]]
        ]
      });
    });



    test("dragging a midpoint of two adjacent road segments where travel modes not the same DOES NOT BATCH", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      drive();
      var secondReq = catchGetRequests();
      gmap.click(0, 2);

      firstReq.respondWithPoints([0, 0], [0, 1]);
      secondReq.respondWithPoints([0, 1], [0, 2]);

      var dragReq = catchGetRequests();
      var dragPairReq = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);

      dragPairReq.assertNotCalled();
      dragReq.assertCalled();
    });



    test("Error on 1st segment then 2nd segment adjusts error endpoint", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      firstReq.error();
      var secondReq = catchGetRequests();
      gmap.click(0, 2);
      secondReq.respondWithPoints([1, 1], [0, 2]);
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        polylines: [
          [[1, 1], [0, 2]]
        ],
        errorPolylines: [
          [[0, 0], [1, 1]]
        ],
        tempPolylines: []
      });
    });


    test("dragging a midpoint of two adjacent road segments where one is error DOES NOT BATCH", function () {
      walk();
      var firstReq = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var secondReq = catchGetRequests();
      gmap.click(0, 2);

      firstReq.error();
      secondReq.respondWithPoints([0, 1], [0, 2]);

      var dragReq = catchGetRequests();
      var dragPairReq = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);

      dragPairReq.assertNotCalled();
      dragReq.assertCalled();
    });


    test("error received on direction response and cleared on next request", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);

      // google responds with error, error segment should be displayed
      req.error();
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [],
        errorPolylines: [
          [[0, 0], [0, 1]]
        ],
        tempPolylines: []
      });

      // drag, error should persist
      var nextReq = catchGetRequests();
      gmap.markers[0].dragStart();
      gmap.markers[0].dragEnd();
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [],
        errorPolylines: [
          [[0, 0], [0, 1]]
        ],
        tempPolylines: []
      });

      // respond success - error should clear
      nextReq.respondWithPoints([0, 0], [0, 1]);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("error received on batched getPair request", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var req2 = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 0], [0, 1]);
      req2.respondWithPoints([0, 1], [0, 2]);

      var dragReq = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);

      dragReq.error();
      stateShouldBe({
        markers: [
          [0, 0],
          [1, 1],
          [0, 2]
        ],
        polylines: [],
        errorPolylines: [
          [[0, 0], [1, 1]],
          [[1, 1], [0, 2]]
        ],
        tempPolylines: []
      });
    });



    test("undo road segments", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var req2 = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 0], [0, 1]);
      req2.respondWithPoints([0, 1], [0, 2]);

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ],
        tempPolylines: []
      });

      route.undoLast();
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ],
        tempPolylines: []
      });

      route.undoLast();
      stateShouldBe({
        markers: [
          [0, 0]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: []
      });

      route.undoLast();
      stateShouldBe({});
    });



    test("undo road segments before responses come in", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var req2 = catchGetRequests();
      gmap.click(0, 2);

      route.clear();

      req.respondWithPoints([0, 0], [0, 1]);
      req2.respondWithPoints([0, 1], [0, 2]);
      stateShouldBe({});
    });


    test("reverse when no route", function () {
      route.reverse();
      stateShouldBe({});
      deepEqual(window.ga.args, [['send', 'event', 'Reverse', 'Reverse']]);
    });


    test("reverse when one point", function () {
      gmap.click(0, 0);
      route.reverse();
      stateShouldBe({
        markers: [
          [0, 0]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse when two points", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      route.reverse();
      stateShouldBe({
        markers: [
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 1], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse when three points", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);
      route.reverse();
      stateShouldBe({
        markers: [
          [0, 2],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [0, 1]],
          [[0, 1], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse routed segment", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithRoundTripPoints(
        [[0, 0], [1, 1], [0, 1]],
        [[0, 1], [0, 0]]
      );
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [0, 1]
          ],
          polylines: [
            [[0, 0], [1, 1], [0, 1]]
          ],
          errorPolylines: [],
          tempPolylines: [],
          dist: "167 mi"
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 1],
            [0, 0]
          ],
          polylines: [
            [[0, 1], [0, 0]]
          ],
          errorPolylines: [],
          tempPolylines: [],
          dist: "69.2 mi"
        });
      };
      assertForward();
      route.reverse();
      assertBackward();
      route.reverse();
      assertForward();
      route.reverse();
      assertBackward();
    });

    test("reverse routed segment after restore", function () {
      OTGM.standalone = true;
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithRoundTripPoints(
        [[0, 0], [1, 1], [0, 1]],
        [[0, 1], [0, 0]]
      );
      var saveRestore = function () {
        OTGM.standalone = true;
        resetMap();
        route.restore();
      };
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [0, 1]
          ],
          polylines: [
            [[0, 0], [1, 1], [0, 1]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 1],
            [0, 0]
          ],
          polylines: [
            [[0, 1], [0, 0]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      saveRestore();
      assertForward();
      route.reverse();
      assertBackward();

      saveRestore();
      assertBackward();
      route.reverse();
      assertForward();
    });


    test("reverse routed segment with error in reverse direction", function () {
      OTGM.standalone = true;
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithRoundTripPoints(
        [[0, 0], [1, 1], [0, 1]],
        null
      );
      var saveRestore = function () {
        OTGM.standalone = true;
        resetMap();
        route.restore();
      };
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [0, 1]
          ],
          polylines: [
            [[0, 0], [1, 1], [0, 1]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 1],
            [0, 0]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 1], [0, 0]]
          ],
          tempPolylines: []
        });
      };
      assertForward();
      route.reverse();
      assertBackward();
      route.reverse();
      assertForward();

      saveRestore();
      assertForward();
      route.reverse();
      assertBackward();

      saveRestore();
      assertBackward();
      route.reverse();
      assertForward();
    });


    test("reverse while routing segment with error in reverse direction", function () {
      OTGM.standalone = true;
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      route.reverse();
      req.respondWithRoundTripPoints(
        [[0, 0], [1, 1], [0, 1]],
        null
      );
      var saveRestore = function () {
        OTGM.standalone = true;
        resetMap();
        route.restore();
      };
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [0, 1]
          ],
          polylines: [
            [[0, 0], [1, 1], [0, 1]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 1],
            [0, 0]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 1], [0, 0]]
          ],
          tempPolylines: []
        });
      };
      assertBackward();
      saveRestore();
      assertBackward();
      route.reverse();
      assertForward();

      saveRestore();
      assertForward();
      route.reverse();
      assertBackward();
    });


    test("error in both directions, saves/restores error", function () {
      OTGM.standalone = true;
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.error();
      var saveRestore = function () {
        OTGM.standalone = true;
        resetMap();
        route.restore();
      };
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [0, 1]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 0], [0, 1]]
          ],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 1],
            [0, 0]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 1], [0, 0]]
          ],
          tempPolylines: []
        });
      };
      assertForward();
      saveRestore();
      assertForward();
      route.reverse();
      assertBackward();

      saveRestore();
      assertBackward();
      route.reverse();
      assertForward();
    });


    test("reverse routed getPair segment with error in reverse direction", function () {
      OTGM.standalone = true;
      gmap.click(0, 0);
      var req = catchGetRequests();
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 1]);
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 2]);
      req = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);
      req.respondWithRoundTripPoints(
        [[0, 0], [1, 1]],
        [[1, 1], [0, 2]],
        null,
        null
      );
      var saveRestore = function () {
        OTGM.standalone = true;
        resetMap();
        route.restore();
      };
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [1, 1],
            [0, 2]
          ],
          polylines: [
            [[0, 0], [1, 1]],
            [[1, 1], [0, 2]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 2],
            [1, 1],
            [0, 0]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 2], [1, 1]],
            [[1, 1], [0, 0]]
          ],
          tempPolylines: []
        });
      };
      saveRestore();
      assertForward();
      route.reverse();
      assertBackward();

      saveRestore();
      assertBackward();
      route.reverse();
      assertForward();
    });


    test("reverse while routing getPair segment with error in reverse direction", function () {
      OTGM.standalone = true;
      gmap.click(0, 0);
      var req = catchGetRequests();
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 1]);
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 2]);
      req = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);
      route.reverse();
      req.respondWithRoundTripPoints(
        [[0, 0], [1, 1]],
        [[1, 1], [0, 2]],
        null,
        null
      );
      var saveRestore = function () {
        OTGM.standalone = true;
        resetMap();
        route.restore();
      };
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [1, 1],
            [0, 2]
          ],
          polylines: [
            [[0, 0], [1, 1]],
            [[1, 1], [0, 2]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 2],
            [0, 1],
            [0, 0]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 2], [0, 1]],
            [[0, 1], [0, 0]]
          ],
          tempPolylines: []
        });
      };
      assertBackward();
      saveRestore();
      assertBackward();
      route.reverse();
      assertForward();

      saveRestore();
      assertForward();
      route.reverse();
      // TODO
      // assertBackward();
    });


    test("getPair error displays error on save/restore/reverse", function () {
      OTGM.standalone = true;
      gmap.click(0, 0);
      var req = catchGetRequests();
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 1]);
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 2]);
      req = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);
      req.error();
      var saveRestore = function () {
        OTGM.standalone = true;
        resetMap();
        route.restore();
      };
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [1, 1],
            [0, 2]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 0], [1, 1]],
            [[1, 1], [0, 2]]
          ],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 2],
            [1, 1],
            [0, 0]
          ],
          polylines: [],
          errorPolylines: [
            [[0, 2], [1, 1]],
            [[1, 1], [0, 0]]
          ],
          tempPolylines: []
        });
      };
      assertForward();
      saveRestore();
      assertForward();
      route.reverse();
      assertBackward();

      saveRestore();
      assertBackward();
      route.reverse();
      assertForward();
    });

    test("reverse routed segment - ends don't line up", function () {
      direct();
      gmap.click(-1, -1);
      gmap.click(0, 0);
      walk();
      var req = catchGetRequests();
      gmap.click(0, 1);
      req.respondWithRoundTripPoints(
        [[0, 0], [0, 1]],
        [[0, 1.1], [0, 0.1]]
      );
      direct();
      gmap.click(0, 2);
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [-1, -1],
            [0, 0],
            [0, 1],
            [0, 2]
          ],
          polylines: [
            [[-1, -1], [0, 0]],
            [[0, 0], [0, 1]],
            [[0, 1], [0, 2]]
          ],
          errorPolylines: [],
          tempPolylines: [],
          dist: "236.2 mi"
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [0, 2],
            [0, 1.1],
            [0, 0.1],
            [-1, -1]
          ],
          polylines: [
            [[0, 2], [0, 1.1]],
            [[0, 1.1], [0, 0.1]],
            [[0, 0.1], [-1, -1]]
          ],
          errorPolylines: [],
          tempPolylines: [],
          dist: "234.3 mi"
        });
      };
      assertForward();
      route.reverse();
      assertBackward();
      route.reverse();
      assertForward();
      route.reverse();
      assertBackward();
    });

    test("reverse adjacent routed segments where ends don't line up", function () {
      walk();
      gmap.click(0, 0);
      var req = catchGetRequests();
      gmap.click(0, 1);
      req.respondWithRoundTripPoints(
        [[0, 0], [-1, 1]],
        [[1, 1], [1, 0]]
      );
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithRoundTripPoints(
        [[1, 1], [0, 2]],
        [[1, 2], [-1, 1]]
      );
      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [1, 1],
            [0, 2]
          ],
          polylines: [
            [[0, 0], [-1, 1], [1, 1]],
            [[1, 1], [0, 2]]
          ],
          errorPolylines: [],
          tempPolylines: [],
          dist: "334 mi"
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [1, 2],
            [1, 0],
            [1, 1]
          ],
          polylines: [
            [[1, 2], [-1, 1]],
            [[1, 1], [1, 0]]
          ],
          errorPolylines: [],
          tempPolylines: [],
          dist: "223.8 mi"
        });
      };
      assertForward();
      route.reverse();
      assertBackward();
      route.reverse();
      assertForward();
      route.reverse();
      assertBackward();
    });


    test("optimized get pair request with different out/back then reverse", function () {
      walk();
      gmap.click(0, 0);
      var req = catchGetRequests();
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 1]);
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 2]);
      req = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);
      req.respondWithRoundTripPoints(
        [[0, 0], [0, 1]],
        [[0, 1], [0, 2]],
        [[1, 2], [1, 1]],
        [[1, 1], [1, 0]]
      );

      var assertForward = function () {
        stateShouldBe({
          markers: [
            [0, 0],
            [0, 1],
            [0, 2]
          ],
          polylines: [
            [[0, 0], [0, 1]],
            [[0, 1], [0, 2]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      var assertBackward = function () {
        stateShouldBe({
          markers: [
            [1, 2],
            [1, 1],
            [1, 0]
          ],
          polylines: [
            [[1, 2], [1, 1]],
            [[1, 1], [1, 0]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
      };
      assertForward();
      route.reverse();
      assertBackward();
      route.reverse();
      assertForward();
      route.reverse();
      assertBackward();
    });


    test("reverse two routed segments", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 0.5], [0, 1]);
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 1.5], [0, 2]);

      route.reverse();
      stateShouldBe({
        markers: [
          [0, 2],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [0, 1.5], [0, 1]],
          [[0, 1], [0, 0.5], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse error segment", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.error();

      route.reverse();
      stateShouldBe({
        markers: [
          [0, 1],
          [0, 0]
        ],
        polylines: [],
        errorPolylines: [
          [[0, 1], [0, 0]]
        ],
        tempPolylines: []
      });
    });


    test("undo after reversing", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);

      route.reverse();
      route.undoLast();
      stateShouldBe({
        markers: [
          [0, 1]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("clear after reversing", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);

      route.reverse();
      route.clear();
      stateShouldBe({});
    });

    test("add direct point after reversing", function () {
      direct();
      gmap.click(0, 1);
      gmap.click(0, 2);

      route.reverse();
      gmap.click(0, 0);
      stateShouldBe({
        markers: [
          [0, 2],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [0, 1]],
          [[0, 1], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    test("add routed segment after reversing", function () {
      direct();
      gmap.click(0, 1);
      gmap.click(0, 2);

      route.reverse();
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      req.assertCalledWith([0, 1], [0, 0], TravelMode.WALKING);
      req.respondWithPoints([0, 1], [0, 0.5], [0, 0]);
      stateShouldBe({
        markers: [
          [0, 2],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [0, 1]],
          [[0, 1], [0, 0.5], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("double reverse", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 1);
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 1.5], [0, 2]);

      route.reverse();
      route.reverse();
      stateShouldBe({
        markers: [
          [0, 1],
          [0, 2]
        ],
        polylines: [
          [[0, 1], [0, 1.5], [0, 2]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse when direction request pending", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);

      route.reverse();
      stateShouldBe({
        markers: [
          [0, 1],
          [0, 0]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: []
      });
      req.respondWithPoints([0, 0], [0, 0.5], [0, 1]);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 1], [0, 0.5], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse twice when direction request pending", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);

      route.reverse();
      route.reverse();
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: []
      });
      req.respondWithPoints([0, 0], [0, 0.5], [0, 1]);
      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 0.5], [0, 1]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse when optimized direction request pair is pending", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 0.5], [0, 1]);
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 1.5], [0, 2]);
      req = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);

      route.reverse();
      stateShouldBe({
        markers: [
          [0, 2],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [0, 1.5], [0, 1]],
          [[0, 1], [0, 0.5], [0, 0]]
        ],
        errorPolylines: []
      });
      req.respondWithRoundTripPoints(
        [[0.1, 0], [0, 1]],
        [[0, 1], [0.2, 2]],
        [[0.2, 2], [0, 1]],
        [[0, 1], [0.1, 0]]
      );
      stateShouldBe({
        markers: [
          [0.2, 2],
          [0, 1],
          [0.1, 0]
        ],
        polylines: [
          [[0.2, 2], [0, 1]],
          [[0, 1], [0.1, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("split a route after reversed", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      route.reverse();
      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0.5, 0)
      });
      var splitter = gmap.splitters[0];
      splitter.dragStart(0.5, 0);

      stateShouldBe({
        markers: [
          [0, 1],
          [0.5, 0],
          [0, 0]
        ],
        polylines: [
          [[0, 1], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: [
          [[0, 1], [0.5, 0]],
          [[0.5, 0], [0, 0]]
        ],
        mouseoverPolylines: []
      });
      splitter.dragEnd(0.5, 0);

      stateShouldBe({
        markers: [
          [0, 1],
          [0.5, 0],
          [0, 0]
        ],
        polylines: [
          [[0, 1], [0.5, 0]],
          [[0.5, 0], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });


    test("reverse while splitting", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0.5, 0)
      });
      var splitter = gmap.splitters[0];
      splitter.dragStart(0.5, 0);
      route.reverse();

      stateShouldBe({
        markers: [
          [0, 1],
          [0.5, 0],
          [0, 0]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ],
        errorPolylines: [],
        tempPolylines: [
          [[0, 1], [0.5, 0]],
          [[0.5, 0], [0, 0]]
        ],
        mouseoverPolylines: [
          [[0, 1], [0.5, 0]],
          [[0.5, 0], [0, 0]]
        ]
      });
      splitter.dragEnd(0.5, 0);

      stateShouldBe({
        markers: [
          [0, 1],
          [0.5, 0],
          [0, 0]
        ],
        polylines: [
          [[0, 1], [0.5, 0]],
          [[0.5, 0], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    test("drag midpoint then reverse", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);
      gmap.markers[1].dragTo(1, 1);
      route.reverse();

      stateShouldBe({
        markers: [
          [0, 2],
          [1, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [1, 1]],
          [[1, 1], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        splitters: []
      });
    });



    test("reverse then drag midpoint", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 1]);
      req = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithPoints([0, 1], [0, 2]);

      route.reverse();
      stateShouldBe({
        markers: [
          [0, 2],
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [0, 1]],
          [[0, 1], [0, 0]]
        ],
        errorPolylines: []
      });

      var dragReq = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);
      dragReq.assertCalledWith([0, 2], [1, 1], [0, 0], TravelMode.WALKING);

      dragReq.respondWithArgs(
        ['warning'],
        'copyright',
        {
          points: [ll(0, 2), ll(1, 1)],
          distance: mi(1)
        },
        {
          points: [ll(1, 1), ll(0, 0)],
          distance: mi(2)
        }, {}, {}, {}, {}
      );

      stateShouldBe({
        markers: [
          [0, 2],
          [1, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 2], [1, 1]],
          [[1, 1], [0, 0]]
        ],
        errorPolylines: []
      });
    });

    //////////////////// display copyrights and warnings

    module("Warning and copyright display", setupTeardown);

    test("Warning and copyright start empty", function () {
      warningsShouldBe();
      copyrightsShouldBe();
    });

    test("Warning and copyrights still empty when segment has none", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithArgs(
        [],   // empty warning array
        null, // empty copyright,
        {
          points: [ll(0, 0), ll(0, 1)],
          distance: mi(1)
        }, {}, {}
      );

      warningsShouldBe();
      copyrightsShouldBe();
    });

    test("Warning and copyrights still empty when segment has empty strings", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithArgs(
        [""],   // empty warning array
        "",     // empty copyright,
        {
          points: [ll(0, 0), ll(0, 1)],
          distance: mi(1)
        }, {}, {}
      );

      warningsShouldBe();
      copyrightsShouldBe();
    });


    test("Warning and copyrights displayed for single segment", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithArgs(
        ["warning"],
        "copyright",
        {
          points: [ll(0, 0), ll(0, 1)],
          distance: mi(1)
        }, {}, {}
      );

      warningsShouldBe("warning");
      copyrightsShouldBe("copyright");
    });


    test("Multiple warnings and copyrights displayed for multiple segments", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var req2 = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithArgs(
        ["warning 1", "warning 2"],
        "copyright 1",
        {
          points: [ll(0, 0), ll(0, 1)],
          distance: mi(1)
        }, {}, {}
      );
      req2.respondWithArgs(
        ["warning 3"],
        "copyright 2",
        {
          points: [ll(0, 1), ll(0, 2)],
          distance: mi(1)
        }, {}, {}
      );

      warningsShouldBe("warning 1", "warning 2", "warning 3");
      copyrightsShouldBe("copyright 1", "copyright 2");
    });


    test("Identical warnings/copyrights only included once", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var req2 = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithArgs(
        ["warning", "warning"],
        "copyright",
        {
          points: [ll(0, 0), ll(0, 1)],
          distance: mi(1)
        }, {}, {}
      );
      req2.respondWithArgs(
        ["warning"],
        "copyright",
        {
          points: [ll(0, 1), ll(0, 2)],
          distance: mi(1)
        }, {}, {}
      );

      warningsShouldBe("warning");
      copyrightsShouldBe("copyright");
    });


    test("Warnings and copyrights updated on route change and getPair", function () {
      var arbitraryPath = { points: [ll(0, 0), ll(0, 1)], distance: mi(1) };
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      var req2 = catchGetRequests();
      gmap.click(0, 2);
      req.respondWithArgs(
        ["warning 1", "warning 2"],
        "copyright 1",
        arbitraryPath,
        arbitraryPath,
        {}
      );
      req2.respondWithArgs(
        ["warning 3"],
        "copyright 2",
        arbitraryPath,
        arbitraryPath,
        {}
      );

      var dragReq = catchGetPairRequests();
      gmap.markers[1].dragTo(1, 1);
      dragReq.respondWithArgs(
        ["warning", "warning"],
        "copyright",
        arbitraryPath,
        arbitraryPath,
        {},
        arbitraryPath,
        arbitraryPath,
        {}
      );

      warningsShouldBe("warning");
      copyrightsShouldBe("copyright");
    });

    //////////////////// mouseover route

    module("Mouseover routes", setupTeardown);

    test("mouseover direct segment sends event", function () {
      var listener = sinon.stub();
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);
      $(route).on('mouseover', listener);
      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 0)
      });
      deepEqual(listener.args[0][1], {
        position: [0, 0],
        distance: 0
      });

      google.maps.event.trigger(gmap.mouseoverPolylines[1], 'mouseover', {
        latLng: ll(0, 1)
      });
      deepEqual(listener.args[1][1], {
        position: [0, 1],
        distance: 111319.49079327357
      });

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });
      deepEqual(listener.args[2][1], {
        position: [0, 1],
        distance: 111319.49079327357
      });
    });

    test("mouseover road segment sends event", function () {
      var listener = sinon.stub();
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [1, 0], [0, 1]);
      $(route).on('mouseover', listener);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 0)
      });
      deepEqual(listener.args[0][1], {
        position: [0, 0],
        distance: 0
      });

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });
      deepEqual(listener.args[1][1], {
        position: [0, 1],
        distance: 268745.02790168556
      });
    });

    test("mouseover restored segment sends event", function () {
      var listener = sinon.stub();
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [1, 0], [0, 1]);

      OTGM.standalone = true;
      gmap.click(0, 0);

      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      $(route).on('mouseover', listener);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 0)
      });
      deepEqual(listener.args[0][1], {
        position: [0, 0],
        distance: 0
      });

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });
      deepEqual(listener.args[1][1], {
        position: [0, 1],
        distance: 268745.02790168556
      });
    });


    test("mousemove", function () {
      var listener = sinon.stub();
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);

      $(route).on('mouseover', listener);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mousemove', {
        latLng: ll(0, 0)
      });
      deepEqual(listener.args[0][1], {
        position: [0, 0],
        distance: 0
      });
    });


    test("mouseout", function () {
      var listener = sinon.stub();
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);

      $(route).on('mouseover', listener);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mousemove', {
        latLng: ll(0, 0)
      });
      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseout', {
        latLng: ll(0, 0)
      });
      strictEqual(listener.args.length, 2);
      ok(!listener.args[1][1]);
    });


    test("mouseover/out out of order on segments", function () {
      var listener = sinon.stub();
      direct();
      gmap.click(0, 0);
      gmap.click(0, 1);
      gmap.click(0, 2);

      $(route).on('mouseover', listener);

      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseover', {
        latLng: ll(0, 1)
      });
      google.maps.event.trigger(gmap.mouseoverPolylines[1], 'mouseover', {
        latLng: ll(0, 1)
      });
      google.maps.event.trigger(gmap.mouseoverPolylines[0], 'mouseout', {
        latLng: ll(0, 1)
      });
      strictEqual(listener.args.length, 2);
      deepEqual(listener.args[0][1], {distance: 111319.49079327357, position: [0, 1]});
      deepEqual(listener.args[1][1], {distance: 111319.49079327357, position: [0, 1]});
    });


    //////////////////// serialize/deserialize

    module("Serialize/deserialize routes", setupTeardown);

    test("serialize/deserialize a point", function () {
      direct();
      OTGM.standalone = true;
      gmap.click(0, 0);

      resetMap();
      route.restore();

      stateShouldBe({
        markers: [
          [0, 0]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: [],
        dist: "0 mi"
      });

      gmap.click(0, 1);

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        dist: "69.2 mi"
      });
    });

    test("serialize/deserialize 2 points - ensure correct order", function () {
      direct();
      OTGM.standalone = true;
      gmap.click(0, 0);
      gmap.click(0, 1);

      resetMap();
      route.restore();

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        dist: "69.2 mi"
      });


      gmap.click(0, 2);

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1],
          [0, 2]
        ],
        polylines: [
          [[0, 0], [0, 1]],
          [[0, 1], [0, 2]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        dist: "138.3 mi"
      });
    });

    test("serialize/deserialize after reverse", function () {
      direct();
      OTGM.standalone = true;
      gmap.click(0, 0);
      gmap.click(0, 1);
      route.reverse();

      resetMap();
      route.restore();

      stateShouldBe({
        markers: [
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 1], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        dist: "69.2 mi"
      });
    });


    test("serialize/deserialize segment with error", function () {
      walk();
      OTGM.standalone = true;
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.error();

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [],
        errorPolylines: [
          [[0, 0], [0, 1]]
        ],
        tempPolylines: [],
        dist: "69.2 mi"
      });

      resetMap();
      route.restore();

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [],
        errorPolylines: [
          [[0, 0], [0, 1]]
        ],
        tempPolylines: [],
        dist: "69.2 mi"
      });
    });


    test("serialize/deserialize in-progress segment", function () {
      walk();
      OTGM.standalone = true;
      catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: [],
        dist: "0 mi"
      });

      var req = catchGetRequests();
      resetMap();
      route.restore();
      req.assertNotCalled();

      stateShouldBe({
        markers: [
          [0, 0]
        ]
      });
    });
    test("serialize/deserialize in-progress segment", function () {
      walk();
      OTGM.standalone = true;
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.respondWithPoints([0, 0], [0, 1]);
      gmap.markers[1].dragStart();
      gmap.markers[1].drag(2, 2);

      req = catchGetRequests();
      resetMap();
      route.restore();
      req.assertNotCalled();

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [0, 1]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        dist: "69.2 mi"
      });
    });

    test("serialize/deserialize with nothing saved", function () {
      direct();
      OTGM.standalone = true;

      resetMap();
      route.restore();

      stateShouldBe({
        markers: [],
        polylines: [],
        errorPolylines: [],
        tempPolylines: [],
        dist: "0 mi"
      });
    });

    $.each([
      {func: walk, mode: TravelMode.WALKING},
      {func: drive, mode: TravelMode.DRIVING},
      {func: bike, mode: TravelMode.BICYCLING}
    ], function (idx, elem) {
      test("serialize/deserialize a " + elem.mode + " segment", function () {
        elem.func();
        var req = catchGetRequests();
        gmap.click(0, 0);
        gmap.click(0, 2);
        req.respondWithPoints([0, 0], [0, 2]);
        req = catchGetRequests();
        OTGM.standalone = true;
        gmap.markers[1].dragTo(0, 1);
        req.respondWithArgs(
          ["warning"],
          "copyright",
          {
            points: [ll(0, 0), ll(1, 1), ll(0, 1)],
            distance: mi(2)
          },
          {},
          {}
        );

        req = catchGetRequests();
        resetMap();
        route.restore();
        req.assertNotCalled();

        stateShouldBe({
          markers: [
            [0, 0],
            [0, 1]
          ],
          polylines: [
            [[0, 0], [1, 1], [0, 1]]
          ],
          errorPolylines: [],
          tempPolylines: [],
          dist: "2 mi"
        });

        req = catchGetRequests();
        gmap.markers[1].dragTo(2, 2);
        req.assertCalledWith([0, 0], [2, 2], elem.mode);
      });
    });


    test("serialize/deserialize garbage", function () {
      direct();
      OTGM.standalone = true;
      Store.set("lastroute", "garbage");
      resetMap();
      route.restore();

      stateShouldBe({
        markers: [],
        polylines: [],
        errorPolylines: [],
        tempPolylines: [],
        dist: "0 mi"
      });
    });

    test("snapshot stored data behavior", function () {
      Store.set("lastroute", '[{"e":{"d":3218.68,"w":["warning"],"c":"copyright","p":"??_ibE_ibE~hbE?","t":"WALKING"},"p":"?_ibE"},{"p":"??"}]');

      var req = catchGetRequests();
      resetMap();
      route.restore();
      route.reverse();
      stateShouldBe({
        markers: [
          [0, 1],
          [0, 0]
        ],
        polylines: [
          [[0, 1], [1, 1], [0, 0]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        dist: "2 mi"
      });
      route.reverse();
      req.assertNotCalled();

      stateShouldBe({
        markers: [
          [0, 0],
          [0, 1]
        ],
        polylines: [
          [[0, 0], [1, 1], [0, 1]]
        ],
        errorPolylines: [],
        tempPolylines: [],
        dist: "2 mi"
      });
      deepEqual(model.getLegal(), {
        warnings: ["warning"],
        copyrights: ["copyright"]
      });

      req = catchGetRequests();
      gmap.markers[1].dragTo(2, 2);
      req.assertCalledWith([0, 0], [2, 2], TravelMode.WALKING);
    });

    test("snapshot stored data behavior after adding route reversal", function () {
      Store.set("lastroute", '[{"e":{"d":2005.9779020170754,"p":"_xwwFtanbMuAjEaClH??}BsAeCeB??Ob@}BlHeC`IwDpLyArE??gC_BgCaB??_@nAc@tAkDpKYdA]|@a@rAoA`Es@bA??A?A?A?A?C?A?A?A?K@C@A@C?A@GDA@A?A@A@A@ABABA@ABABA@AB?B_@h@ABCBC@A?C@C?u@@???D?FAFAHCH_BfFa@`A","dr":2443,"pr":"}xywF`}pbMAB_@pAUd@??bCzA??zBzAdC`B??~BqHnBgGpAeEzCmJzA{ErC}IxAsEvDqLdCaI|BmHNc@`CsH??fCbBzBzAzBxA??fCaI??{ByA??q@tB","w":[],"c":"Map data ©2013 Google","t":"DRIVING","e":false,"er":false},"p":"}xywF`}pbM"},{"p":"_xwwFtanbM"}]');
      resetMap();
      route.restore();
      route.update();
      distShouldBe("1.2 mi");
      route.reverse();
      distShouldBe("1.5 mi");
    });

    test("merge & serialze empty", function () {
      direct();
      OTGM.standalone = true;
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [],
        polylines: [],
        errorPolylines: [],
        tempPolylines: [],
        dist: "0 mi"
      });
    });

    test("merge & serialze one point direct", function () {
      direct();
      gmap.click(0, 0);
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [
          [0, 0]
        ],
        polylines: [],
        errorPolylines: [],
        tempPolylines: [],
        dist: "0 mi"
      });
    });

    test("merge & serialze two points direct", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(1, 2);
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 2]
        ],
        polylines: [
          [[0, 0], [1, 2]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    test("merge & serialze three points direct", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(1, 2);
      gmap.click(3, 4);
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [
          [0, 0],
          [3, 4]
        ],
        polylines: [
          [[0, 0], [1, 2], [3, 4]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    test("merge & serialze follow road segment", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 2);
      req.respondWithPoints([0, 0], [0, 2], [1, 2]);
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 2]
        ],
        polylines: [
          [[0, 0], [0, 2], [1, 2]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    test("merge & serialze 2 follow road segments", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 2);
      req.respondWithPoints([0, 0], [0, 2], [1, 2]);
      req = catchGetRequests();
      gmap.click(3, 4);
      req.respondWithPoints([1, 2], [0, 0], [3, 4]);
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [
          [0, 0],
          [3, 4]
        ],
        polylines: [
          [[0, 0], [0, 2], [1, 2], [0, 0], [3, 4]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    test("merge & serialze follow road segment with error", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 2);
      req.error();
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 2]
        ],
        polylines: [
          [[0, 0], [1, 2]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    test("merge & serialze follow road segment with drag in progress", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 2);
      req.respondWithPoints([0, 0], [1, 2]);
      gmap.markers[1].dragStart();
      gmap.markers[1].drag(9, 9);
      var merged = route.getMergedEncodedPath();
      resetMap();
      route.restoreMerged(merged);

      stateShouldBe({
        markers: [
          [0, 0],
          [1, 2]
        ],
        polylines: [
          [[0, 0], [1, 2]]
        ],
        errorPolylines: [],
        tempPolylines: []
      });
    });

    asyncTest("build & parse URL with route", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(1, 2);
      req.respondWithPoints([0, 0], [1, 1], [1, 2]);

      URL.build(gmap, model, route, function (link) {
        URL.params = URL.parse(link);
        resetMap();
        route.restoreMerged(URL.params.r);
        stateShouldBe({
          markers: [
            [0, 0],
            [1, 2]
          ],
          polylines: [
            [[0, 0], [1, 1], [1, 2]]
          ],
          errorPolylines: [],
          tempPolylines: []
        });
        URL.params = {};
        start();
      });
    });

    ///////////////////////////////////////////////////////////////
    module("route bounds", setupTeardown);
    var assertBoundsAre = function (north, east, south, west) {
      var bounds = new LatLngBounds(new LatLng(south, west), new LatLng(north, east));
      ok(route.getBounds().equals(bounds), "expected " + bounds + " got " + route.getBounds());
    };
    test("bounds of route with no points", function () {
      ok(!route.getBounds(), "should return null when no bounds");
    });
    test("bounds of route with 1 point", function () {
      direct();
      gmap.click(0, 0);
      assertBoundsAre(0, 0, 0, 0);
    });
    test("bounds of route with 2 points", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(1, 1);
      assertBoundsAre(1, 1, 0, 0);
    });
    test("bounds of route with 3 points", function () {
      direct();
      gmap.click(0, 0);
      gmap.click(1, 1);
      gmap.click(-1, -1);
      assertBoundsAre(1, 1, -1, -1);
    });
    test("bounds of route with 2 points but interesting segment between", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0.5, 0.5);
      req.respondWithPoints([0, 0], [1, 1], [0.5, 0.5]);
      assertBoundsAre(1, 1, 0, 0);
    });

    ///////////////////////////////////////////////////////////////
    module("route behavior when shared", {
      setup: function () {
        URL.params.context = "embed";
        setupTeardown.setup();
      },
      teardown: function () {
        setupTeardown.teardown();
        URL.params.context = null;
      }
    });

    test("shared route can't be modified", function () {
      walk();
      var req = catchGetRequests();
      gmap.click(0, 0);
      gmap.click(0, 1);
      req.assertNotCalled();

      stateShouldBe({
        markers: [],
        polylines: [],
        errorPolylines: [],
        tempPolylines: [],
        dist: "0 mi"
      });
    });
  }
);