define([
    "jquery",
    "url",
    "store",
    "map",
    "model",
    "route",
    "util"
  ],

  function (
    $,
    URL,
    Store,
    Map,
    Model,
    Route,
    Util) {

    "use strict";
    var MapTypeId = google.maps.MapTypeId;
    var LatLng = google.maps.LatLng;
    var map, model, route;
    var iframeUrl = function (url) {
      return $(url).attr("src");
    };

    module("URL tests", {
      setup: function () {
        Store.clear();
        map = Map.attachTo("#qunit-fixture");
        model = new Model();
        route = new Route(map, model);
        window.ga = sinon.stub();
      },
      teardown: function () {
        Store.clear();
        URL.params = {};
      }
    });

    test("iframe", function () {
      var url = "http://onthegomap.com/";
      var expected = '<iframe width="425" height="350" frameborder="0" scrolling="no" marginheight="0" marginwidth="0" src="http://onthegomap.com/"></iframe><br/><small><a href="http://onthegomap.com" style="color:#0000FF;text-align:left" target="_blank">On The Go Map</a></small>';
      strictEqual(URL.iframe(url), expected, "constructed iframe code");
    });

    test("iframe encodes apersands", function () {
      var url = "http://onthegomap.com/?a=b&c=d";
      var expected = '<iframe width="425" height="350" frameborder="0" scrolling="no" marginheight="0" marginwidth="0" src="http://onthegomap.com/?a=b&amp;c=d"></iframe><br/><small><a href="http://onthegomap.com" style="color:#0000FF;text-align:left" target="_blank">On The Go Map</a></small>';
      strictEqual(URL.iframe(url), expected, "constructed iframe code");
    });

    asyncTest("share context=share", function () {
      URL.build(map, model, route, function (link) {
        var parsed = URL.parse(link);
        strictEqual(parsed.context, "share", "parsed context=?");
        start();
      });
    });

    asyncTest("embed context=embed", function () {
      URL.build(map, model, route, function (link, iframe) {
        var src = iframeUrl(iframe);
        var parsed = URL.parse(src);
        strictEqual(parsed.context, "embed", "parsed context=?");
        start();
      });
    });

    test("parse function corner cases", function () {
      deepEqual(
        URL.parse(""),
        {},
        "parse empty"
      );
      deepEqual(
        URL.parse("http://onthegomap.com/"),
        {},
        "parse no search"
      );
      deepEqual(
        URL.parse("http://onthegomap.com/?"),
        {},
        "parse empty search"
      );
      deepEqual(
        URL.parse("http://onthegomap.com/?="),
        {"": ""},
        "parse 1-empty-param search"
      );
      deepEqual(
        URL.parse("http://onthegomap.com/?a="),
        {a: ""},
        "parse 1-param-empty-value search"
      );
      deepEqual(
        URL.parse("http://onthegomap.com/?a"),
        {a: "true"},
        "parse key with no value"
      );
      deepEqual(
        URL.parse("http://onthegomap.com/?a=b"),
        {a: "b"},
        "parse 1-param well-formed search"
      );
    });

    test("encode", function () {
      deepEqual(
        URL.encode({a: true}),
        "a",
        "encode key with true value"
      );
      deepEqual(
        URL.encode({a: false}),
        "",
        "encode key with false value"
      );
      deepEqual(
        URL.encode({a: [1, 2, 3]}),
        "a%5B%5D=1,2,3",
        "encode array"
      );
      deepEqual(
        URL.encode({a: []}),
        "",
        "encode array"
      );
      deepEqual(
        URL.encode({a: "a b c"}),
        "a=a+b+c",
        "encode array"
      );
    });

    test("build/parse query", function () {
      var runTest = function (obj) {
        var encoded = URL.encode(obj);
        var decoded = URL.decode(encoded);
        deepEqual(decoded, obj, "intermediate: " + encoded);
      };
      runTest({a: "b"});
      runTest({a: "&"});
      runTest({a: ";!32"});
      runTest({a: "!' ()*"});
      runTest({a: ["1", "2", "3,4"]});
    });

    var testParseIframeAndLink = function (attr, expected, descr) {
      URL.build(map, model, route, function (link, iframe) {
        ok(link.indexOf("http://onthegomap.com") === 0, "should start with onthegomap.com, but instead was: " + link);
        var parsedLink = URL.parse(link);
        var parsedIframe = URL.parse(iframeUrl(iframe));
        strictEqual(parsedIframe[attr], expected, descr + " (from iframe)");
        strictEqual(parsedLink[attr], expected, descr + " (from link)");
        start();
      });
    };

    $.each([MapTypeId.ROADMAP, MapTypeId.HYBRID, MapTypeId.TERRAIN], function (idx, value) {
      asyncTest("build & parse URL with " + value + " map type", function () {
        map.setMapTypeId(value);
        testParseIframeAndLink("m", value, "extracted map type");
      });
    });

    asyncTest("build & parse URL with lat", function () {
      map.setCenter(new LatLng(1, 2));
      testParseIframeAndLink("lat", "1", "extracted lat");
    });

    asyncTest("build & parse URL with lng", function () {
      map.setCenter(new LatLng(1, 2));
      testParseIframeAndLink("lng", "2", "extracted lng");
    });

    asyncTest("lat/lng rounded", function () {
      map.setCenter(new LatLng(1.1111111111, 2.2222222222));
      testParseIframeAndLink("lat", "1.111111", "extracted lat");
    });

    asyncTest("lat/lng rounded", function () {
      map.setCenter(new LatLng(1.1111111111, 2.2222222222));
      testParseIframeAndLink("lng", "2.222222", "extracted lng");
    });

    asyncTest("build & parse URL with zoom", function () {
      map.setZoom(9);
      testParseIframeAndLink("zoom", "9", "extracted zoom");
    });

    asyncTest("build & parse URL with unit", function () {
      model.setUnit("km");
      testParseIframeAndLink("u", "km", "extracted unit");
    });

    asyncTest("build & parse URL with unit", function () {
      model.setUnit("mi");
      testParseIframeAndLink("u", "mi", "extracted unit");
    });

    asyncTest("build & parse URL with default", function () {
      testParseIframeAndLink("u", "mi", "extracted unit");
    });

    asyncTest("build & parse URL with bike layer", function () {
      model.setBikeLayer(true);
      testParseIframeAndLink("b", "true", "extracted bike layer");
    });

    asyncTest("build & parse URL with bike layer OFF", function () {
      model.setBikeLayer(false);
      testParseIframeAndLink("b", undefined, "extracted bike layer");
    });

    asyncTest("build & parse URL with warn & copyright", function () {
      model.setLegal({
        warnings: ["warning 1", "warning 2"],
        copyrights: ["copyright 1", "copyright 2"]
      });

      URL.build(map, model, route, function (link) {
        URL.params = URL.parse(link);
        var newModel = new Model();
        deepEqual(newModel.getLegal(), {
          warnings: ["warning 1", "warning 2"],
          copyrights: ["copyright 1", "copyright 2"]
        }, "extracted legal");
        start();
      });
    });

    asyncTest("build & parse URL with bike layer OFF", function () {
      model.setMeters(100.1);
      deepEqual(window.ga.args, []);
      URL.build(map, model, route, function (link) {
        URL.params = URL.parse(link);
        var newModel = new Model();
        strictEqual(newModel.getMeters(), 100);
        setTimeout(function () {
          deepEqual(window.ga.args[0], ['send', 'event', 'Share Link', 'Create Link', '0.1 mi', 0]);
          ok(/send,timing,CPU,Encode Route,[0-9]*,0.1 mi/.test(window.ga.args[1].toString()),
            "GA args 1: " + window.ga.args[1].toString());
          start();
        }, 0);
      });
    });

    test("Snapshot released URL API - a path", function () {
      var sample = "http://onthegomap.com/?m=t&u=mi&w%5B%5D=Walking+directions+are+in+beta.++++Use+caution+%E2%80%93+This+route+may+be+missing+sidewalks+or+pedestrian+paths.&c%5B%5D=Map+data+%C2%A92013+Google&d=8849&context=share&r=u%7Bw%60GbxmsLa@mAg@uA%5Ds@aD_Gc@%7B@e@iA%5B_AOy@OqBe@eIEsACm@?w@@%5B@m@ToEHcAFoAD%7D@?y@?a@CgAGkASoBGy@Iy@YeCaAwGc@oCOeAKu@i@kDSqAQq@gAoDe@mAuAwCw@kBwCiHa@_Ak@yAcAqCi@_Bi@oAa@qAcAqCsAyDcBwE%7BBmGe@sAcCyGy@%7DBg@sAm@iBo@%7DBeAmEOg@O_@%5Bm@%5Di@%5Dk@_@e@WYi@q@QS%5Dc@%7D@aASSc@Yo@Yu@YaAa@YOOI%7D@%7B@%5D_@m@k@%5BYq@g@s@c@s@c@%7B@k@mBkAm@a@_@Q_Aq@YYY%5DeA%7BA_DwE_@i@_@e@%5B%5BmEgEgBmBw@%7D@IIMOGIKMKUEIQi@IYIc@Ga@CWIaA?SGk@KmAM_ASmA_@oCiAoFe@yBu@sDm@aDoBgJeAcFUyA%5B_Ci@mECMAOCQCMa@%7BDQ%7B@Oo@Qe@eAiC%5Dw@g@gAEIQ%5BMW%5Bm@%7B@yAc@s@_AoAs@w@s@s@mAg@i@Q%7B@%5Bk@Wu@c@i@i@g@m@eAsAc@c@m@e@wAy@c@g@_AgAg@i@e@_@g@YgAi@qCiA%7B@Y%7B@%5BgAg@yBs@%7D@U_@Ke@Ks@Qa@K%7B@Wc@QwBoAMKmBaB_BkAUUIGGKOUIQIO%5DcA%7B@uCe@oAu@mBkAoCeBaEe@aA%7B@kBGSGSCGCIGWCQAICUCSGs@?E?q@@Y@Q@g@??w@Bw@BcBFSB??A%5BEi@ImCMyJEkBGaACYEc@u@uHEi@AQCs@@m@DaAJwB??OA??CXA@cCU@K??e@C??@o@Gc@?A";
      URL.params = URL.parse(sample);
      map = Map.attachTo("#qunit-fixture");
      model = new Model();
      route = new Route(map, model);
      route.restoreMerged(URL.params.r);
      deepEqual(model.getLegal(), {
        "copyrights": [
          "Map data ©2013 Google"
        ],
        "warnings": [
          "Walking directions are in beta.    Use caution – This route may be missing sidewalks or pedestrian paths."
        ]
      });
      strictEqual(model.getMeters(), 8849);
      strictEqual(model.getUnit(), "mi");
      strictEqual(map.getMapTypeId(), google.maps.MapTypeId.TERRAIN);
      var path = route.lastPoint.entering.polyline.getPath();
      strictEqual(path.getLength(), 240);
      strictEqual(Math.round(google.maps.geometry.spherical.computeLength(path)), 8860);
    });

    test("Snapshot released URL API - a path with new encoding", function () {
      var sample = "http://onthegomap.com/?m=r&u=mi&w%5B%5D=Walking+directions+are+in+beta.++++Use+caution+%E2%80%93+This+route+may+be+missing+sidewalks+or+pedestrian+paths.&c%5B%5D=Map+data+%C2%A92013+Google%2C+Sanborn,Map+data+%C2%A92013+Google&d=42134&context=share&r2=y~tX8limqDc1W1Y1c1i3_2Y3W2Y2q1Y2m2Og1UW2c4WBY2Y5_1e6g1u3e2a5y4uAa7mBY2g4Gw1Go3c1cAAY43Y4j1YD0y2Cq4W2gBW6YPw2Y7c1k2s2u4_8mHk3o7g1m2w5_CiFcX1k1g3m1_3c2k6W1e2u2a4o4g5o2s2a1Qe5s2g1Qg4e4k2Y2WAY7W1IW2o1q1u1c9aCg7a7i6_6a1g2K_1Cw1Um6Y2_8m3gBw8YTg3cFY1y5Iy1Gm1Ic1i4g8Y2i3W3k4W2m2e3i3q5s2Y3y1g1g1k3Y4a1a1k1c1u2w1i4w4c1W1m3a2o4g2u3s1e2e1w3q1w6i2W3g1u3m2y3k3W4g3i1a2w2w6y2_5o5oAY3k5W1o2Mw35i4Te4h3uH~2q9Tc2v3W7Pu1Bc22e16SGc1eAsJg3y6y5wIi4uF_1W6a1W4i2s8e1g5u4YMq3kIa1o4W7qVY4aHe6gTW1a4Aa2Eg3c1iI6k18g5W1WDq1_QUeJ4qF4s21w36c9UY7k6ic18a39y3Fs2n5sJd3aEBY44W8As3Mk5_2wKGm4W4_Ti1qBe1m7w6YY1Ue3k4sCu8yMo7uHsI_Y1iAsIq1Y4q1c5g2q7m3uCq3aBa1q4W1i4u1yAIq4IY5MqCIWM0q5DeEPmHJe94o3Im7Gi3Iw5k1q8W1Y7m5yGc1m2Y2q2qIWDc3k2k1g1o1c1W6m5m6W7_4m6e8uByEmKi7w7c7a7W6y5s5c5w2w2g2o2g1u1w4W8k3g5e5e7c5s7W8kBy2c4_2s4c2w2W2u1q3o2W6W3k7_2i3s1u3q2o3c3q5y6Ok1Gm1Iy3Ao5g1s9Kg2m1W4g3WAOi1y2u3c1w1G4WBgGu1i2m1q2q1e4m1q4m1q2SW1W2i1y5m4m3W3c2m2o7q9g5a7g5y7g1k1e1W1y8m5kCW7W3_1g2q1W3k2Y3Y3b5w7Z1W2r1k3Vu2Hw1Jy2HY55k5DeNBi4Z1q5j1m4z1g4~3w6Z3s4r1_2d1i3Lu2BY46m5Gg5Y1Y4Sc2a2y3Km24o13c3Hg2n1c4n1g5B_11u2W1m3y1w4m1W3k1_3My2Km34e25Y4Fu45a42s2Cq2Ig2a1g3Sa2Uw3AY28a32a35e2Jc4Vk4d2_6Ha3Bu31o28_4Gg44o20y27u4Da2h1g3j1m2v1s2l3s3f2W2PY1h1w2Hy2Cm38e1g1e4GY2I_50_C6o7SeDk1gA6u3Fo9Hq5v1kEFi54i2Ia3Og2Ik1c1i2g2c4QY2Qq38q20W2b1g83q12w18u1Is1Kg1KSi6_6_1m2s1W3Mo1c1Y3w1a6q1e8Se8Aa54Y7JiJBi4Fs3r1Y8r1q5~1c5v3_7z1o47c17o2Aw1d5w5X5c6Ba1n2w25Y11q1q4wJg1e6Ga3a2m9s5iRs7mVEw26c20kCCk6Mo5k1k7g2u7o2q6q1_4e2w8u1k8a1m4e1s3u5aLo3YCq1g6i5eJY2Y8m1u4w5gLg1q4Y2s5y1c6yBYh12w7CoE0iE2s54c13K4k86k1c4mFl7s3n2a1o2Y9a3_B";
      URL.params = URL.parse(sample);
      map = Map.attachTo("#qunit-fixture");
      model = new Model();
      route = new Route(map, model);
      route.restoreMerged(URL.params.r);
      deepEqual(model.getLegal(), {
        "copyrights": [
          "Map data ©2013 Google, Sanborn",
          "Map data ©2013 Google"
        ],
        "warnings": [
          "Walking directions are in beta.    Use caution – This route may be missing sidewalks or pedestrian paths."
        ]
      });
      strictEqual(model.getMeters(), 42134);
      strictEqual(model.getUnit(), "mi");
      strictEqual(map.getMapTypeId(), google.maps.MapTypeId.ROADMAP);
      var path = route.lastPoint.entering.polyline.getPath();
      strictEqual(path.getLength(), 422);
      strictEqual(Math.round(google.maps.geometry.spherical.computeLength(path)), 42170);
    });

    test("Snapshot released URL API - no path", function () {
      var sample = "http://onthegomap.com/?m=r&u=mi&lat=42.179496&lng=-71.01273&zoom=10&context=share";
      URL.params = URL.parse(sample);
      map = Map.attachTo("#qunit-fixture");
      model = new Model();
      route = new Route(map, model);

      ok(Util.numbersAreClose(map.getCenter().lat(), 42.179496));
      ok(Util.numbersAreClose(map.getCenter().lng(), -71.01273));
      strictEqual(map.getZoom(), 10);
      strictEqual(model.getUnit(), "mi");
      strictEqual(map.getMapTypeId(), google.maps.MapTypeId.ROADMAP);
      strictEqual(route.lastPoint, null);
    });
  }
);