define([
    'jquery',
    'urlshortener'
  ],

  function (
    $,
    URLShortener) {

    // global gapi
    "use strict";
    var readyCalled = false;
    var failHandler;

    var loadMockGoogleClient = function () {
      window.gapi = {
        client: {
          load: sinon.stub(),
          setApiKey: sinon.stub()
        }
      };
    };
    var execute;
    var successfullyLoadAPI = function () {
      URLShortener.load();
      loadMockGoogleClient();
      window.OTGMgapiclientload();
      window.gapi.client.urlshortener = {
        url: {
          insert: function () {}
        }
      };
      execute = sinon.stub();
      sinon.stub(window.gapi.client.urlshortener.url, 'insert', function () {
        return {
          execute: execute
        };
      });
      gapi.client.load.args[0][2]();
    };

    module("URL Shortener tests", {
      setup: function () {
        var $doc = $(document);
        sinon.stub(window.jQuery, "getScript");
        window.ga = sinon.stub();
        window.gapi = null;
        $doc.off('urlshortenerready');
        readyCalled = false;
        $doc.one('urlshortenerready', function () {
          readyCalled = true;
        });
        $.getScript.returns({
          fail: function (arg) {
            failHandler = arg;
          }
        });
      },
      teardown: function () {
        window.jQuery.getScript.restore();
      }
    });

    //////////////////////////// loading

    test("load script called", function () {
      URLShortener.load();
      deepEqual($.getScript.args, [["https://apis.google.com/js/client.js?onload=OTGMgapiclientload"]], "expected getScript args on load");
      ok(!readyCalled, "ready not called yet");
    });

    test("load script failure", function () {
      URLShortener.load();
      ok(failHandler, "fail() handler should be installed");
      failHandler({}, "Error", "not found");

      ok(!readyCalled, "ready not called yet");
      deepEqual(window.ga.args, [['send', 'event', 'error', 'Load client.js error', "Error not found"]], "load script error logged");
    });

    test("load script sucess - client load called", function () {
      URLShortener.load();
      loadMockGoogleClient();
      window.OTGMgapiclientload();
      strictEqual(gapi.client.load.args[0][0], 'urlshortener', "loader arg 1");
      strictEqual(gapi.client.load.args[0][1], 'v1', "loader arg 2");
      strictEqual(typeof gapi.client.load.args[0][2], 'function', "loader arg 3 type");
      strictEqual(gapi.client.load.args.length, 1, "num times load called");

      ok(!readyCalled, "ready not called yet");
    });

    test("load script error", function () {
      URLShortener.load();
      loadMockGoogleClient();
      window.OTGMgapiclientload();
      gapi.client.load.args[0][2]("error string");

      deepEqual(window.ga.args, [['send', 'event', 'error', 'URL shortener gapi.client load', '"error string"']], "load urlshortener error logged");
      ok(!readyCalled, "ready not called yet");
    });

    test("load urlshortener succes", function () {
      successfullyLoadAPI();
      ok(readyCalled, "ready HAS BEEN called");
    });

    test("Call ready after API loaded", function () {
      var stub = sinon.stub();
      successfullyLoadAPI();
      URLShortener.ready(stub);
      ok(stub.calledOnce, "ready() arg should be called immediately");
    });

    //////////////////////////// invoking API

    test("shorten() prior to API load", function () {
      var stub = sinon.stub();
      URLShortener.shorten("url", stub, stub);
      ok(!stub.called, "no args should have been called");
    });

    test("call shorten() makes request", function () {
      var stub = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten", stub, stub);
      deepEqual(gapi.client.urlshortener.url.insert.args, [[{
        resource: {
          longUrl: "url to shorten"
        }
      }]], "args to url.insert()");
      strictEqual(execute.args.length, 1, "execute() times invoked");
      strictEqual(typeof execute.args[0][0], 'function', "typeof execute() first arg");

      ok(!stub.called, "no args should have been called");
    });

    test("call shorten() before load makes request after load", function () {
      var stub = sinon.stub();
      URLShortener.shorten("url to shorten", stub, stub);
      successfullyLoadAPI();
      strictEqual(gapi.client.urlshortener.url.insert.args.length, 1, "# calls to url.insert()");
      strictEqual(execute.args.length, 1, "execute() times invoked");
      strictEqual(typeof execute.args[0][0], 'function', "typeof execute() first arg");
      ok(!stub.called, "no args should have been called");
    });

    test("shorten success response: logged and succ callback invoked", function () {
      var succ = sinon.stub();
      var err = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten", succ, err);
      execute.args[0][0]({
        id: "short URL"
      });
      deepEqual(succ.args, [["short URL"]], "success callback invoked with");
      ok(!err.called, "error handler should not have been called");
      deepEqual(window.ga.args, [['send', 'event', 'Share Shorten', 'URL Shortener', 'OK']], "use URL shortener logged");
    });

    test("shorten error response: logged and err callback invoked", function () {
      var succ = sinon.stub();
      var err = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten", succ, err);
      execute.args[0][0]({
        "code": 403,
        "message": "Access Not Configured",
        "data": [
          {
            "domain": "usageLimits",
            "reason": "accessNotConfigured",
            "message": "Access Not Configured"
          }
        ],
        "error": {
          "code": 403,
          "message": "Access Not Configured",
          "data": [
            {
              "domain": "usageLimits",
              "reason": "accessNotConfigured",
              "message": "Access Not Configured"
            }
          ]
        }
      });
      ok(err.called, "error callback invoked");
      ok(!succ.called, "success handler should not have been called");
      deepEqual(window.ga.args, [['send', 'event', 'error', 'URL Shortener', '{"code":403,"message":"Access Not Configured","data":[{"domain":"usageLimits","reason":"accessNotConfigured","message":"Access Not Configured"}]}']], "use URL shortener error logged");
    });

    test("shorten error response then re-request same: invoked twice", function () {
      var succ = sinon.stub();
      var err = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten", succ, err);
      execute.args[0][0]({"error": true});
      URLShortener.shorten("url to shorten", succ, err);
      strictEqual(execute.args.length, 2, "execute() times invoked");
    });

    test("shorten in progress: 1 reqest, both callbacks invoked", function () {
      var succ1 = sinon.stub();
      var succ2 = sinon.stub();
      var err = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten", succ1, err);
      URLShortener.shorten("url to shorten", succ2, err);
      strictEqual(execute.args.length, 1, "num API calls made");
      execute.args[0][0]({id: "short URL"});
      strictEqual(succ1.args.length, 1, "callback A invoked once");
      strictEqual(succ2.args.length, 1, "callback B invoked once");
      ok(!err.called, "error handler not invoked");
    });

    test("shorten in progress different: 2 reqests, both callbacks invoked", function () {
      var succ1 = sinon.stub();
      var succ2 = sinon.stub();
      var err = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten1", succ1, err);
      URLShortener.shorten("url to shorten2", succ2, err);
      strictEqual(execute.args.length, 2, "num API calls made");

      execute.args[0][0]({id: "short URL1"});
      deepEqual(succ1.args, [["short URL1"]]);
      deepEqual(succ2.args, []);

      execute.args[1][0]({id: "short URL2"});
      deepEqual(succ1.args, [["short URL1"]]);
      deepEqual(succ2.args, [["short URL2"]]);
      ok(!err.called, "error handler not invoked");
    });


    test("shorten cached same: 1 request, second invoked immediately", function () {
      var succ1 = sinon.stub();
      var succ2 = sinon.stub();
      var err = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten", succ1, err);
      execute.args[0][0]({id: "short URL"});
      URLShortener.shorten("url to shorten", succ2, err);
      strictEqual(execute.args.length, 1, "num API calls made");
      deepEqual(succ1.args, [["short URL"]]);
      deepEqual(succ2.args, [["short URL"]]);
      ok(!err.called, "error handler not invoked");
    });


    test("shorten cached different: 2 requests, both cb's invoked", function () {
      var succ1 = sinon.stub();
      var succ2 = sinon.stub();
      var err = sinon.stub();
      successfullyLoadAPI();
      URLShortener.shorten("url to shorten1", succ1, err);
      execute.args[0][0]({id: "short URL1"});
      URLShortener.shorten("url to shorten2", succ2, err);
      execute.args[1][0]({id: "short URL2"});
      strictEqual(execute.args.length, 2, "num API calls made");
      deepEqual(succ1.args, [["short URL1"]]);
      deepEqual(succ2.args, [["short URL2"]]);
      ok(!err.called, "error handler not invoked");
    });
  }
);